#include <math.h>
#include <stdlib.h>
#include "eigen.h"
#include "deferreur.h"

/* Echange des colonnes a et b (k elements) */
static void swapcol (double *a, double *b, int k)
{
double bid ;
    while (k--) {bid = a[k] ; a[k] = b[k] ; b[k] = bid ; }
}

/* Echange des lignes a et b (k elements) d'une matrice d'ordre n */
static void swaplin (double *a, double *b, int k, int n)
{
int i = 0 ;
double bid ;
    while (k--) {bid = a[i] ; a[i] = b[i] ; b[i] = bid ; i += n ; }
}

/* Division complexe */
static void cpxdiv (double AR, double AI, double BR, double BI,
						double *CR, double *CI)
{
double S = fabs(BR) + fabs(BI) ;
      AR /= S ; AI /= S ; BR /= S ; BI /= S ;
      S = BR*BR + BI*BI ;
      *CR = (AR*BR + AI*BI)/S ;
      *CI = (AI*BR - AR*BI)/S ;
}

/*           BALANCEMENT D'UNE MATRICE (D'apres EISPACK)                    */
void balanc (int N, double *A, int *low, int *igh, double *scale)
/*----------------------------------------------------------------------------
    ENTREE :
	N : ordre de la matrice
	A(N,N) matrice a balancer
    SORTIE :
	LOW et IGH sont tels que A(I,J) == 0 si
	    I > J && (J=1:LOW-1 || I=IGH+1:N
	    cad que des colonnes 1 a LOW-1, A est triangularisee ainsi
	    que des lignes IHG+1 a n.
	SCALE contient l'information permettant de debalancer A
	    Pour J = 1:LOW-1  SCALE(J) = indice colonne echangee avec J
		 J = LOW:IGH  SCALE(J) = elemnt diag utilise
		 J = IGH+1:N  SCALE(J) = indice colonne echangee
	    les echanges etant effecetues de N a IGH+1, puis de 1 a LOW-1.
	Remarque : Si IGH = 0, 1 est renvoye a la place.
----------------------------------------------------------------------------*/
{
int i, j, jj, jcol, k, n, nm1, fconv, flag, np1 ;
double c, f, g, r, s, b2, radix ;

    radix = 16.0 ;
    b2 = radix * radix ;
    n = N ; nm1 = N-1 ; np1 = N+1 ;
  /* Recherche ligne contenant des v.p. isolees de la derniere a la 1ere */
    for (j = nm1, jj = nm1*np1 ; j >= 0 ; j--, jj -= np1)
    {
      flag = 1 ;
      for (i = 0 ; i < n ; i++)
	if (i != j && A[j+i*N] != 0.) {flag = 0 ; break ; }
      if (flag)
      { /* on a isole une v.p. dans ligne j */
	scale[nm1] = j+1 ; /* memo numereo de ligne echangee */
	if (j != nm1)
	{
	/* Echange colonne j et derniere colonne */
	  swapcol (&A[j*N], &A[nm1*N], n) ;
	/* Echange ligne j et derniere ligne */
	  swaplin (&A[j], &A[nm1], N, N) ;
	}
	if(n == 1) { *low = *igh = 1 ; return ; }
	/* Diminue la taille de la matrice de 1 unite et recommence
	   la recherche sur la matrice diminue de ses dernieres li/co */
	j = n = nm1-- ; /* j va subir j-- avant le test de fin de boucle */
      }
    }
/* Recherche colonnes contenanat des v.p. isolees */
    for (j = k = 0, jj = 0 ; j < n ; j++, jj+= np1)
    {
      flag = 1 ;jcol = j*N ;
      for (i = k ; i < n ; i++)
	if (i != j && A[i+jcol] != 0.) { flag = 0 ; break ; }
      if (flag)
      {
	scale[k] = j+1 ; /* memo numero de colonne echangee */
	if (j != k)
	{
	  jcol = k*N ;
	  /* Echange colonnes j et k des lignes 1 a n */
	  swapcol (&A[j*N], &A[jcol], n) ;
	  /* Echange ligne j et k des colonnes k a N */
	  swaplin (&A[j+jcol], &A[k+jcol], N-k, N) ;
	}
	/* Diminue la taille recherche de 1 unite et recommence
	   la recherche sur la matrice diminue de ses premieres li/co */
	j = k++ ;
      }
    }
/* Balance la sous-matrice des lignes k a n */
    for (i = k ; i < n; i++) scale[i] = 1. ;
/* Boucle de reduction de la norme */
    do /* while (fconv) */
    {
      fconv = 0 ;
      for (i = k ; i < n ; i++)
      {
	c = 0. ;
	r = 0. ;
	jcol = i*N ;
	for (j = k ; j < n ; j++)
	  if (j != i) { c += fabs(A[j+jcol]) ; r += fabs(A[i+j*N]) ; }
	/* teste underflow */
	if (c == 0. || r == 0.) continue ;
	g = r / radix ; f = 1. ; s = c + r ;
	while (c < g) { f *= radix ; c *= b2 ; }
	g = r * radix ;
	while (c >= g) { f /= radix ; c /= b2 ; }
	if ((c + r)/f >= 0.95 * s) continue ;
	  fconv = 1 ; scale[i] *= f ;
	  /* Divise la ligne i par f */
	  for (j = k ; j < N ; j++) A[i+j*N] /= f ;
	  /* Multiplie la colonne i par f */
	  jcol = i*N ; for (j = 0 ; j < n ; j++) A[j+jcol] *= f ;
      }
    } while (fconv) ;
   *low = k+1 ; *igh = n ;
   return ;
}

/*          FORME DE HESSENBERG D'UNE MATRICE BALANCEE PAR BALANC           */
/*          calculee par transformations orthogonales                       */
static void orthes (int N, int LOW, int IGH, double *A, double *ORT)
/*----------------------------------------------------------------------------
    ENTREE :
	A matrice NxN orientee colonne
	N : dimension de la matrice A
	LOW et IGH fournis par BALANC (sinon mettre a 1 et N)
    SORTIE :
	A : sous forme de Hessenberg. Les multiplicateurs utilises sont
	    memorises dans le triangle inferieur.
	ORT[IGH] : information sur les substitutions de LOW a IGH
----------------------------------------------------------------------------*/
{
int I, J, M, jcol, mm1 ;
double F,G,H,SCALE ;

      A -= N+1 ; ORT-- ;
      for (M = LOW+1 ; M < IGH ; M++)
      {
	 ORT[M] = H = SCALE = 0. ;
	 mm1 = (M-1)*N ;
	 for (I = M ; I <= IGH ; I++)
	     SCALE = SCALE + fabs(A[I+mm1]) ;
	 if (SCALE == 0.) continue ;
	 for (I = IGH ; I >= M ; I--)
	 {
	    ORT[I] = A[I+mm1] / SCALE ;
	    H += ORT[I] * ORT[I] ;
	 }
	 G = ORT[M] >= 0. ? -sqrt(H) : sqrt(H) ;
	 H -= ORT[M] * G ;
	 ORT[M] -= G ;
/*     .......... FORM (I-(U*UT)/H) * A .......... */

	 for (J = M, jcol=J*N ; J <= N ; J++,jcol+=N)
	 {
	    F = 0.0 ;
	    for (I = IGH ; I >= M ; I--) F += ORT[I] * A[I+jcol] ;
	    F /= H ;
	    for (I = M ; I <= IGH ; I++) A[I+jcol] -= F * ORT[I] ;
	 }
/*     .......... FORM (I-(U*UT)/H)*A*(I-(U*UT)/H) .......... */
	 for (I = 1 ; I <= IGH ; I++)
	 {
	    F = 0. ;
	    for (J = IGH,jcol=J*N ; J >= M ; J--,jcol-=N)
		F += ORT[J] * A[I+jcol] ;
	    F /= H ;
	    for (J = M,jcol=J*N ; J <= IGH ; J++,jcol+=N)
		A[I+jcol] -= F * ORT[J] ;
	 }
	 ORT[M] *= SCALE ;
	 A[M+mm1] = SCALE * G ;
    }
}

/*          FORME DE HESSENBERG D'UNE MATRICE BALANCEE PAR BALANC           */
/*          calculee par transformations elementaires                       */
void elmhes (int N, int LOW, int IGH, double *A, int *INT)
/*----------------------------------------------------------------------------
    ENTREE :
	A matrice NxN orientee colonne
	N : dimension de la matrice A
	LOW et IGH fournis par BALANC (sinon mettre a 1 et N)
    SORTIE :
	A : sous forme de Hessenberg. Les multiplicateurs utilises sont
	    memorises dans le triangle inferieur.
	INT[IGH] : information sur les substitutions de LOW a IGH
----------------------------------------------------------------------------*/
{
int i, j, m, mm1, mcol, mm1col, jcol, icol ;
double x , y ;

      A -= N+1 ; INT-- ;

      for (m = LOW+1 ; m < IGH ; m++)
      {
	mm1 = m-1 ; mcol = m*N ; mm1col = mcol-N ; x = 0. ; i = m ;
	for (j = m ; j <= IGH; j++)
	  if (fabs(A[j+mm1col]) > fabs(x)) { i = j ; x = A[j+mm1col] ; }
	INT[m] = i ;
	if (i != m)
	{
	/* Echange lignes i et m de mm1 a N */
	  swaplin (&A[i+mm1col], &A[m+mm1col], N-mm1+1, N) ;
	/* Echange colonnes i et m  de 1 a IGH */
	  swapcol (&A[1+i*N], &A[1+mcol],IGH) ;
	}

	if (x == 0.) continue ;
	for (i = m+1 ; i <= IGH ; i++)
	{
	  y = A[i+mm1col] ;
	  if (y == 0.) continue ;
	  y /= x ; A[i+mm1col] = y ;
	  for (j = m, jcol = j*N ; j <= N ; j++, jcol += N)
	  { A[i+jcol] -= y * A[m+jcol] ; }
	  icol = i*N ;
	  for (j = 1 ; j <= IGH ; j++)
	  { A[j+mcol] += y * A[j+icol] ; }
	}
    }
    return ;
}

/*--------------------------------------------------------------------------*/
int hqr(int N, int LOW, int IGH, double *H, double *wr, double *wi)
/*--------------------------------------------------------------------------*/
/*   Recherche de valeurs propres d'une matrice sous forme de Hessemberg    */
/*   Cf : "The algebraic eigenvalue problem" J.H.Wilkinson, Oxford, (1965). */
/*   Procedure algol hqr dans num. math. 14, 219-231(1970), Martin, Peters, */
/*   Wilkinson, Handbook for Auto. Comp., vol.II-Linear Algebra, (1971).    */
/*   Procedure fortran hqr dans Eispack guide, B.S. Garbow (1976)           */
/*----------------------------------------------------------------------------
   ENTREE :
	H	Matrice sous forme hessenberg superieure
	N	Ordre de la matrice
	LOW et IGH fournis par BALANC qui a deja isole des v.p. Si BALANC
	n'a pas ete utilise, mettre LOW=1 et IGH=N en argument.
   SORTIE :
	H	H est detruite
	wr, wi  Valeurs propres (parties reelles et imaginaires
	hqr	Valeur renvoyee : Si 0 toutes les v.p. sont bonnes,
		sinon, les v.p. de hqr+1 a n sont les seules bonnes
		apres 30n iterations.
----------------------------------------------------------------------------*/
{
int i, j, ij, IJ, k,l,m,Np1,n,nm1,nm2,itn,its,mp2 ;
int n_n, n1_n1, l_l, m_m, m1_m1, k_k ;
double p,q,r,s,t,w,x,y,z,norm,tst1,tst2 ;

   Np1 = N+1 ;
   H -= Np1 ;                            /* Pour acceder a Hij par H[i+j*N] */
   wr-- ; wi-- ;            /* Pour acceder a WRi et WIi par WR[i] et WI[i] */

   /* Calcul de la norme de la matrice de Hessenberg */
   norm = 0.0 ;
   for (i = k = 1 ; i <= N ; i++)
   {
      for (j = k ; j <= N ; j++) norm += fabs(H[i+j*N]) ;
      k = i ;
      if (i < LOW || i > IGH) { wr[i] = H[i*Np1] ; wi[i] = 0. ; }
   }

   n = IGH ;
   t = 0.0 ;
   itn = 30*N ;
   /* Recherche valeur propre suivante */
   while (n >= LOW)
   {
      its = 0 ;
      n_n = n*Np1 ;
      n1_n1 = n_n - Np1 ;
      nm1 = n - 1 ;
      nm2 = nm1 - 1 ;
      while (1)
      {
	 /* Examine si un element sub-diagonal est negligeable */
	 for (l = n, l_l = n*Np1 ; l > LOW ; l--, l_l -= Np1)
	 {
	    s = fabs(H[l_l-Np1]/*h(l-1,l-1)*/) + fabs(H[l_l]/*h(l,l)*/) ;
	    if (s == 0.0) s = norm ;
	    tst1 = s + fabs(H[l_l-N]/*h(l,l-1)*/) ;
	    if (tst1 == s) break ;
	 }
	 /* si l > 1 : indice ligne element subdiagonal negligeable */
	 /* Forme le decalage */
	 x = H[n_n]/* h(n,n)*/ ;
	 if (l == n) break ;
	 y = H[n1_n1]/*h(n-1,n-1)*/ ;
	 w = H[n1_n1+1]/*h(n,n-1)*/ * H[n1_n1+N]/*h(n-1,n)*/ ;
	 if (l == nm1) break ;
	 if (itn == 0) return n ;
	 if (its == 10 || its == 20)
	 {
	    /* Decalage elements diagonaux -= x (i = 1 a n) */
	    for (ij = LOW*Np1 ; ij <= n_n ; ij += Np1) H[ij]/*h(i,i)*/ -= x ;
	    t += x ;                                 /* Total des decalages */
	    s = fabs(H[n1_n1+1]/*h(n,n-1)*/) + fabs(H[n1_n1-N]/*h(n-1,n-2)*/) ;
	    y = x = 0.75 * s ; w = -0.4375 * s * s ;
	 }
	 its++ ; itn-- ;
	 /* Recherche 2 elements sub-diagonaux successifs negligeables */
	 for (m = nm2, m_m = nm2*Np1 ; m >= l ; m--, m_m -= Np1)
	 {
	    z = H[m_m]/*h(m,m)*/ ;
	    r = x - z ;
	    s = y - z ;
	    p = (r * s - w) / H[m_m+1]/*h(m+1,m)*/ + H[m_m+N]/*h(m,m+1)*/ ;
	    m1_m1 = m_m + Np1 ;
	    q = H[m1_m1]/*h(m+1,m+1)*/ - z - r - s ;
	    r = H[m1_m1+1]/*h(m+2,m+1)*/ ;
	    s = fabs(p) + fabs(q) + fabs(r) ;
	    p /= s ; q /= s ; r /= s ;
	    if (m == l) break ;
	    tst1 = fabs(p)* ( fabs(H[m_m-Np1]/*h(m-1,m-1)*/) + fabs(z) +
					     fabs(H[m1_m1]/*h(m+1,m+1)*/) ) ;
	    tst2 = tst1 + fabs(H[m_m-N]/*h(m,m-1)*/)*( fabs(q) + fabs(r) ) ;
	    if (tst2 == tst1) break ;
	 }
	 /* Si m >= l : indice ligne du couple d'el. negligeables */
	 /* Raz 2eme et 3eme sous-diagonales */
	 for (i = mp2 = m+2 ; i <= n ; i++)
	 {
	    ij = i+(i-2)*N ;
	    H[ij]/*h(i,i-2)*/ = 0.0 ;
	    if (i != mp2) H[ij-N]/*h(i,i-3)*/ = 0.0 ;
	 }
	 /* Double QR pour les lignes l -> n et les colonnes m -> n */
	 for (k = m, k_k = m_m ; k <= nm1 ; k++, k_k += Np1)
	 {
	  ij = k_k-N;
	  if (k != nm1) /* Not the last */
	  {
	    if (k != m)
	    {
	       p = H[ij]/*h(k,k-1)*/ ;
	       q = H[ij+1]/*h(k+1,k-1)*/ ;
	       r = H[ij+2]/*h(k+2,k-1)*/ ;
	       x = fabs(p) + fabs(q) + fabs(r) ;
	       if (x == 0.0) continue ;
	       p /= x ; q /= x ; r /= x ;
	    }
	    s = sqrt(p*p+q*q+r*r) ;
	    if (p < 0.) s = -s ;
	    if (k != m) H[ij]/*h(k,k-1)*/ = -s * x ;
	    else if (l != m) H[ij]/*h(k,k-1)*/ = -H[ij] ;
	    p += s ;
	    x = p / s ; y = q / s ; z = r / s ;
	    q /= p ; r /= p ;
	    /* modification ligne */
	    for (j = k, ij = k_k ; j <= n ; j++, ij += N)
	    {
	       p = H[ij]/*h(k,j)*/ + q * H[ij+1]/*h(k+1,j)*/
		   + r * H[ij+2]/*h(k+2,j)*/ ;
	       H[ij]/*h(k,j)*/ -= p * x ;
	       H[ij+1]/*h(k+1,j)*/ -= p * y ;
	       H[ij+2]/*h(k+2,j)*/ -= p * z;
	    }
	    j = n < k+3 ? n : k+3 ;
	    /* modification colonne */
	    for (i = l, ij = l+k*N ; i <= j ; i++, ij++)
	    {
	       IJ = ij+N ;
	       p = x * H[ij]/*h(i,k)*/ + y * H[IJ]/*h(i,k+1)*/
		   + z * H[IJ+N]/*h(i,k+2)*/ ;
	       H[ij]/*h(i,k)*/ -= p ;
	       H[IJ]/*h(i,k+1)*/ -= p * q ;
	       H[IJ+N]/*h(i,k+2)*/ -= p * r;
	    }
	  }
	  else
          {
	    if (k != m)
	    {
	       p = H[ij]/*h(k,k-1)*/ ;
	       q = H[ij+1]/*h(k+1,k-1)*/ ;
	       x = fabs(p) + fabs(q) ;
	       if (x == 0.0) continue ;
	       p /= x ; q /= x ;
	    }
	    s = sqrt(p*p+q*q) ;
	    if (p < 0.) s = -s ;
	    if (k != m) H[ij]/*h(k,k-1)*/ = -s * x ;
	    else if (l != m) H[ij]/*h(k,k-1)*/ = -H[ij] ;
	    p += s ;
	    x = p / s ; y = q / s ; q /= p ;
	    /* modification ligne */
	    for (j = k, ij = k_k ; j <= n ; j++, ij += N)
	    {
	       p = H[ij]/*h(k,j)*/ + q * H[ij+1]/*h(k+1,j)*/ ;
	       H[ij]/*h(k,j)*/ -= p * x ;
	       H[ij+1]/*h(k+1,j)*/ -= p * y ;
	    }
	    j = n < k+3 ? n : k+3 ;
	    /* modification colonne */
	    for (i = l, ij = l+k*N ; i <= j ; i++, ij++)
	    {
	       IJ = ij+N ;
	       p = x * H[ij]/*h(i,k)*/ + y * H[IJ]/*h(i,k+1)*/ ;
	       H[ij]/*h(i,k)*/ -= p ;
	       H[IJ]/*h(i,k+1)*/ -= p * q ;
	    }
	  }
	 }
      } /* Fin du while (itn). Sortie avec l = n ou l = nm1 ou itn = 0 */
      if (l == n)
      {
      /* trouve une valeur propre */
	 wr[n] = x + t ;
	 wi[n] = 0.0 ;
	 n = nm1 ;
      }
      else
      {
      /* Trouve deux val. propres */
	 p = (y - x) / 2.0 ;
	 q = p * p + w ;
	 z = sqrt(fabs(q)) ;
	 x = x + t ;
	 if (q >= 0.0)
	 {
	 /* couple reel */
	    z = p > 0. ? p+z : p-z ;
	    wr[nm1] = x + z ;
	    wr[n] = wr[nm1] ;
	    if (z != 0.0) wr[n] = x - w / z ;
	    wi[nm1] = wi[n] = 0.0 ;
	 }
	 else
	 {
	 /* couple complexe */
	    wr[nm1] = wr[n] = x + p ;
	    wi[nm1] = z ; wi[n] = -z ;
	 }
	 n = nm2 ;
      }
   } /* fin du while (n >= LOW) */
   return n ;
}


/* Accumulation dans Z des transformations effectuees par elmhes */
static void eltran (int N, int LOW, int IGH, double *A, int *INT, double *Z)
{
int i, j, jcol, kl, mm, mp, mpN, mpm1N ;

    Z -= N+1 ; A -= N+1 ; INT-- ;            /* Pour simplifier l'adressage */
/* Matrice identite */
    for (j = 1 ; j <= N ; j++)
    {
      jcol = j*N ; for (i = 1 ; i <= N ; i++) Z[i+jcol] = 0. ;
      Z[j+jcol] = 1. ;
    }
    kl = IGH - LOW - 1 ;
    for (mm = 1 ; mm <= kl ; mm++)
    {
      mp = IGH - mm ; mpN = mp*N ; mpm1N = mpN - N ;
      for (i = mp+1 ; i <= IGH ; i++) Z[i+mpN] = A[i+mpm1N] ;
      i = INT[mp] ;
      if (i == mp) continue ;
      for (j = mp, jcol = j*N ; j <= IGH ; j++, jcol += N)
      { Z[mp+jcol] = Z[i+jcol] ; Z[i+jcol] = 0. ; }
      Z[i+mpN] = 1. ;
    }
}

/* Accumulation dans Z des transformations effectuees par orthes */
static void ortran (int N, int LOW, int IGH, double *A, double *ORT, double *Z)
{
int I,J,M,mm1,jcol;
double G ;
  /* Matrice Z unite */
    /* raz Z */  mm1 = N*N-1 ; for (J = 1 ; J < mm1 ; J++) Z[J] = 0. ;
    /* Z = 1 */  M = N+1 ; for (J = 0 ; J <= mm1 ; J+=M) Z[J] = 1. ;
    Z -= N+1 ; A -= N+1 ; ORT-- ;            /* Pour simplifier l'adressage */
    for (M = IGH-1, mm1 = (M-1)*N ; M > LOW ; M--, mm1 -= N)
    {
	if (A[M+mm1] == 0.) continue ;
	for (I = M+1 ; I <= IGH ; I++) ORT[I] = A[I+mm1] ;
	for (J = M, jcol = J*N ; J <= IGH ; J++, jcol+=N)
	{
	    G = 0. ;
	    for (I = M ; I <= IGH ; I++) G += ORT[I] * Z[I+jcol] ;
	    G = (G / ORT[M]) / A[M+mm1] ;
	    for (I = M ; I <= IGH ; I++) Z[I+jcol] += G * ORT[I] ;
	}
    }
}

static void balbak (int N, int LOW, int IGH, double *scale, int M, double *Z)
{
/* Calcule les vecteurs propres d'une matrice a partir de ceux de sa matrice
   prealablement transformee par balanc. */
int i, j, ii, k ;
double s ;
    scale-- ; Z -= N+1 ;          /* pour simplifier le calcul des adresses */
    if (M <= 0) return ;
    if (IGH != LOW) for (i = LOW ; i <= IGH ; i++)
    { s = scale[i] ; for (j = 1 ; j <= M ; j++) Z[i+j*N] *= s ; }
    for (ii = 1 ; ii <= N ; ii++)
    {
      i = ii ;
      if (i >= LOW && i <= IGH) continue ;
      if (i < LOW) i = LOW - ii ;
      k = (int) scale[i] ;
      if (k != i) swaplin (&Z[i+N], &Z[k+N], M, N) ;
    }
}

/*--------------------------------------------------------------------------*/
static int hqr2(int N, int LOW, int IGH, double *H, double *wr, double *wi, double *V)
/*--------------------------------------------------------------------------*/
/*   Recherche de valeurs et propres et des vecteurs propres d'une matrice  */
/*   sous forme de Hessemberg                                               */
/*----------------------------------------------------------------------------
   ENTREE :
	H	Matrice sous forme hessenberg superieure
	N	Ordre de la matrice
	LOW et IGH fournis par BALANC qui a deja isole des v.p. Si BALANC
	n'a pas ete utilise, mettre LOW=1 et IGH=N en argument.
	V       matrice de la transformation produite par eltran (Fournir
		une matrice unite si pas de transformation par balanc/elmhes)
   SORTIE :
	H	H est detruite
	wr, wi  Valeurs propres (parties reelles et imaginaires
	hqr	Valeur renvoyee : Si 0 toutes les v.p. sont bonnes,
		sinon, les v.p. de hqr+1 a n sont les seules bonnes
		apres 30n iterations.
	V       matrice des vecteurs propres
----------------------------------------------------------------------------*/
{
int i, j, jN, ij, IJ, k,ik, l,m,Np1,n,nN,nm1,nm1N, nm2,itn,its,mp2 ;
int n_n, n1_n1, i_i, l_l, m_m, m1_m1, k_k ;
double p,q,r,s,t,w,x,y,z,ra,sa,norm,tst1,tst2,vr,vi ;
   Np1 = N+1 ;
   H -= Np1 ; V -= Np1 ;    /* Pour acceder a Hij at Vij par H[i+j*N] ...   */
   wr-- ; wi-- ;            /* Pour acceder a WRi et WIi par WR[i] et WI[i] */

   /* Calcul de la norme de la matrice de Hessenberg */
   norm = 0.0 ;
   for (i = k = 1 ; i <= N ; i++)
   {
      for (j = k ; j <= N ; j++) norm += fabs(H[i+j*N]) ;
      k = i ;
      if (i < LOW || i > IGH) { wr[i] = H[i*Np1] ; wi[i] = 0. ; }
   }
   if (IGH < LOW || norm == 0.)  return 0 ;
   n = IGH ;
   t = 0.0 ;
   itn = 30*N ;
   /* Recherche valeur propre suivante */
   while (n >= LOW)
   {
      its = 0 ;
      n_n = n*Np1 ;
      nN = n*N ;
      n1_n1 = n_n - Np1 ;
      nm1 = n - 1 ;
      nm1N = nm1*N ;
      nm2 = nm1 - 1 ;
      while (1)
      {
	 /* Examine si un element sub-diagonal est negligeable */
	 for (l = n, l_l = l*Np1 ; l > LOW ; l--, l_l -= Np1)
	 {
	    s = fabs(H[l_l-Np1]) + fabs(H[l_l]) ;
	    if (s == 0.0) s = norm ;
	    tst1 = s + fabs(H[l_l-N]) ;
	    if (tst1 == s) break ;
	 }
	 /* si l > 1 : indice ligne element subdiagonal negligeable */
	 /* Forme le decalage */
	 x = H[n_n] ;
	 if (l == n) break ;
	 y = H[n1_n1] ;
	 w = H[n1_n1+1] * H[n1_n1+N] ;
	 if (l == nm1) break ;
	 if (itn == 0) return n ;
	 if (its == 10 || its == 20)
	 {
	    /* Decalage elements diagonaux -= x (i = 1 a n) */
	    for (ij = LOW*Np1 ; ij <= n_n ; ij += Np1) H[ij] -= x ;
	    t += x ;                                 /* Total des decalages */
	    s = fabs(H[n1_n1+1]) + fabs(H[n1_n1-N]) ;
	    y = x = 0.75 * s ; w = -0.4375 * s * s ;
	 }
	 its++ ; itn-- ;
	 /* Recherche 2 elements sub-diagonaux successifs negligeables */
	 for (m = nm2, m_m = m*Np1 ; m >= l ; m--, m_m -= Np1)
	 {
	    z = H[m_m] ;
	    r = x - z ;
	    s = y - z ;
	    p = (r * s - w) / H[m_m+1] + H[m_m+N] ;
	    m1_m1 = m_m + Np1 ;
	    q = H[m1_m1] - z - r - s ;
	    r = H[m1_m1+1] ;
	    s = fabs(p) + fabs(q) + fabs(r) ;
	    p /= s ; q /= s ; r /= s ;
	    if (m == l) break ;
	    tst1 = fabs(p)* ( fabs(H[m_m-Np1]) + fabs(z) + fabs(H[m1_m1]) ) ;
	    tst2 = tst1 + fabs(H[m_m-N])*( fabs(q) + fabs(r) ) ;
	    if (tst2 == tst1) break ;
	 }
	 /* Si m >= l : indice ligne du couple d'el. negligeables */
	 /* Raz 2eme et 3eme sous-diagonales */
	 for (i = mp2 = m+2 ; i <= n ; i++)
	 {
	    ij = i+(i-2)*N ;
	    H[ij] = 0.0 ;
	    if (i != mp2) H[ij-N] = 0.0 ;
	 }
	 /* Double QR pour les lignes l -> n et les colonnes m -> n */
	 for (k = m, k_k = m_m ; k <= nm1 ; k++, k_k += Np1)
	 {
	  ij = k_k-N;
	  if (k != nm1) /* Not the last */
	  {
	    if (k != m)
	    {
	       p = H[ij] ;
	       q = H[ij+1] ;
	       r = H[ij+2] ;
	       x = fabs(p) + fabs(q) + fabs(r) ;
	       if (x == 0.0) continue ;
	       p /= x ; q /= x ; r /= x ;
	    }
	    s = sqrt(p*p+q*q+r*r) ;
	    if (p < 0.) s = -s ;
	    if (k != m) H[ij] = -s * x ;
	    else if (l != m) H[ij] = -H[ij] ;
	    p += s ;
	    x = p / s ; y = q / s ; z = r / s ;
	    q /= p ; r /= p ;
	    /* modification ligne */
	    for (j = k, ij = k_k ; j <= N ; j++, ij += N)
	    {
	       p = H[ij] + q * H[ij+1] + r * H[ij+2] ;
	       H[ij] -= p*x ;
	       H[ij+1] -= p*y ;
	       H[ij+2] -= p*z;
	    }
	    j = n < k+3 ? n : k+3 ;
	    /* modification colonne */
	    for (i = 1, ik = i+k*N ; i <= j ; i++, ik++)
	    {
	       IJ = ik+N ;
	       p = x * H[ik] + y * H[IJ] + z * H[IJ+N] ;
	       H[ik] -= p ;
	       H[IJ] -= p*q ;
	       H[IJ+N] -= p*r;
	    }
	    /* accumulation des transformation */
	    for (i = LOW, ik = i+k*N ; i <= IGH ; i++, ik++)
	    {
	       IJ = ik + N ;
	       p = x * V[ik] + y * V[IJ] + z * V[IJ+N] ;
	       V[ik] -= p ;
	       V[IJ] -= p*q ;
	       V[IJ+N] -= p*r;
	    }
	  }
	  else
	  {
	    if (k != m)
	    {
	       p = H[ij] ;
	       q = H[ij+1] ;
	       x = fabs(p) + fabs(q) ;
	       if (x == 0.0) continue ;
	       p /= x ; q /= x ;
	    }
	    s = sqrt(p*p+q*q) ;
	    if (p < 0.) s = -s ;
	    if (k != m) H[ij] = -s * x ;
	    else if (l != m) H[ij] = -H[ij] ;
	    p += s ;
	    x = p / s ; y = q / s ; q /= p ;
	    /* modification ligne */
	    for (j = k, ij = k_k ; j <= N ; j++, ij += N)
	    {
	       p = H[ij] + q * H[ij+1] ;
	       H[ij] -= p * x ;
	       H[ij+1] -= p * y ;
	    }
	    j = n < k+3 ? n : k+3 ;
	    /* modification colonne */
	    for (i = 1, ik = i+k*N ; i <= j ; i++, ik++)
	    {
	       IJ = ik+N ;
	       p = x * H[ik] + y * H[IJ] ;
	       H[ik] -= p ;
	       H[IJ] -= p*q ;
	    }
	    /* accumulation des transformation */
	    for (i = LOW, ik = i+k*N ; i <= IGH ; i++, ik++)
	    {
	       IJ = ik + N ;
	       p = x * V[ik] + y * V[IJ] ;
	       V[ik] -= p ;
	       V[IJ] -= p*q ;
	    }
	  }
	 }
      } /* Fin du while (itn). Sortie avec l = n ou l = nm1 ou itn = 0 */
      if (l == n)
      {
      /* trouve une valeur propre */
	 H[n_n] = wr[n] = x + t ;
	 wi[n] = 0.0 ;
	 n = nm1 ;
      }
      else
      {
      /* Trouve deux val. propres */
	 p = (y - x) / 2.0 ;
	 q = p * p + w ;
	 z = sqrt(fabs(q)) ;
	 x = H[n_n] = x + t ;
	 H[n1_n1] = y + t ;
	 if (q >= 0.0)
	 {
	 /* couple reel */
	    z = p > 0. ? p+z : p-z ;
	    wr[nm1] = x + z ;
	    wr[n] = wr[nm1] ;
	    if (z != 0.0) wr[n] = x - w / z ;
	    wi[nm1] = wi[n] = 0.0 ;
	    x = H[n1_n1+1] ;
	    s = fabs(x) + fabs(z) ;
	    p = x / s ;
	    q = z / s ;
	    r = sqrt(p*p+q*q) ;
	    p /= r ; q /= r ;
	  /* modif ligne */
	    for (j = nm1, ij = n1_n1 ; j <= N ; j++, ij += N)
	    {
	       IJ = ij+1 ;
	       z = H[ij] ; H[ij] = q * z + p * H[IJ] ;
	       H[IJ] = q * H[IJ] - p * z ;
	    }
	  /* modif colonne */
	    for (i = 1, ij = i+nm1N ; i <= n ; i++, ij++)
	    {
	       IJ = ij + N ;
	       z = H[ij] ;
	       H[ij] = q * z + p * H[IJ] ;
	       H[IJ] = q * H[IJ] - p * z ;
	    }
	  /* Accumule les transfos */
	    for (i = LOW, ik = i + nm1N ; i <= IGH ; i++, ik++)
	    {
	       IJ = ik + N ;
	       z = V[ik] ;
	       V[ik] = q * z + p * V[IJ] ;
	       V[IJ] = q * V[IJ] - p * z ;
	    }
	 }
	 else
	 {
	 /* couple complexe */
	    wr[nm1] = wr[n] = x + p ;
	    wi[nm1] = z ; wi[n] = -z ;
	 }
	 n = nm2 ;
      }
   } /* fin du while (n >= LOW) */

/* Calcul des vecteurs propres */
   for (n = N ; n >= 1 ; n--)
   {
     nm1 = n - 1 ;
     nN = n * N ;
     nm1N = nN - N ;
     n_n = nN + n ;
     n1_n1 = n_n - Np1 ;
     p = wr[n] ; q = wi[n] ;
     if (q > 0.) continue ; /* Si v.p. X+iY et X-iY, ne traite que X-iY */
     else if (q == 0.0)
     { /* Vecteur reel */
	m = n ;
	H[n_n] = 1. ;
	for (i = nm1, i_i = i*Np1 ; i >= 1 ; i--, i_i -= Np1)
	{
	  w = H[i_i] - p ;
	  r = 0. ;
	  for (j = m, ij = i+j*N ; j <= n ; j++, ij += N)
	    r += H[ij] * H[j+nN] ;
	  if (wi[i] < 0.) { z = w ; s = r ; }
	  else
	  {
	    m = i ;
	    if (wi[i] == 0.)
	    {
	      if (w != 0.) t = w ;
	      else
	      { t = tst1 = norm ;
		do { t *=  0.01 ; tst2 = norm + t ; } while (tst2 > tst1) ;
	      }
	      H[i+nN] = -r/t ;
	    } /* fin wi[i] = 0. */
	    else
	    { /* resolution equation reelle */
	      /* [ w  x ] ( h(i,n) )   (-r)  */
	      /* [ y  z ]*(h(i+1,n)) = (-s) */
	      x = H[i_i+N] ; y = H[1+i_i] ;
	      q = wr[i] - p ;
	      q = q * q + wi[i] * wi[i] ;
	      H[i+nN] = t = (x * s - z * r) / q ;
	      H[i+1+nN] = fabs(x) > fabs(z)
			? (-r - w * t) / x : (-s - y * t) / z ;
	    } /* fin wi[i] > 0. */
	    /* Verification overflow */
	    t = fabs(H[i+nN]) ;
	    if (t == 0.0) continue ;
	    tst1 = t ;
	    tst2 = tst1 + 1.0/tst1 ;
	    if (tst2 > tst1) continue ;
	    for (j = i ; j <= n ; j++) H[j+nN] /= t ;
	  } /* fin wi[i] >= 0. */
	} /* fin for i = n-1 -> 1 */
    } /* fin q == 0. c-a-d du cas vecteur reel */
    else
    { /* Cas vecteur complexe (q < 0) */
      m = nm1 ;
/* Dernieres composantes imaginaires  =>  la mat. des v.p. triangulaire */
      if (fabs(H[1+n1_n1]) > fabs(H[n1_n1+N]))
      {
	 H[n1_n1] = q / H[1+n1_n1] ;
	 H[n1_n1+N] = -(H[n_n] - p) / H[1+n1_n1] ;
      }
      else
      {
	 cpxdiv(0.0,-H[n1_n1+N],H[n1_n1]-p, q, &H[n1_n1], &H[n1_n1+N]) ;
      }
      H[1+n1_n1] = 0.0 ;
      H[n_n] = 1.0 ;
      nm2 = nm1 - 1 ;
      for (i = nm2, i_i = i*Np1 ; i >= 1 ; i--, i_i -= Np1)
      {
	w = H[i_i] - p ; ra = sa = 0.0 ;
	for (j = m, ij = i+j*N ; j <= n ; j++, ij += N)
	{
	  ra += H[ij] * H[j+nm1N] ;
	  sa += H[ij] * H[j+nN] ;
	}
	if (wi[i] < 0.0)
	{ z = w ; r = ra ; s = sa ; }
	else
	{
	  m = i ;
	  if (wi[i] == 0.)
	    cpxdiv(-ra,-sa,w,q,&H[i+nm1N],&H[i+nN]) ;
	  else
	  {
	  /* resolution eq. cpx */
	    x = H[i_i+N] ; y = H[1+i_i] ;
	    vr = wr[i] - p ;
	    vi = vr * 2.0 * q ;
	    vr = vr * vr + wi[i] * wi[i] - q * q ;
	    if (vr == 0.0 && vi == 0.0)
	    {
	      vr = tst1 =
		norm * (fabs(w) + fabs(q) + fabs(x) + fabs(y) + fabs(z)) ;
	      do { vr *= 0.01 ; tst2 = tst1 + vr ; } while (tst2 > tst1) ;
	    }
	    cpxdiv (x*r-z*ra+q*sa,x*s-z*sa-q*ra,vr,vi,&H[i+nm1N],&H[i+nN]) ;
	    if (fabs(x) > fabs(z) + fabs(q))
	    {
	      H[i+1+nm1N] = (-ra - w * H[i+nm1N] + q * H[i+nN]) / x ;
	      H[i+1+nN] = (-sa - w * H[i+nN] - q * H[i+nm1N]) / x ;
	    }
	    else
	      cpxdiv(-r-y*H[i+nm1N],-s-y*H[i+nN],z,q,&H[i+1+nm1N],&H[i+1+nN]) ;
	  } /* fin cas wi[i] != 0. */
	/* Test overflow */
	  t = fabs(H[i+nm1N]) ;
	  tst1 = fabs(H[i+nN]) ;
	  if (tst1 > t) t = tst1 ;
	  if (t != 0.0)
	  {
	    tst1 = t ;
	    tst2 = tst1 + 1.0/tst1 ;
	    if (tst2 <= tst1) for (j = i; j <= n ; j++)
	      { H[j+nm1N] /= t ; H[j+nN] /= t ; }
	  }
	}
      } /* fin for i = nm2 -> 1 */
    } /* fin q < 0. c-a-d du cas vecteur cpx */
  } /* fin for n = N -> 1 */

/* Vect.p. des val.p. isolees */
  for (i = 1 ; i <= N ; i++)
  { if (i < LOW || i > IGH) for (j = i, ij = i+j*N ; j <= N ; j++, ij += N)
			    V[ij] = H[ij] ;
  }

/* Retour a la matrice originale */
  for (j = N ; j >= LOW ; j--)
  {
    m = j < IGH ? j : IGH ;
    jN = j*N ; itn = jN - N ;
    for (i = LOW ; i <= IGH ; i++)
    {
	z = 0.0 ;
	for (k = LOW, ik = i+k*N ; k <= m ; k++, ik += N)
	  z += V[ik] * H[k+jN] ;
	V[i+jN] = z ;
    }
  }
  return n ;
}

/*--------------------------------------------------------------------------*/
/*               VALEURS PROPRES D'UNE MATRICE CARREE                       */
/*----------------------------------------------------------------------------
   Entree :                                                                  
  	    n	    ordre de la matrice (n x n)                                  
        A       matrice carree  
        mode	indicateur de la methode : 
               si mode & 1 => il y aura balancement prealable  
                si mode & 2 => transformations par orthogonalisation (+lent) 
  		sinon              //          elementaires (+rapide) 
   Sortie :    
        eig	0 si convergence dans le calcul des valeurs propres. Sinon 
                nombre de valeurs calculees avant abandon du calcul. 
        wr,wi   partie reelle et imaginaire des valeurs propres  
  	Retour :
		 0  : toutes les v.p. sont bonnes,
		r>0 : les v.p. de r+1 a n sont les seules bonnes apres 30n iterations
		 -1 : echec allocation
/*--------------------------------------------------------------------------*/
int eigval (double *wr, double *wi, double *A, int n, int mode)
{
double *scl, *ort ;
int I, J, *iv ;

	scl = (double *) malloc (2*n*sizeof(double));
	if (scl == NULL) { perreur("Echec allocation dans eigval"); return -1 ;}

    if (mode & 1) balanc (n, A, &I, &J, scl) ;
    else { I = 1 ; J = n ; }

    if (mode & 2)
    {
		ort = scl + n ;
		orthes (n, I, J, A, ort) ;
    }
    else
    {
		iv = (int *) (scl + n) ;
		elmhes (n, I, J, A, iv) ;
    }
	free(scl);
    return hqr (n, I, J, A, wr, wi) ;
}

/*--------------------------------------------------------------------------*/
/*      VALEURS PROPRES ET VECTEURS PROPRES D'UNE MATRICE CARREE            */
/*----------------------------------------------------------------------------
   Entree :                                                                   
  		n	ordre de la matrice (n x n) 
        A       matrice carree  
        mode	indicateur de la methode : 
                si mode & 1 => il y aura balancement prealable 
                si mode & 2 => transformations par orthogonalisation (+lent) 
  		sinon              //          elementaires (+rapide) 
                si mode & 4 => unitarisation des vecteurs propres 
   Sortie : 
        eig	0 si convergence dans le calcul des valeurs propres. Sinon 
                nombre de valeurs calculees avant abandon du calcul. 
        wr,wi   partie reelle et imaginaire des valeurs propres 
        A	matrice bloc diagonale avec les valeurs propres reelles sur 
                la diagonale ou un bloc du type    a  b 
                                                  -b  a     si a + i b est 
                valeur propre complexe. 
        M       matrice des vecteures propres correspondants. Dans le cas  
                valeurs propres complexes conjugues, M contient d'abord le 
                vecteur partie reelle et apres le vecteur partie imaginaire 
                Cette disposition est telle que M * A * inv(M) redonne la 
                matrice A initiale. 
  	Retour :
		 0  : toutes les v.p. sont bonnes,
		r>0 : les v.p. de r+1 a n sont les seules bonnes apres 30n iterations
		 -1 : echec allocation
/*--------------------------------------------------------------------------*/
int eigvec (double *wr, double *wi, double *M, double *A, int n, int mode) 
{
double tot, *scl, *ort ;
int i, I, J, Jf, *iv ;

	scl = (double *) malloc (2*n*sizeof(double));
	if (scl == NULL) { perreur("Echec allocation dans eigvec"); return -1 ;}

    if (mode & 1) balanc (n, A, &I, &J, scl) ;
    else { I = 1 ; J = n ; }

    if (mode & 2)
    {
		ort = scl+n ;
		orthes (n, I, J, A, ort) ;
		ortran (n, I, J, A, ort, M) ;
    }
    else
    {
		iv = (int *) (scl+n) ;
		elmhes (n, I, J, A, iv) ;
		eltran (n, I, J, A, iv, M) ;
    }

    i = hqr2 (n, I, J, A, wr, wi, M) ;
    if (mode & 1)
    {
		balbak (n, I, J, scl, n, M) ;
    }
    if (i) {free(scl); return i;} 
  /* Unitarise les vecteurs propres */
    if (mode & 4)
    {
	for (i = I = 0 ; i < n ; i++)
	{

	    Jf = I+n ;
	    if (wi[i] != 0.) { Jf += n ; i++ ; }
	    {
		tot = 0 ;
		for (J = I ; J < Jf ; J++) tot += M[J]*M[J] ;
		tot = sqrt(tot) ;
		for (J = I ; J < Jf ; J++) M[J] /= tot ;
	    }
	    I = Jf ;
	}
    }
 /* Matrice bloc diagonale */
    I = n*n ;
    for (i = 0 ; i < I ; i++) A[i] = 0. ;
    for (i = I= 0 ; i < n ; i++, I += n+1) A[I] = wr[i] ;
    for (i = I= 0 ; i < n ; i++, I += n+1)
    if (wi[i] != 0.) {A[I+1] = wi[i+1] ; A[I+n] = wi[i] ; i++ ; I += n+1 ; }

	free(scl);
    return 0 ;
}

