/*-------------------------------------------------------------------------*/
/*                        Routines d'algebre lineaire                      */
/*-------------------------------------------------------------------------*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "linsolve.h"
#include "deferreur.h"

/*-------------------------------------------------------------------------*/
/* Pour acceder en global au nombre d'iteration max pour convergence SVD   */
/* decommenter la ligne suivante                                           */
/* #define ACCES_NBITERMAX */
/*-------------------------------------------------------------------------*/
#ifdef ACCES_NBITERMAX
int nbitermaxsvd;        /* On le met en global. Il est affecte dans svdc. */
#endif
/*-------------------------------------------------------------------------*/
/*                          Utilitaires internes                           */
/*-------------------------------------------------------------------------*/
static void cum1vec (double *y, double a, double *x, int n)
{
/* Cumulation de vecteurs */
    if (n > 0 && a != 0.) while (n--) y[n] += a * x[n] ;
}
/*-------------------------------------------------------------------------*/
static void sca1vec (double *x, double a, int n)
{ /* multiplie un vecteur par une constante */

    if (n > 0) while (n--) x[n] = a * x[n] ;
}
/*-------------------------------------------------------------------------*/
static void shholder (double *y, double *x, int n)
{
/*
Transformation de Householder reduite :  y = y - x *[(x'*y)/x[0]]
*/
int i ;
double t = 0. ;
    for (i = 0 ; i < n ; i++) t += x[i] * y[i] ;
    t = -t / *x ;
    while (n--) y[n] += t * x[n] ;
    return ;
}
/*-------------------------------------------------------------------------*/
static void srot (double *x, double *y, double c, double s, int n)
{ /* Rotation plane de n couples x,y */
double temp ;

    while (n--)
    {
		temp = c * x[n] + s * y[n] ;
		y[n] = c * y[n] - s * x[n] ;
		x[n] = temp ;
    }
}
/*-------------------------------------------------------------------------*/
static void srotg (double *sa, double *sb, double *c, double *s)
{ /* Construction d'une rotation plane de Givens */
double asa, asb, rho, fact, r, z ;

    asa = *sa ;
    asb = *sb ;
    if (asa < 0.) asa = -asa ;
    if (asb < 0.) asb = -asb ;
    if (asa > asb) rho = *sa ; else rho = *sb ;
    fact = asa + asb ;
    if (fact == 0.) { *c = 1. ; *s = 0. ; r = 0. ; }
    else
    {
		r = sqrt (asa*asa + asb*asb) ;
		if (rho < 0.) r = - r ;
		*c = *sa / r ;
		*s = *sb / r ;
    }
    z = 1.0 ;
    if (asa > asb) z = *s ;
    if (asb >= asa && *c != 0.) z = 1./ *c ;
    *sa = r ;
    *sb = z ;
}
/*--------------------------------------------------------------------*/
static double lufactor (double *a, int lda, int n, int *ipvt)
{
/*
     lufactor effectue la factorisation L U d'une matrice n x n
     extraite d'une matrice rangees par vecteurs de lda composantes.
     Les facteurs sont renvoyes dans A et les indices des pivots
     des lignes dans ipvt.
     Renvoi le determinant

     Adapte en C par MLL le 28/04/92 d'aprs
     LINPACK. 08/14/78 CLEVE MOLER, UNIVERSITY OF NEW MEXICO
     ARGONNE NATIONAL LAB.
*/

double t, det=1. ;
int j, k, l, nm1, lk, kk, lj, kj, klda;
/* int info; */ /* indice fortran du 1er element nul de la diag */
int lda1 = 1 + lda ;

/*  Elimination de Gauss avec pivotage partiel */

/*    info = 0 ; */
    nm1 = n - 1 ;
    kk = 0 ;
    if (nm1 >= 1)
    {
		for (k = 0 , klda = 0 , kk = 0 ; k < nm1 ;
						 k++ , klda += lda, kk += lda1)
		{
		/*  Recherche l = indice du pivot */
			l = iamaxvec (&a[kk], n-k) + k ;
			ipvt[k] = l ;
		/* Si 0 la colonne est deja triangularisee */
			lk = l + klda ;
			if ( a[lk] != 0.)
			{
			if (l != k) { t = a[lk] ; a[lk] = a[kk] ; a[kk] = t; det = -det;}
			det *= a[kk];
			/* Calcul des multiplicateurs */
			t = -1./a[kk] ;
			sca1vec (&a[kk+1], t, nm1-k) ;
			/* Elimination des colonnes */
			lj = lk ; kj = kk ;
			for (j = k+1 ; j < n ; j++)
			{
				lj += lda ; kj += lda ; t= a[lj] ;
				if (l != k) { a[lj] = a[kj] ; a[kj] = t ; }
				cum1vec (&a[kj+1], t, &a[kk+1], nm1-k) ;
			}
			}
/*			else info = k+1 ; */
			else det = 0.;
		}
    }
    ipvt[nm1] = nm1 ;
    kk = nm1*lda1 ;
/*    if ( a[kk] == 0.) info = n ; */
	det *= a[kk];
    return det ;
}
/*-------------------------------------------------------------------------*/
static void lusolv (double *a, int lda, int n, int *ipvt, double *b, int trp)
{
/*
      Renvoi le determinant. Si 0 les calculs sont inacheves
      Adapte en C par MLL le 28/04/92 d'aprs :
      LINPACK. THIS VERSION DATED 08/14/78 .
      CLEVE MOLER, UNIVERSITY OF NEW MEXICO, ARGONNE NATIONAL LAB.
*/
double t, diag ;
int k, kp1, klda, l, nm1 ;

nm1 = n - 1 ;
if (trp == 0 )
{
/*
         trp = 0 , resout  A * X = B
         resout d'abord  L*Y = B
*/
	if (nm1 > 0)
    {
        for (k = 0 , klda = 0 ; k < nm1 ; k++, klda += lda )
        {
            kp1 = k + 1 ;
            l = ipvt[k] ; t = b[l] ;
			if (l != k) { b[l] = b[k] ; b[k] = t ;}
			cum1vec (&b[kp1], t, &a[klda+kp1], n-k-1) ;
        }
    }
/*
         resout  U*X = Y
*/
    for (k = nm1 , klda = nm1 * lda ; k >= 0 ; k--, klda -= lda)
	{
	    if ( (diag = a[k+klda]) == 0.) return ;
        b[k] /= diag ;
	    t = - b[k] ;
	    cum1vec (b, t , &a[klda], k) ;
    }
}
else
{
/*
     trp = 1, resout  A' * X = B
     resout d'abord   U' * Y = B
*/
    for (k = 0, klda = 0 ; k < n ; k++, klda += lda)
    {
		t = dotvec (&a[klda], b, k) ;
        if ( (diag = a[k+klda]) == 0.) return ;
        b[k] = ( b[k] - t ) / diag ;
    }
/*
    resout L' * X = Y
*/
	if (nm1 > 0)
	{
	    for (k = n-2, klda = k * lda ; k >= 0 ; k-- , klda -= lda)
	    {
			kp1 = k + 1 ;
			b[k] += dotvec (&a[klda+kp1], &b[kp1], nm1 - k) ;
			l = ipvt[k] ;
			if (l != k) {t = b[l]; b[l] = b[k] ; b[k] = t ;}
	    }
	}
}
}
/*-------------------------------------------------------------------------*/
/*                              qrfactor                                   */
/*-------------------------------------------------------------------------*/
/*
   Factorisation QR d'une matrice A de dimension N x P avec permutation
   de colonnes en option.
   En entree :
		*X		: Tableau de dimension >= a celle de A. Il contient les
				  elements de A, ranges colonne par colonne.
		ldx		: Nombre de lignes du tableau contenant la matrice A
		n		: Nombre de lignes de A
		*p		: Nombre de colonnes de A
		*jpvt	: Indicateur precisant la nature de chacune des p colonnes
					jpvt[k] == 0 : la colonne peut tre permutee
					jpvt[k]  > 0 : la colonne ne peut tre permutee.
					Si job != 0 sera placee parmi les premires puis figee.
					jpvt[k]  < 0 : la colonne ne peut tre permutee.
					Si job != 0 sera placee parmi les dernires puis figee
					INUTILISE si job = 0.
		job		: Si = 0 : Pas de permutation de colonne
				  Sinon les colonnes libres sont permutees pour tre
				  ordonnees en fonction de leur norme sous-diagonale
		*work	: Tableau de travail de dimension p qui n'est pas
				  utilise si job == 0.
	En sortie :
		*X		: Contient dans son triangle superieur le triangle
		          superieur de R et
				  sous sa diagonale une partie des infos pour reconstituer
				  Q.

		*qraux    de dimension p, contient les infos supplementaires
		          pour reconstituer Q
		*jpvt     jpvt(k) est l'indice dans la matrice originale de la
		          colonne placee en k-ime position dans la solution
				  (INUTILISE si job = 0.)

		rang      valeur thorique du rang avec epsilon = 0.

    Adapte en C par MLL le 29/04/92 a partir de DQRDC de LINPACK :
C     LINPACK. THIS VERSION DATED 08/14/78 .
C     G.W. STEWART, UNIVERSITY OF MARYLAND, ARGONNE NATIONAL LAB.
*/
/*-------------------------------------------------------------------------*/
static int qrfactor(double *x, int ldx, int n, int p, double *qraux, int *jpvt,
						      double *work, int job)
{
int swapj, rang = 0;
int j, jldx, jp, l, ll, lj, lup, maxj, pl, pu , lldx ;
double maxnrm, tt, nrmxl, t ;

    pl = 0 ;
    pu = -1 ;
    if (job != 0)
    { /* Permutations des colonnes selon l'indicateur jpvt */
		for (j = 0 ; j < p ; j++)
		{
			swapj = jpvt[j] > 0 ;
			if (jpvt[j] >= 0) jpvt[j] = j+1 ;
			else jpvt[j] = - (j+1) ;
			if (swapj)
			{
				if (j != pl) swapvec (&x[pl*ldx], &x[j*ldx], n) ;
				jpvt[j] = jpvt[pl] ;
				jpvt[pl] = j+1 ;
				pl++ ;
			}
		}
        pu = p - 1 ;
        for (j = pu ; j >= 0 ; j--)
        {
            if (jpvt[j] < 0)
            {
				jpvt[j] = - jpvt[j] ;
                if (j != pu)
				{
					swapvec (&x[pu*ldx], &x[j*ldx], n) ;
                    jp = jpvt[pu] ;
                    jpvt[pu] = jpvt[j] ;
                    jpvt[j] = jp ;
                }
				pu-- ;
            }
        }
    }
/* Norme des colonnes permutables */
    for (j = pl, jldx = pl * ldx ; j <= pu ; j++, jldx += ldx)
		work[j] = qraux[j] = longvec(&x[jldx], n) ;
/* Triangularisation par transformation de Householder */
    lup = n < p ? n : p ;
    for (l = 0, lldx = 0 ; l < lup ; l++, lldx += ldx)
    {
		ll = l + lldx ;
/*		nrmxl = qraux[l] ;
		supprime car recalcule pour ameliorer la precision */
        if (l >= pl && l < pu)
        {
/* Slection colonne de norme sous-diag. maximale pour pivoter */
            maxnrm = 0. ;
			maxj = l ;
            for (j = l ; j <= pu ; j++)
                if (qraux[j] > maxnrm) { maxnrm = qraux[j] ; maxj = j ; }
            if (maxj != l)
            {
				swapvec (&x[lldx], &x[maxj*ldx], n) ;
/*				nrmxl = qraux[maxj];
				supprime car recalcule pour ameliorer la precision */
                qraux[maxj] = qraux[l] ;
                work[maxj] = work[l] ;
                jp = jpvt[maxj] ;
				jpvt[maxj] = jpvt[l] ;
                jpvt[l] = jp ;
			}
		}
        if (l == n-1) {qraux[l] = 0. ; if (x[ll] != 0) rang++;}
        else
        {
/* Transfo. Househoder de la colonne l */
			/* nrmxl recalcule pour ameliorer la precision */
			nrmxl = longvec (&x[ll], n-l) ;
			if (nrmxl != 0.)
            {
				rang++;
                if (x[ll] < 0.) nrmxl = - nrmxl ;
				sca1vec (&x[ll], 1./nrmxl, n-l) ;
                x[ll] += 1. ;
/* Transfo. des autres colonnes, et mise a jour des normes */
				for (j = l+1, jldx = j*ldx ; j < p ; j++, jldx += ldx)
                {
					lj = l+jldx ;
					shholder (&x[lj], &x[ll], n-l ) ;
					if (j >= pl && j <= pu && qraux[j] != 0.)
					{
						tt = fabs (x[lj]) /qraux[j] ;
						tt = 1. - tt*tt ;
						if (tt < 0.) tt = 0. ;
						t = tt ;
						tt = qraux[j] / work[j] ;
						tt = 1. + 0.5 * t * tt*tt ;
						if (tt != 1.) qraux[j] *= sqrt(t) ;
						else work[j] = qraux[j] = longvec (&x[lj+1], n-l-1) ;
					}
				}
/* Memorisation des transfo. */
				qraux[l] = x[ll] ;
				x[ll] = -nrmxl ;
			}
		}
    }
	return rang;

}
/*-------------------------------------------------------------------------*/
/*                  qrsolv  (SQRSL et DQRSL de LINPACK)                    *
/*-------------------------------------------------------------------------*/
/*
	qrsolv utilise les sorties de qrfactor, prealablement appele
	pour calculer les matrices de permutation de colonne E et de
	projection Q, et la solution des moindres carrs.
	Posons k = min(n,p) et C = A(:,jpvt(1:k))  (syntaxe Matlab)
	Cas redondant     : C est une sous-matrice carre extraite de A
	Cas sur-dtermin : C  la matrice A
	qrfactor fournit la dcomposition QR de C telle que C = Q [R]
	                                                          [0]
	avec Q et R codes dans x et qraux.
En entree :
	*x		: sortie de qrfactor
	ldx		: nombre de lignes du tableau contenant la matrice A
	n		: nombre de lignes de A (p colonnes)
	k		: nombre de colonnes de C <= min(n,p)
	*qraux	: sortie de qrfactor
	*y		: vecteur de dimension n a manipuler
	job		: Indicateur binaire
			  (job&16) : Calculer Q*Y
			  (job&15) : Calculer Q'*Y
			  (job&4)  : Calculer b
			  (job&2)  : Calculer rsd
			  (job&1)  : Calculer C*b
En sortie :
    *qy		: Contient Q*Y (dimension n) si calcule
	*qty	: Contient Q'*Y ((dimension n) si calcule
	*b		: Solution de min ||y-C*b||**2 (dimension k)
	          (b est permute par jpvt)
	*rsd	: (dimension n) erreur rsiduelle y-C*b
	          (projection orthogonale de y sur le noyau de C').
	*xb     : (dimension n) Produit C*b qui est la projection
	          orthogonale de y sur l'image de C.
	qrsolv  : information significatif si calcul de b demande
	          0  Rsolution Ok
			  k>0 R est singulire. k est l'indice du 1er element
			  diagonal nul de R. b n'est pas modifie.
Remarque : Les tableaux qy, qty, b, rsd et xb ne sont pas accedes
           si leur calcul n'est pas requis. On peut mettre n'importe
		   quel argument de type compatible. Par ailleurs on peut
		   fournir le mme tableau en argument a certains groupes
		   de vecteurs accedes. Les partages de memoire suivants sont
		   autorises :

		   1 : tab pour (y,qty,b)
		   2 : tab pour (y,qty,rsd)
		   3 : tab pour (y,qty,xb)

		   4 : tab1 pour (y,qy) et tab2 pour (qty,b)
		   5 : tab1 pour (y,qy) et tab2 pour (qty,rsd)
		   6 : tab1 pour (y,qy) et tab2 pour (qty,xb)

		   tab ou tab1 et tab2 contiennent en sortie le dernier
		   element de la la liste.

	 Adapte en C par MLL le 29/04/92 a partir de DQRSL de LINPACK :
C     LINPACK. THIS VERSION DATED 08/14/78 .
C     G.W. STEWART, UNIVERSITY OF MARYLAND, ARGONNE NATIONAL LAB.
*/
/*-------------------------------------------------------------------------*/
static int qrsolv (double *x, int ldx, int n, int k, double *qraux, double *y,
		double *qy, double *qty, double *b, double *rsd, double *xb,
		int job)
{
int info, i, j, jj, nc, jldx ;
int ldx1 = ldx+1 ;
int cb, cqy, cqty, cr, cxb ;
double temp ;

    info = 0 ;
/* Indicateurs du travail a faire */
    cqy  = job & 16 ;
    cqty = job & 15 ;
    cb   = job & 4 ;
    cr   = job & 2 ;
    cxb  = job & 1 ;

    nc = k < n-1 ? k : n-1 ;
/* Cas particulier n=1 */
    if (nc == 0)
    {
        if (cqy) *qy = *y ;
        if (cqty) *qty = *y ;
		if (cxb) *xb = *y ;
        if (cb)
        {
            if (*x == 0.) info = 1 ;
            else *b = *y / *x ;
        }
		if (cr) *rsd = 0. ;
        return info ;
    }

/* Calcul de Q*y */
    if (cqy)
    {
		for (i = 0 ; i < n; i++)  qy[i] = y[i] ;
        for (j = nc-1, jj = j * ldx1 ; j >= 0 ; j--, jj -= ldx1)
		{
            if (qraux[j] != 0.)
            {
                temp = x[jj] ;
                x[jj] = qraux[j] ;
				shholder (&qy[j], &x[jj], n-j ) ;
				x[jj] = temp ;
            }
		}
    }

/* Calcul de Q'*y */
    if (cqty)
    {
		for (i = 0 ; i < n; i++) qty[i] = y[i] ;
        for (j = 0, jj = 0 ; j < nc ; j++, jj += ldx1)
        {
            if (qraux[j] != 0.)
            {
				temp = x[jj] ;
                x[jj] = qraux[j] ;
				shholder (&qty[j], &x[jj], n-j) ;
                x[jj] = temp ;
            }
        }
    }
/* Initialisations pour le calcul de b, rsd, C=xb */
    if (cxb)
	{
		for (i = 0 ; i < k ; i++) xb[i] = qty[i] ;
		for (i = k ; i < n ; i++) xb[i] = 0. ;
	}
    if (cr)
	{
		for (i = k ; i < n ; i++) rsd[i] = qty[i] ;
		for (i = 0 ; i < k ; i++) rsd[i] = 0. ;
	}
    if (cb) /* Calcul de b */
    {
		for (i = 0 ; i < k ; i++)   b[i] = qty[i] ;
		for (j = k-1, jldx = j*ldx ; j >= 0 ; j--, jldx -= ldx)
        {
            jj = j + jldx ;
            if (x[jj] == 0.) { info = j+1 ; break ; }
            b[j] = b[j] / x[jj] ;
			if (j != 0) cum1vec (b, -b[j] , &x[jldx], j) ;
		}
    }

    if (cr || cxb) /* Calcul de rsd et/ou xb */
    {
        for (j = nc - 1, jj = j * ldx1 ; j >= 0 ; j--, jj -= ldx1)
        {
			if (qraux[j] != 0.)
            {
                temp = x[jj] ;
                x[jj] = qraux[j] ;
				if (cr) shholder (&rsd[j], &x[jj], n-j) ;
				if (cxb) shholder (&xb[j], &x[jj], n-j) ;
                x[jj] = temp ;
            }
		}
    }
    return info ;
}
/*-------------------------------------------------------------------------*/
int solvqr (double *by, double *x, int n, int p, int ns, double eps)
/*---------------------------------------------------------------------------

    !!!!!!!!!!    ATTENTION la matrice x est detruite       !!!!!!!!!!
    Calcule la solution moindre-carres du systeme Y = X * B
    qui peut etre sur ou sous determine.
    L'usager doit fournir une epsilon a partir duquel les colonnes de X
    ne sont plus utilisees pour calculer la solution. Les composantes de
    B correspondantes sont mises a zero.

    Entree :

    X(n,p)     matrice n lignes, p colonnes
    by(n,ns)   ns vecteurs a decomposer sur les colonnes de X
    eps        seuil relatif de singularit

    Sortie :

    by(p, ns) composantes des ns vecteurs solutions
    Renvoi le rang de X
           -1 si echec allocation.
---------------------------------------------------------------------------*/
{
int i, ii, j, k, m, np1 = n+1, nsp = ns*p, nsn=ns*n, rank ;
double *rsd=NULL, *y;
double temp ;

int *jpvt;
double *qraux, *work;

	qraux = malloc((2*p+nsn)*sizeof(double)+p*sizeof(int));
	if (qraux == NULL)
	{perreur("Echec allocation dans solvqr"); return -1;}
	work = qraux + p;
	y = work + p;
	jpvt = (int *) (y+nsn);

	for (i=0;i<p;i++) jpvt[i]=0;/* Toutes colonnes libres */

	if (p <= n) y = by;
	else for(i=0;i<nsn;i++) y[i]=by[i];

    qrfactor (x, n, n, p, qraux, jpvt, work, 1) ;
    rank = 0 ;
    m = n < p ? n : p ;
    eps *= fabs (*x) ;
    for (i = 0, ii = 0 ; i < m ; i++, ii += np1)
    {
		if ( fabs (x[ii]) <= eps) break ;
		rank = i+1 ;
    }

/* Resolution des ns problemes tronques */
    if (rank != 0) for (j = 0 ; j < ns ; j++)
		qrsolv (x, n, n, rank, qraux, &y[j*n], rsd, work, &by[j*p], rsd, rsd, 4) ;
/* Initialise jpvt et met a 0 les lignes k  p de b */
    for (i = 0 ; i < p ; i++)
    {
		jpvt[i] = -jpvt[i] ;
        if (i >= rank) for (j = 0 ; j < nsp ; j += p) by[i+j] = 0. ;
    }

/* Reordonne la solution */
    for (i = 0 ; i < p ; i++)
    {
        if (jpvt[i] > 0) continue ;
		jpvt[i] = - jpvt[i] ;
		k = jpvt[i] - 1 ;
		while (k != i)
        {
            for (j = 0 ; j < nsp ; j += p)
            {
                temp = by[i+j] ;
                by[i+j] = by[k+j] ;
                by[k+j] = temp ;
            }
			jpvt[k] = - jpvt[k] ;
			k = jpvt[k] - 1 ;
		}
    }
	free(qraux);
    return rank ;
}

/*-------------------------------------------------------------------------*/
int rankqr (double *x, int n, int p, double eps)
/*---------------------------------------------------------------------------

    !!!!!!!!!!    ATTENTION la matrice x est detruite       !!!!!!!!!!
    Calcule le rang de la matrice x
    L'usager doit fournir une epsilon relatif a partir duquel on considere
	que les colonnes sont liees.

Entree :
    X(n,p)     matrice n lignes, p colonnes
    eps        seuil relatif de singularit
Sortie :
    Renvoi le rang r de X
	       -1 si echec allocation
---------------------------------------------------------------------------*/
{
int i, ii, m, np1=n+1, rank;
int *jpvt;
double *qraux, *work;

	qraux = malloc((2*p)*sizeof(double)+p*sizeof(int));
	if (qraux == NULL)
	{perreur("Echec allocation dans rankqr"); return -1;}
	work = qraux + p;
	jpvt = (int *) work + p;
	for (i=0;i<p;i++) jpvt[i]=0;/* Toutes colonnes libres */

    qrfactor (x, n, n, p, qraux, jpvt, work, 1) ;
    rank = 0 ;
    m = n < p ? n : p ;
    eps *= fabs (*x) ;
    for (i = 0, ii = 0 ; i < m ; i++, ii += np1)
    {
		if ( fabs (x[ii]) <= eps) break ;
		rank = i+1 ;
    }
	free(qraux);
    return rank ;
}


/*-------------------------------------------------------------------------*/
int nullqr (double *noy, double *x, int n, int p, double eps)
/*---------------------------------------------------------------------------

    !!!!!!!!!!    ATTENTION la matrice x est detruite       !!!!!!!!!!
    Calcule le noyau de la matrice x
    L'usager doit fournir une epsilon relatif a partir duquel on considere
	que les colonnes sont liees.

Entree :
    X(n,p)     matrice n lignes, p colonnes
    eps        seuil relatif de singularit
Sortie :
    noy(n, n-r) Noyau orthonorme  gauche de X tel que N'*X = 0
    Renvoi le rang r de X
	       -1 si echec allocation
---------------------------------------------------------------------------*/
{
int i, ii, j, jn, m, np1=n+1, rank;
int *jpvt;
double *qraux, *work;

	qraux = malloc((2*p)*sizeof(double)+p*sizeof(int));
	if (qraux == NULL)
	{perreur("Echec allocation dans nullqr"); return -1;}
	work = qraux + p;
	jpvt = (int *) work + p;
	for (i=0;i<p;i++) jpvt[i]=0;/* Toutes colonnes libres */

    qrfactor (x, n, n, p, qraux, jpvt, work, 1) ;
    rank = 0 ;
    m = n < p ? n : p ;
    eps *= fabs (*x) ;
    for (i = 0, ii = 0 ; i < m ; i++, ii += np1)
    {
		if ( fabs (x[ii]) <= eps) break ;
		rank = i+1 ;
    }

	for (j = rank, jn=0 ; j < n ; j++, jn+=n)
	{
		for(i=0;i<n;i++) noy[jn+i]=0; noy[jn+j]=1.;
		qrsolv (x, n, n, rank, qraux, &noy[jn], &noy[jn], work, work, work, work, 16) ;
	}
	free(qraux);
    return rank ;
}

/*-------------------------------------------------------------------------*/
int mat2QR (double *Q, double *AR, int n, int p)
/*---------------------------------------------------------------------------

    Calcule la decomposition QR de la matrice A
	            A = QR

Entree :
    XR(n,p)    matrice A n lignes, p colonnes
Sortie :
    Q(n, n)  matrice orthogonale Q
	XR(n, p)  matrice R n lignes, p colonnes
	Renvoi 0 normalement
		-1 si echec allocation
---------------------------------------------------------------------------*/
{
int i, j, jn, m, rank;
double *qraux, bidon[1];

	qraux = malloc(p*sizeof(double));
	if (qraux == NULL)
	{perreur("Echec allocation dans mat2QR"); return -1;}

    rank = qrfactor (AR, n, n, p, qraux, &i, bidon, 0) ;
	m = n < p ? n : p;
	for (j = 0, jn=0 ; j < n ; j++, jn+=n)
	{
		for(i=0;i<n;i++) Q[jn+i]=0; Q[jn+j]=1.;
		qrsolv (AR, n, n, m, qraux, &Q[jn], &Q[jn], bidon, bidon, bidon, bidon, 16) ;
	}
	for (j=jn=0;j<p;j++,jn+=n)
		for(i=j+1;i<n;i++) AR[jn+i]=0.;

	free(qraux);
	return 0;
}

/*-------------------------------------------------------------------------*/
int mat2QRE (double *Q, double *AR, int *jpvt, int n, int p, double eps)
/*---------------------------------------------------------------------------

    Calcule la decomposition QRE de la matrice A
					A = QRE'

Entree :
    XR(n,p)    matrice A n lignes, p colonnes
Sortie :
    Q(n, n)  matrice orthogonale Q
	XR(n, p)  matrice R n lignes, p colonnes
	jpvt[p]   permutation des colonnes QR qui permet de reconstituer A
	Renvoi le rang ou -1 si echec allocation
	Pour retrouver A, il faut faire Q*R, puis appeler
	permcol(QR, jpvt, nlin, ncol);
---------------------------------------------------------------------------*/
{
int i, ii, j, jn, m, rank, np1=n+1;
double *qraux, *work;

	qraux = malloc(2*p*sizeof(double));
	if (qraux == NULL)
	{perreur("Echec allocation dans mat2QRE"); return -1;}
	work = qraux + p;

	for (i=0;i<p;i++) jpvt[i]=0;/* Toutes colonnes sont libres */

    rank = qrfactor (AR, n, n, p, qraux, jpvt, work, 1) ;
    rank = 0 ;
    m = n < p ? n : p ;
    eps *= fabs (*AR) ;
    for (i = 0, ii = 0 ; i < m ; i++, ii += np1)
    {
		if ( fabs (AR[ii]) <= eps) break ;
		rank = i+1 ;
    }

	for (j = 0, jn=0 ; j < n ; j++, jn+=n)
	{
		for(i=0;i<n;i++) Q[jn+i]=0; Q[jn+j]=1.;
		qrsolv (AR, n, n, rank, qraux, &Q[jn], &Q[jn], work, work, work, work, 16) ;
	}
	for (j=jn=0;j<p;j++,jn+=n)
		for(i=j+1;i<n;i++) AR[jn+i]=0.;

	free(qraux);
	return rank;
}


/*-------------------------------------------------------------------------*/
void permcol(double *A, int *jpvt, int nlin, int ncol)
/*-------------------------------------------------------------------------*/
/*  Permutation des colonnes d'une matrice A(nlin,ncol) selon les          */
/*  indications du tableau de permutation jpvt (non modifie)               */
/*-------------------------------------------------------------------------*/
{
int i, k;

	for (i=0;i<ncol;i++) jpvt[i] = -jpvt[i] ;
	for (i=0;i<ncol;i++)
    {
        if (jpvt[i] > 0) continue ;
		jpvt[i] = - jpvt[i] ;
		k = jpvt[i] - 1 ;
		while (k != i)
        {
			swapvec(A+i*nlin,A+k*nlin,nlin);
			jpvt[k] = - jpvt[k] ;
			k = jpvt[k] - 1 ;
		}
    }
}
/*-------------------------------------------------------------------------*/
void permlin(double *A, int *jpvt, int nlin, int ncol)
/*-------------------------------------------------------------------------*/
/*  Permutation des lignes d'une matrice A(nlin,ncol) selon les            */
/*  indications du tableau de permutation jpvt (non modifie)               */
/*-------------------------------------------------------------------------*/
{
int i, j0, k, ntot = nlin*ncol;
double temp;

	for (i=0;i<ncol;i++) jpvt[i] = -jpvt[i] ;
	for (i=0;i<ncol;i++)
    {
        if (jpvt[i] > 0) continue ;
		jpvt[i] = - jpvt[i] ;
		k = jpvt[i] - 1 ;
		while (k != i)
        {
			for (j0 = 0; j0 < ntot; j0 += nlin)
			{
				temp    = A[i+j0];
				A[i+j0] = A[k+j0];
				A[k+j0] = temp;
			}
			jpvt[k] = - jpvt[k] ;
			k = jpvt[k] - 1 ;
		}
    }
}

/*-------------------------------------------------------------------------*/
/*          DECOMPOSITION EN VALEURS SINGULIERES                           */
/*-------------------------------------------------------------------------*/
static int svdc (double *x, int n, int p, double *s, double *e, double *u,
		double *v, double *work, int wantu, int wantv)
{
/*
      Dimensions : x(n,p), s(min(n+1,p)),
      svdc reduit la matrice reelle x(n,p) par les tranf. orthogonales
      u(n,n) et v(p,p) a une forme diagonale dont les elements seront dans
      s(n).
      Les colonnes de u sont les vecteurs singuliers a gauche.
      Les colonnes de v sont les vecteurs singuliers a droite.
   En entree :
      x (n,p)	matrice dont les valeurs singulieres sont a calculer .
		x est detruit par le calcul.
      work(n)   buffer de travail
      wantu :
	   == 0 Pas de calcul des u
	   == 1 calcule n vecteurs singuliers dans u
	   >= 2 calcule les min(n,p) vecteurs singuliers dans u
      wantv :
	   == 0 Pas de calcul de v
       == 1 Calcul des v
   En sortie
       s(min (n+1,p))
		les min(n,p) premiers elements sont les valeurs singulieres
		rangees en taille decroissante
       e(p)     Normallement des 0
       u(n,n)   contient les vecteurs singuliers calcules selon wantu
       v(p,p)   contient les vecteurs singuliers calules selon wantv
		Remarque :
		    si n <= p ou si wantu = 2 u peut partager la meme place
		    que x.
		    si p <= n v peut partager la meme place que x.
       svdc :   si < 0 : Nombre de valeurs singulieres manquant a calculer.
		si >= 0 les valeurs singulieres et les vecteurs singuliers
		   sont corrects. La matrice B = uT * x * v est la matrice
		   bidiagonale avec les elements de de s sur sa diagonale
		   et ceux de e sur sa super-diagonale. Les valeurs singulieres
		   de b et x sont les memes. svdc contient alors le nombre
		   d'iterations effectuees.

   Adapte en C par MLL le 29/04/92 a partir de DSVDC de LINPACK :
C     LINPACK. THIS VERSION DATED 03/13/79 .
C     G.W. STEWART, UNIVERSITY OF MARYLAND, ARGONNE NATIONAL LAB.
*/
int iter, itermax, kase ;
int np1, nm1, pp1, pm1, ncu, nct, nrt, m, mm, mm1, mm2 ;
int lu, l, lp1, ll, j, lj, lp1j, i, il, ij, jj, ls, k ;
double test, ztest, f, g, cs, sn, scale, t1 ;
double sm, sl, b, c;

#define SIGNESHIFT
#define MYSHIFT

#ifdef SIGNESHIFT
double shift ;
#ifdef MYSHIFT
double A,a,E;
#endif /* #ifdef MYSHIFT */
#endif /* #ifdef SIGNESHIFT */
    np1 = n+1 ;
    nm1 = n-1 ;
    pp1 = p+1 ;
    pm1 = p-1 ;
    ncu = n ;
    if (wantu > 1) ncu = n < p ? n : p ;

/*  Reduction de X en bidiagonale. Diagonale dans S et surdiag. dans E */
    nct = nm1 < p ? nm1 : p ;
    nrt = p-2 < n ? p-2 : n ; if (nrt < 0) nrt = 0 ;
    lu = nct > nrt ? nct : nrt ;
    for (l = 0 , ll = 0 ; l < lu ; l++, ll += np1)
    {
		lp1 = l+1 ;
		if (l < nct)
		{
/*      Transformation de la colonne l et place l'element diagonal dans s */
			s[l] = longvec (&x[ll], n-l) ;               /* longueur colonne */
			if (s[l] != 0.)
			{              /* Unitarise la colonne avec 1er element positif */
				if (x[ll] < 0.) s[l] = -s[l] ;
				sca1vec (&x[ll], 1./s[l], n-l) ;
				x[ll] += 1. ;
			}
			s[l] = -s[l] ;
		}
		for (j = lp1, lj = ll + n ; j < p ; j++, lj += n)
		{
/*          Applique la transformation */
			if (l < nct && s[l] != 0.) shholder (&x[lj], &x[ll], n-l) ;
/*          Place la ligne l dans E pour les futures transf. des lignes */
			e[j] = x[lj] ;
		}
/*      Place la transformation dans U pour multiplication futures */
		if (wantu && l < nct)
			for (il = ll ; il < ll+n-l ; il++) u[il] = x[il] ;
/*       Calcule la transf. de la ligne l et place la surdiag dans e[l] */
		if (l < nrt)
		{
			e[l] = longvec (&e[lp1], pm1-l) ;
			if (e[l] != 0.)
			{
				if (e[lp1] < 0.) e[l] = - e[l] ;
				sca1vec (&e[lp1], 1./e[l], pm1-l) ;
				e[lp1] += 1. ;
			}
			e[l] = - e[l] ;
/*          Applique la transformation */
			if (lp1 < n && e[l] != 0.)
			{
				for (i = lp1 ; i < n ; i++) work[i] = 0. ;
				for (j = lp1, lp1j = ll+np1 ; j < p ; j++,lp1j += n)
					cum1vec (&work[lp1], e[j], &x[lp1j], nm1-l) ;
				for (j = lp1, lp1j = ll+np1 ; j < p ; j++, lp1j += n)
					cum1vec (&x[lp1j], -e[j]/e[lp1], &work[lp1], nm1-l) ;
			}
/*          Place la trasf. dans V pour futures multiplications */
			if (wantv)
				for (i = lp1, il =lp1+l*p ; i < p ; i++, il++) v[il] = e[i] ;
		}
    }
/* Termine la bidiagonalisation d'ordre m */
    m = p < np1 ? p : np1 ;
    if (nct < p) s[nct] = x[nct*np1] ;
    if (n < m) s[m-1] = 0. ;
    if (nrt+1 < m) e[nrt] = x[nrt+(m-1)*n] ;
/*  
    La diagonale s comporte m lments numrots de 0  m-1
    La sup-diago e comporte m-1 lments numrots de 0  m-2
*/
    e[m-1] = 0. ; /* Est en dehors de la sup-diago */
/* Calcule U si demande */
    if (wantu)
    {
		for (j=nct, ij=nct*n, jj=ij+j  ; j < ncu ; j++, jj += np1)
		{
			for (i = 0 ; i < n ; i++) u[ij++] = 0. ;
			u[jj] = 1. ;
		}
		for (l = nct - 1, ll = l*np1 ; l >= 0 ; l--, ll -= np1)
		{
			if (s[l] != 0.)
			{
				lp1 = l+1 ;
				for (j = lp1, lj = ll+n ; j < ncu ; j++, lj += n)
					shholder (&u[lj], &u[ll], n-l) ;
				sca1vec (&u[ll], -1., n-l) ;
				u[ll] += 1. ;
				for (i = 0, il = l*n ; i < l ; i++) u[il++] = 0. ;
			}
			else
			{
				for (i = 0, il = l*n ; i < n ; i++) u[il++] = 0. ;
				u[ll] = 1. ;
			}
		}
    }
/* Calcule V si demande */
    if (wantv)
    {
		for ( l = p-1, ll = l*pp1; l >= 0 ; l--, ll -= pp1)
		{
			lp1 = l + 1 ;
			if (l < nrt && e[l] != 0.)
			{
				for (j = lp1, lp1j = ll+pp1 ; j < p ; j++, lp1j += p)
					shholder (&v[lp1j], &v[ll+1], pm1-l) ;
			}
			for (i = 0, il = l*p ; i < p ; i++) v[il++] = 0. ;
			v[ll] = 1. ;
		}
    }
/* Boucle principale pour valeurs singulieres */
    mm = m ;
    itermax = iter = 0 ;
/*  Tant que toutes les valeurs singulieres n'ont pas ete trouvees */
    while (m != 0)
    {
		mm1 = m-1 ;
		mm2 = m-2 ;
	/* Si trop d'iterations -> flag et sortie */
		if (iter >= MAXIT)
		{
#ifdef ACCES_NBITERMAX
			nbitermaxsvd = itermax ;
#endif /* #ifdef ACCES_NBITERMAX */ 
			return m ;
		}
	/* Recherche des element negligeables dans s et e. Positionne :
	   kase = 1     si s(m) et e(l-1) negligeables et  l < m
	   kase = 2     si s(l) negligeable et l < m
	   kase = 3     si e(l-1) negligeable et l < m et s(l) .. s(m) non
			negligeable (etape QR)
	   kase = 4     si e(m-1) est negligeable (convergence).
	*/
		for (l = mm2 ; l >= 0 ; )
		{
			test = fabs (s[l]) + fabs(s[l+1]) ;
			ztest = test + fabs(e[l]) ;
			if (ztest == test) { e[l] = 0. ; break ;}
			l-- ;
		}
		if (l != mm2)
		{
			for (ls = mm1 ; ls > l ; )
			{
				test = 0. ;
				if (ls != mm1) test += fabs(e[ls]) ;
				if (ls != l+1) test += fabs(e[ls-1]) ;
				ztest = test + fabs(s[ls]) ;
				if (ztest == test) { s[ls] = 0. ; break ; }
				ls-- ;
			}
			if (ls != l)
			{
				if (ls != mm1) { kase = 2 ; l = ls ;} else kase = 1 ;
			}
			else kase = 3 ;
		}
		else kase = 4 ;
		l++ ;
	/* Effectue le traitement selon kase */
		switch (kase)
		{
/*       Elimine s(m) negligeable */
		case 1 :
			f = e[mm2] ;
			e[mm2] = 0. ;
			for (k = mm2 ; k >= l ; k--)
			{
				srotg (&s[k], &f, &cs, &sn) ;
				if (k != l) { f = -sn * e[k-1] ; e[k-1] *= cs ; }
				if (wantv) srot (&v[k*p], &v[(mm1)*p], cs, sn, p) ;
			}
			break ;
/*      Partage au s(l) negligeable */
		case 2 :
			f = e[l-1] ;
			e[l-1] = 0. ;
			for (k = l ; k < m ; k++)
			{
				srotg (&s[k], &f, &cs, &sn) ;
				f = - sn * e[k] ;
				e[k] *= cs ;
				if (wantu) srot (&u[k*n], &u[(l-1)*n], cs, sn, n) ;
			}
			break ;
/*      Une etape de QR */
		case 3 :
/*          calcule le deplacement */
			scale = fabs(s[mm1]) ;
			t1 = fabs(s[mm2]) ; if (t1 > scale) scale = t1 ;
			t1 = fabs(e[mm2]) ; if (t1 > scale) scale = t1 ;
			t1 = fabs(s[l]) ; if (t1 > scale) scale = t1 ;
			t1 = fabs(e[l]) ; if (t1 > scale) scale = t1 ;
			sm = s[mm1] / scale ;
			sl = s[l] / scale ;
#ifdef SIGNESHIFT
			b = s[mm2] / scale ;
			c = e[mm2] / scale ;
#ifdef MYSHIFT
			g = mm2 < 1 ? 0. : e[mm2-1]/scale; /* VOIR SI CE TEST EST UTILE */
			A = sm*sm + c*c; /* Element c(b,b) */
			a = b*b + g*g ;  /* Element c(b-1,b-1) */
			E = b*c;		 /* Element c(b-1,b) */
			b = (a-A)/2. ;	 /* Distance A - demisomme val prop matrice c */
			shift = sqrt (b*b + E*E) ;
			f = sl*sl - (A + b + (b > 0. ? -shift : shift));
/* Dans la ligne prcdente le shift est gal  sm*sm - (A + b + (b<0 ? shift : -shift)) */
#else /* #ifdef MYSHIFT */
			b = ((b+sm)*(b-sm) + c*c)/2. ;
			c *= sm ; c *= c ;
			shift = 0. ;
			if (b != 0. || c != 0.)
			{
				shift = sqrt (b*b + c) ;
				if (b < 0.) shift = - shift ;
				shift = c / (b + shift) ;
			}
			f = (sl + sm)*(sl - sm) - shift ;
#endif /* #ifdef MYSHIFT */
#else /* #ifdef SIGNESHIFT */
			f = (sl + sm)*(sl - sm) ;
#endif /* #ifdef SIGNESHIFT */
			g = sl * (e[l] / scale) ;
/*          chasse les zeros */
			for (k = l ; k < mm1 ; k++)
			{
				srotg (&f, &g, &cs, &sn) ;
				if (k != l) e[k-1] = f ;
				f = cs * s[k] + sn * e[k] ;
				e[k] = cs * e[k] - sn * s[k] ;
				g = sn * s[k+1] ;
				s[k+1] *= cs ;
				if (wantv) srot (&v[k*p], &v[(k+1)*p], cs, sn, p) ;
				srotg (&f, &g, &cs, &sn) ;
				s[k] = f ;
				f = cs * e[k] + sn * s[k+1] ;
				s[k+1] = - sn * e[k] + cs * s[k+1] ;
				g = sn * e[k+1] ;
				e[k+1] *= cs ;
				if (wantu && k < n-1) srot (&u[k*n], &u[(k+1)*n], cs, sn, n) ;
			}
			e[mm2] = f ;
			iter++ ;
			if (iter > itermax) itermax = iter ;
			break ;
/*       Convergence */
		case 4 :
/*          Rend positive la valeur singuliere */
			if (s[l] < 0.)
			{
				s[l] = - s[l] ;
				if (wantv) sca1vec (&v[l*p], -1., p) ;
			}
/*          Ordonne les valeurs singulieres */
			while (l != mm-1 && s[l] < s[l+1])
			{
				t1 = s[l] ;
				s[l] = s[l+1] ;
				s[l+1] = t1 ;
				if (wantv) swapvec (&v[l*p], &v[(l+1)*p], p) ;
				if (wantu) swapvec (&u[l*n], &u[(l+1)*n], n) ;
				l++ ;
			}
			iter = 0 ;
			m-- ;
			break ;
		}
    }
#ifdef ACCES_NBITERMAX
	nbitermaxsvd = itermax ;
#endif /* #ifdef ACCES_NBITERMAX */
    return itermax ;
}

/*-------------------------------------------------------------------------*/
int solvsvd (double *xy, double *a,int n,int p,int ns, double eps)
/*---------------------------------------------------------------------------
    !!!!!!!!!!    ATTENTION la matrice a est detruite       !!!!!!!!!!
    Calcule la solution moindre-carres de norme minimale du
    systeme y = a * x qui peut etre sur ou sous determine.
Entree :
    a(n,p)    matrice n lignes, p colonnes
    xy(n,ns)  ns vecteurs a decomposer sur les colonnes de a
    eps       seuil relatif de singularit
Sortie :
    xy(p, ns) composantes des ns vecteurs solutions
	Renvoi le rang de la matrice ou
		-1 si echec allocation
		-2 si echec factorisation
---------------------------------------------------------------------------*/
{
double *s, *e, *u, *v, *work, *uy ;
int nbiter ; /* nombre max d'iterations pour converger */
int wantu, wantv, i, k, iq, mm, mnp, rank, np1=n+1;

    mm = np1 ;
    if (p < mm) mm = p ;
    mnp = n < p ? n : p ;
    s = (double *) malloc ((mm+p+n*mnp+p*p+n+mnp*ns)*sizeof(double)) ;
    if (s == NULL)
	{perreur ("Echec allocation dans solvsvd !") ; return -1 ;}
    e = s + mm ;
    u = e + p ;
    v = u + n*mnp ;
    work = v + p*p ;
    uy = work + n ;
    wantu = 2 ;    /* seulement les min(n,p) premiers vecteurs singuliers u */
    wantv = 1 ;                           /* tous les vecteurs singuliers v */

    nbiter = svdc (a, n, p, s, e, u, v, work, wantu, wantv) ;
    if (nbiter < 0) {free(s) ; return -2;}
/*  printf ("Echec decomposition. Il manque %d val. singulieres !",-nbiter) ; */

/* Rang de la matrice */
	eps = s[0]*eps;
	rank=0;
    for ( k = 0 ; k < mnp ; k++)
	{
		if (s[k] > eps) rank++ ;
		a[k*np1]=s[k]; /* Pour pouvoir calculer le det  partir de la diagonale */
	}

	/*  Moindre carres  x = V1 * D1-1 * U11T * y + V2 * w */
    mnp = rank * ns ;
    tmatmat (uy, u, xy, rank, n, ns) ;
    for (i = 0 ; i < rank ; i++)
		for (iq = i ; iq < mnp  ; iq += rank) uy[iq] /= s[i] ;
    matmat (xy, v, uy, p, rank, ns) ;
    free (s) ;
    return rank ;
}

/*-------------------------------------------------------------------------*/
int nullsvd (double *Noy, double *a, int n, int p, double eps, int fleft)
/*---------------------------------------------------------------------------
    !!!!!!!!!!    ATTENTION la matrice x est detruite       !!!!!!!!!!
    Calcule le noyau  droite ou  gauche (transpose) de la matrice x
    L'usager doit fournir une epsilon relatif a partir duquel on considere
	que les colonnes sont liees.

Entree :
    X(n,p)     matrice n lignes, p colonnes
    eps        seuil relatif de singularit
	fleft	   0 : calcul du noyau  droite, sinon noyau a gauche transpose.
Sortie :
    noy(p, p-r) Noyau orthonorme  droite de X si fleft = 0
    noy(n, n-r) Noyau orthonorme  gauche de X si fleft != 0
    Renvoi le rang r de X ou
		-1 si echec allocation
		-2 echec factorisation
---------------------------------------------------------------------------*/
{
double *s, *e, *work, *noy ;
int wantu, wantv, i, k, rank, mm, dim, dim2, nbit, np1=n+1, mnp;
/* #define IMPRIM */

	if(fleft) {wantu=1;wantv=0;dim=n;}
	else {wantv=1;wantu=0;dim=p;}
	dim2=dim*dim;
    mm = np1 ;
    if (p < mm) mm = p ;
	s = (double *) malloc ((mm+p+n+dim2)*sizeof(double)) ;
    if (s == NULL){perreur ("Echec allocation dans nullsvd !") ; return -1 ;}
    e = s + mm ;
    work = e + p ;
    noy = work + n ;
    nbit = svdc (a, n, p, s, e, noy, noy, work, wantu, wantv) ;
    if (nbit < 0) {free(s); return -2;};
/* Rang de la matrice */
	eps = s[0]*eps;
	rank=0;
	mnp = n <p ? n  :p;
    for ( k = 0 ; k < mnp ; k++)
	{
		a[k*np1]=s[k]; /* Pour pouvoir calculer le det  partir de la diagonale */
		if (s[k] > eps) rank++ ;
	}
    for (i=0, k = rank*dim ; k < dim2 ; ) Noy[i++]=noy[k++];
    free (s) ;
    return rank ;
}


/*-------------------------------------------------------------------------*/
int mat2SVD (double *u,double *S,double *v,double *a,int n,int p,double eps)
/*---------------------------------------------------------------------------
    !!!!!!!!!!    ATTENTION la matrice a est detruite       !!!!!!!!!!
    Calcule la solution moindre-carres de norme minimale du
    systeme y = a * x qui peut etre sur ou sous determine.
Entree :
    a(n,p)    matrice n lignes, p colonnes
    eps       seuil relatif de singularit
Sortie :
    u(n, n)
    s(k)	  k = min(n,p)
    v(p, p)
	Renvoi le rang de la matrice ou
		-1 si echec allocation
		-2 si echec factorisation
---------------------------------------------------------------------------*/
{
double *e, *work, *s;
int nbiter ; /* nombre max d'iterations pour converger */
int wantu, wantv, k,mm, rank;

    k = n+1 ;
    if (p < k) k = p ;
	s = (double *) malloc ((k+p+n)*sizeof(double)) ;
    if (s == NULL){perreur ("Echec allocation dans mat2SVD !") ; return -1 ;}
    e = s + k ;
    work = e + p ;

    wantu = 1 ;
    wantv = 1 ;

    nbiter = svdc (a, n, p, s, e, u, v, work, wantu, wantv) ;
    if (nbiter < 0) {free(s); return -2;}
/*  printf ("Echec decomposition. Il manque %d val. singulieres !",-nbiter) ; */

/* Rang de la matrice */
	eps = s[0]*eps;
	rank = 0;
    mm = n<p?n:p ;
    for ( k = 0 ; k < mm ; k++) if ((S[k]=s[k]) > eps) rank++ ;
	free(s);
	return rank;
}

/**************************************************/

/*---------------------------------------------------------------------
	Dterminant d'une matrice
	-------------------------

  !!!!!!!!!  ATTENTION LA MATRICE A EST DETRUITE   !!!!!!!!!!!!!!!!

	Entrees:
		double *A	: Adresse de A
		int    n	: Nombre de lignes de A

    Sortie :
	    A est dtruite
		Retour	: Dterminant de A
 Si Det == 0, il y a peut tre eu echec allocation
----------------------------------------------------------------------*/
double detmat(double *a, int n)
{
int *ipvt ;
double det;

	ipvt = (int *) malloc (n*sizeof(int)) ;
    if (ipvt == NULL){perreur ("Echec allocation dans detmat !") ; return 0. ;}
	det = lufactor (a, n, n, ipvt);
	free(ipvt);
	return det;
}


/*---------------------------------------------------------------------
	Dterminant et Inversion d'une matrice (par dcomposition LU)
	-------------------------------------------------------------

	Entrees:
		double *A	: Adresse de A
		int    n	: Nombre de lignes de A

    Sortie :
		inv(A) si det != 0.
		Retour	: dterminant de A
 Si Det == 0, il y a peut tre eu echec allocation
----------------------------------------------------------------------*/
double invmat(double *a, int n)
{
int *ipvt;
double det, *work;

	work = (double *) malloc (n*(sizeof(double)+sizeof(int))) ;
    if (work == NULL){perreur ("Echec allocation dans invmat !") ; return 0. ;}
	ipvt = (int *) ( work+n);

    if ( (det = lufactor (a, n, n, ipvt)) != 0.)
	{
		double t ;
		int j, jn, k, kk, kp1, kn, jk, kj ;
		int n1 = n + 1 ;

		/* Calcul de l'inverse de U */
		for (k = 0, kn = 0 ; k < n ; k++, kn += n)
		{
			kk = k + kn ;
			a[kk] = 1./a[kk] ; t = - a[kk] ;
			sca1vec(&a[kn], t, k) ;
			kp1 = k + 1 ;
			if (kp1 < n) for (j = kp1, jn = kp1 * n ; j < n ; j++, jn += n)
			{
				kj = k + jn ;
				t = a[kj] ; a[kj] = 0. ;
				cum1vec (&a[jn], t, &a[kn], kp1) ;
			}
		}
		/* Forme inverse de U * inverse de L */
		if ( n >= 2)
		{
			for (k = n-2 , kn = k * n ; k >= 0 ; k--, kn -= n)
			{
				kp1 = k + 1 ;
				for (j = kp1 ; j < n ; j++)
				{ jk = j + kn ; work[j] = a[jk] ; a[jk] = 0. ; }
				for (j = kp1 ; j < n ; j++)	cum1vec (&a[kn], work[j] , &a[j*n], n) ;
				j = ipvt[k] ;
				if (j != k) swapvec (&a[kn], &a[j*n], n) ;
			}
		}
	}
	free(work);
    return det ;
}

/*---------------------------------------------------------------------
	               /  Y = A  * X  (si trp = 0)  -> X = inv(A)  * Y
	Rsolution de  |
	               \  Y = A' * X  (si trp = 1)  -> X = inv(A') * Y
	-------------------------------------------------------------------

    !!!!!!!!!!  ATTENTION : LA MATRICE A EST DETRUITE  !!!!!!!!!!!

	Entrees:
		double *XY		: Adresse de Y vecteur donn et X vecteur rsultat
		double *A		: Adresse de A (qui sera dtruite)
		int    n		: dimension de A (n x n)
		int    ns		: Nombre de colonnes de XY
		int    trp		: Flag de transposition de A

    Sortie :
		X est calcule
		Renvoi le determinant
 Si Det == 0, rsolution impossible (peut tre echec allocation ?)
----------------------------------------------------------------------*/
double solvlu (double *XY, double *A, int n, int ns, int trp)
{
int *ipvt ;
double det ;

	ipvt = (int *) malloc (n*sizeof(int)) ;
    if (ipvt == NULL){perreur ("Echec allocation dans solvlu !") ; return 0. ;}

    if ( (det=lufactor (A, n, n, ipvt)) != 0)
    {
		while (ns--) lusolv (A, n, n, ipvt, &XY[n*ns], trp) ;
    }
	free(ipvt);
    return det ;
}
/*---------------------------------------------------------------------*
  ROUTINE solvchud : RESOLUTION SYSTEME LINEAIRE SYMETRIQUE PAR CHOLESKI
                     SANS RACINE CARREE

                        det = solvchud(a,b,n,q)

                        -1
               [C] = [B]  *  [A]  avec [B] matrice symetrique

     Remarque: Seule la partie superieure de B est utilisee.

     Le resultat [C] est renvoye dans A

     Entree :  n        Ordre du systeme
               q        Nombre de seconds membres
               a(n,q)   Valeurs des seconds membres
               b(n,n)   Matrice symetrique du systeme

     Sortie:   a(n,q)   Solution du systeme.
	 Retour    det    Valeur du dterminant
					   Si 0 resultat incorrect

     Methode : Algorithme de Cholevski lineaire sans racine carree:

          L' equation  A = B . C   avec B symetrique peut s'ecrire :

                       A = R'.D.R. C  avec:

              R matrice triangulaire superieure a diagonale unite
              D matrice diagonale.

          La resolution du probleme
                       A = R'.D.R.C
          est alors ramenee a la resolution de
                       A = R'. E
          puis
                       (E/D) = R . C

          Les elements de la partie superieure de la matrice R sont
		  calcules et ranges dans la partie superieure de B.
          Les elements de D sont calcules simultanement et ranges dans
		  la diagonale de B.
          Les elements de E sont calcules et ranges dans A.
          Les elements de C sont calcules et ranges dans A.
-----------------------------------------------------------------------*/
double solvchud(double* a, double* b, int n, int q)
{
int i, j, k, l;
double diag, term, det;
double *pa, *pb, *p1, *p2, *p3;

	det = 1.;
/*    invs = 0;*/
/*
Calcul des elements Rij et Dii en lieu et place de B
*/
    pb = b;                         /* b(i,i) pour i=1,... */
    for (i = 0; i < n; ++i)
    {
		diag = *pb;                  /* diag = b(i,i) */
        if (i)
        {
			p1 = b;                   /* b(k,k) pour k=1,... */
			p2 = pb -i;               /* b(k,i) pour k=1,... */
            for (k = 0; k < i; ++k)
			{ /* diag = diag - b(k,k)*b(k,i)*b(k,i) */
				diag -= *p1 * *p2 * *p2;
                p1 += n + 1;
                ++p2;
            }
		}
        if (diag == 0.) return 0.;
/*        if (diag < 0.) ++invs; */
        det *= diag;
		*pb = diag;

        p3 = pb + n;                 /* b(i,j) pour j=i+1,... */
        for (j = i+1; j < n; ++j)
        {
			term = *p3;               /* term = b(i,j) */
            if (i)
            {
				p1 = b;                /* b(k,k) pour k=1,... */
                p2 = pb - i;           /* b(k,i) pour k=1,... */
				p3 -= i;               /* b(k,j) pour k=1,... */
                for (k = 0; k < i; ++k)
				{
					/* term = term - b(k,i)*b(k,k)*b(k,j) */
					term -= *p2++ * *p1 * *p3++;
                    p1 += n + 1;
                }
			}
            *p3 = term/diag;          /* b(i,j) = term / diag */
            p3 += n;
		}
        pb += n + 1;                 /* b(i+1,i+1) */
	}
/*
calcul des elements de E en lieu et place de A
*/
	pa = a;                         /* a(1,l) pour l=1,... */
    for (l = 0; l < q; ++l)
    {
		p1 = pa;                     /* a(i,l) pour i=1,... */
		pb = b;                      /* b(1,i) pour i=1,... */
        for (i = 0; i < n; ++i)
		{
			term = *p1;               /* term = a(i,l) */
/*          if(i)   */   /* supprim car p2 utilis non initialis pour n=1 */
            {
				p2 = pb;                /* b(k,i) pour k=1,... */
				p3 = pa;                /* a(k,l) pour k=1,... */
                for (k = 0; k < i; ++k) /* term = term - b(k,i)*a(k,l) */
					term -= *p2++ * *p3++;
            }
            *p1++ = term;             /* a(i,l) = term, puis a(i+1,l)*/
            pb += n;                  /* b(1,i+1) */
		}
/*
calcul des elements de C en lieu et place de A
*/
		p1 = pa + n - 1;             /* a(i,l) pour i=n,n-1,... */
        pb = p2;                     /* b(i,i) pour i=n,n-1,... */
		for (i = n; i > 0; --i)
        {
			term = *p1 / *pb;         /* term = a(i,l)/b(i,i) */
			if(i < n)
            {
				p2 = pb + n;            /* b(i,i+1) */
                p3 = p1 + 1;            /* a(i+1,l) */
				for (k = i+1; k <= n; k++)
                {/* term = term - b(i,k)*a(k,l) */
					term -= *p2 * *p3++;
					p2 += n;             /* b(i,k+1) */
                }
			}
            *p1-- = term;             /* a(i,l) = term, puis a(i-1,l)*/
            pb -= n + 1;              /* b(i-1,i-1) */
		}
        pa += n;                     /* a(1,l+1) */
	}
	return det ;
}

/*---------------------------------------------------------------------*
  ROUTINE solvchsq : RESOLUTION SYSTEME LINEAIRE SYMETRIQUE PAR CHOLESKI
                    AVEC RACINE CARREE

                        det = solvchsq(a,b,n,q)

                        -1
               [C] = [B]  *  [A]  avec [B] matrice symetrique

     Remarque: Seule la partie superieure de B est utilisee.

     Le resultat [C] est renvoye dans A

     Entree :  n        Ordre du systeme
               q        Nombre de seconds membres
               a(n,q)   Valeurs des seconds membres
               b(n,n)   Matrice symetrique du systeme

     Sortie:   a(n,q)   Solution du systeme.
                 det    Valeur du dterminant
					Si 0 resultat incorrect (peut etre echec allocation)
					Si < 0 rsultat incertain y compris
					le signe du determinant.

ATTENTION : Si B negative : resultat aberrant !  -> Utiliser solvchud !!!
-----------------------------------------------------------------------*/
double solvchsq(double *a, double *b, int n, int q)
{
    double *p, *pr, *mp, det;
    register double *pki, *pkj;
    register int k;
    double diag, term, bmul;
    int i, j, nq, n1, *is;
	det = 1.;

	is = (int *) malloc (n*sizeof(int)) ;
    if (is == NULL){perreur ("Echec allocation dans solvchsq !") ; return 0. ;}

/*    index = 0;*/
    nq = n + q;
    n1 = n + 1;
/*
Calcul des elements Rij en lieu et place de B et A
*/
    p = b - n1;
    for (i = 0; i < n; i++)
	{             /* resultat ligne par ligne  */
		p += n1;                          /* element diagonal suivant */
		diag = *p;                              /* Bii */
		if (i != 0)
		{
			p -= i;                             /* Boi */
			for (k = 0; k < i; k++)
			{
				bmul = *p++;                    /* Bki */
				if (is[k]) diag -= bmul * bmul; else diag += bmul * bmul;
            }
        }
        if (diag == 0.) { free(is); return 0.;}
		det *= diag;
        if (diag < 0.)
		{
            is[i] = 0;
/*            index += 1; */
            *p = diag = -sqrt(-diag);         /* Bii */
        }
		else
		{
            is[i] = 1;
            *p = diag = sqrt(diag);           /* Bii */
        }
        pr = p;                                 /* &Bii */
        for (j = i+1; j < nq; j++)
		{
            pr += n;                            /* &Bij */
            if (j == n) pr = a + i;             /* &Ai,j-n */
            term = *pr;                         /* Bi,j ou Ai,j-n */
			if (i != 0)
			{
                pkj = pr;                       /* &Bij ou &Ai,j-n */
				pki = p;                        /* &Bii */
                for (k = i - 1; k >= 0; k--)
				{
                    bmul = *--pki * *--pkj;    /* Bk,i*Bk,j (ou Ak,j-n) */
                    if (is[k]) term -= bmul; else term += bmul;
                }
            }
            *pr = term / diag;
        }
    }
/*
Calcul 2emes membres
*/
    mp = p;                                     /* memo &Bnn */
    pkj = a - 1;                                /* &An,-1 */
    while (q--)
	{                               /* Pour les j=1,q col de A */
        pr = pkj + n;                           /* &An,j */
        p = mp;                                 /* &Bnn */
        for (i = n; i > 0; i--)
		{             /* En remontant la col.(math) */
			term = *pr;                         /* Ai,j */
			if (i != n)
			{
                pki = p;                        /* &Bii */
                pkj = pr;                       /* &Aij */
                for (k = i; k < n; k++)
				{       /* (math) k=i+1 a n */
                    pki += n;                   /* &Bik */
                    term -= *pki * *++pkj;      /* Bik * Akj */
                }
            }
            *pr-- = term / *p;                  /* Aij = term / Bii */
            p -= n1;                            /* element diagonale prec. */
        }
    }
	free(is);
    return det;
}

/*-------------------------------------------------------------------------*/
/*     Exemple de resolution systeme lineaire quelconque par la
/*     presqu'inverse avec CHOLESKI                                        */
/*-------------------------------------------------------------------------*/
double solvpi (double *x, double *a, double *y, int n, int p, int ns, double t)
/*---------------------------------------------------------------------------
    !!!!!!!!!!    ATTENTION la matrice a est detruite       !!!!!!!!!!
					   2         2
    Calcule la solution minimisant ||y-ax||  + t||x||
Entree :
    a(n,p)    matrice n lignes, p colonnes
    y(n,ns)   ns vecteurs a decomposer sur les colonnes de a
Sortie :
    x(p, ns)  composantes des ns vecteurs solutions
    Retour !=0.  dterminant de A'*A
	       0. resultat incorrect (echec alloc eventuellement)
----------------------------------------------------------------------------*/
{
int i, nns = n*ns, pp = p*p, pp1 = p+1 ;
double *tempo, det ;

    tempo = (double *) malloc ((pp+nns) * sizeof (double)) ;
    if (tempo == NULL)
    { perreur ("Echec allocation dans solvpi") ; return 0. ; }
    cpyvec (tempo+pp, y, nns) ;                  /* pour permettre &x = &y */
    tmatmat (x, a, tempo+pp, p, n, ns) ;
    tAAmat (tempo, a, p, n) ;
    for (i = 0 ; i < pp ; i += pp1) tempo[i] += t ;
    det = solvchud (x, tempo, p, ns) ;
    free (tempo) ;
    return det ;
}

/****************************************************************************
 Pseudo-Inverse (SV) de la matrice a(n,p) :
	 0 : OK
	-1 : echec allocation
	-2 : echec resolution

 Attention a est detruite
****************************************************************************/
int pinvmat(double *x, double *a, int n, int p)
{
double *s, *e, *u, *v, *work, *uy ;
int wantu, wantv, i, k, iq, mm, mnp ;

    mm = n+1 ;
    if (p < mm) mm = p ;
    mnp = n < p ? n : p ;
    s = (double *) malloc ((mm+p+n*mnp+p*p+n+mnp*n)*sizeof(double)) ;
    if (s == NULL){perreur ("Echec allocation dans pinvmat !") ; return -1 ;}
    e = s + mm ;
    u = e + p ;
    v = u + n*mnp ;
    work = v + p*p ;
    uy = work + n ;
    wantu = 2 ;  /* seulement les min(n,p) premiers vecteurs singuliers u */
    wantv = 1 ;  /* tous les vecteurs singuliers v */

    wantu = svdc (a, n, p, s, e, u, v, work, wantu, wantv) ;
    if (wantu < 0) {free(s); return -2 ;}
/* Rang de la matrice */
    for ( k = 0 ; k < mm ; k++) if (s[k] == 0.) break ;
/*  Moindre carres  x = V1 * D1-1 * U11T * y + V2 * w */
    mnp = k * n ;
    trmat (uy, u, k, n) ;
    for (i = 0 ; i < k ; i++)
	for (iq = i ; iq < mnp  ; iq += k) uy[iq] /= s[i] ;
    matmat (x, v, uy, p, k, n) ;
    free (s) ;
    return 0 ;
}
