#include "polynomes.h"
#include "deferreur.h"

#include <math.h>
#include <stdlib.h>
#include <float.h> /* Pour DBL_MAX*/

/*--------------------------------------------------------------------------*/
/* Degre a partir duquel les procedures roots_hqr et roots balhqr calculent */
/* les racines par iteration plutot que par les formules analytiques        */
/* Ex: Si kRootsIter = 2, les iterations sont utilisees meme pour le degre 2*/
/*--------------------------------------------------------------------------*/
static int kRootsIter = 4;      /* On n'utilise pas les formules de Ferrari */
#include "roots234.c"

/*--------------------------------------------------------------------------*/
/*       Division euclidienne du polynome a(x) par le polynome b(x).        */
/*--------------------------------------------------------------------------*/
/*       Les coefficients sont ranges en degre croissant.                   */
/*               a(x)   =    b(x)   *    q(x)   -    r(x)                   */
/*    degres :    na          nb      nq=na-nb      nr < nb                 */
/*    q et r et nq sont calcules.                                           */
/*    Renvoi nr degre de r.                                                 */
/*--------------------------------------------------------------------------*/
/*    q et r doivent etre differents de a et b !!!!!!!!!!!!!                */
/*--------------------------------------------------------------------------*/
int divpol (double *a, int na, double *b, int nb, double *q, int *Nq,
								  double *r)
{
int i, nq ;
double Q, B = b[nb] ;

    *Nq = na - nb ;                                           /* C'est fait */
  /* Copie de travail de a(x) dans A(x) */
    for (i = 0 ; i <= na ; i++) r[i] = a[i] ;
  /* Calcul quotient et reste */
    while ((nq = na - nb) >= 0)
    {
		q[nq] = Q = r[na--] / B ;
		for (i = 0 ; i < nb ; i++) r[i+nq] -= Q*b[i] ;
    }
  /* Elimination termes de degre superieurs negligeables */
    while (a[na]+r[na] == a[na]) na-- ;
    if (na < 0) { na = 0 ; r[na] = 0. ; }
  /* Retour reste */
    for (i = 0 ; i <= na ; i++) r[i] = -r[i] ;
    return na ;
}

/*--------------------------------------------------------------------------*/
/*    Valeur du polynome coef (croissants de degre 0 a n) au point x        */
/*--------------------------------------------------------------------------*/
double valpol (double *coef, int n, double x)
{ double r = coef[n] ;
    while (n-- > 0)
	r = coef[n] + r*x ;
	return r ; }

/*--------------------------------------------------------------------------*/
/* Valeur derivee du polynome coef (croissants de degre 0 a n) au point x   */
/*--------------------------------------------------------------------------*/
double valderpol (double *coef, int n, double x)
{ double r = n*coef[n] ; while (n-- > 1) r = n*coef[n] + r*x ; return r ; }

/*--------------------------------------------------------------------------*/
/* Polynome derivee du polynome coef (croissants de degre 0 a n)            */
/* Renvoi le degre np-1                                                     */
/*--------------------------------------------------------------------------*/
int derpol (double *coefd, double *coef, int np)
{int n = np ;  do { coefd[n-1] = n*coef[n] ; } while (--n > 0) ; return np-1;}

/*--------------------------------------------------------------------------*/
/* Utilitaire interne pour impression                                       */
/*--------------------------------------------------------------------------*/
#include <stdio.h>
static void pp (double a, int i)
{
    switch (i) {
    case 0 :
	printf ("%lg\n",a) ;
	break ;
    case 1 :
	if (a == 1.) printf ("X ") ;
	else if (a == -1.) printf ("-X ") ;
	else printf ("%lgX ", a) ;
	break ;
    default :
	if (a == 1.) printf ("X^%d ",i) ;
	else if (a == -1.) printf ("-X^%d ",i) ;
	else printf ("%lgX^%d ",a, i) ;
	break ;
    }
}
/*--------------------------------------------------------------------------*/
/*       Impression du polynome a(x) (coef puissances croissantes)          */
/*--------------------------------------------------------------------------*/
void printpol (double *a, int n)
{
    pp (a[n],n) ;
    while (n-- > 0)
    {
	if (a[n] >= 0.) printf ("+") ;
	pp (a[n], n) ;
    }
}

/*--------------------------------------------------------------------------*/
/* Ajoute au polynome coef (croissant) de degre n la racine a. Renvoi n+1   */
/*--------------------------------------------------------------------------*/
int apend_realroot (double *coef, int n, double a)
{
int i ;
    a = - a ;
    if (n == 0)
    { coef[1] = 1. ; coef[0] = a ; }
    else
    {
	coef[n+1] = coef[n] ;
	for (i = n ; i > 0 ; i--) coef[i] = coef[i-1] + a*coef[i] ;
	coef[0] *= a ;
    }
    return n+1 ;
}

/*--------------------------------------------------------------------------*/
/* Ajoute au pol coef (croissant) de degre n la racine a+ib. Renvoi n+2     */
/*--------------------------------------------------------------------------*/
int apend_cpxroot (double *coef, int n, double a, double b)
{
int i ;
double A0, A1 ;
    a = - a ;
    A1 = 2.*a ; A0 = a*a+b*b ;
    if (n == 0)
    { coef[2] = 1. ; coef[1] = A1 ; coef[0] = A0 ; }
    else
    {
	coef[n+2] = coef[n] ;
	coef[n+1] = A1*coef[n]+coef[n-1] ;
	for (i = n ; i > 1 ; i--)
	    coef[i] = A0*coef[i] + A1*coef[i-1] + coef[i-2] ;
	coef[1] = A0*coef[1] + A1*coef[0] ;
	coef[0] *= A0 ;
    }
    return n+2 ;
}

/*--------------------------------------------------------------------------*/
/* Constitue le polynome caracteristique (croissant) de racines wr+i*wi     */
/*--------------------------------------------------------------------------*/
void rac2pol (double *coefs, double *wr, double *wi, int n)
{
int i, m = 0 ;

    for (i = 0 ; i < n ; i++)
    {
	if (wi[i] == 0.) m = apend_realroot (coefs, m, wr[i]) ;
	else { m = apend_cpxroot (coefs, m, wr[i], wi[i]) ; i++ ; }
    }
}

/*--------------------------------------------------------------------------*/
/*  Calcul d'une suite de sturm avec arret au premier facteur commun        */
/*--------------------------------------------------------------------------*/
/* Le tableau A[(M+1)*(M+1)] doit contenir en A[0:M] le polynome a traiter  */
/* Son degre M doit etre indique dans nd[0].                                */
/* La routine calcule le polynomes A1 polynome derivee de A0, puis          */
/* Ai reste de la division euclidienne Ai-2 = Ai-1*Q - Ai pour              */
/* i = 2, jusqu'a l'iteration i = I ou le reste est une constante.          */
/* Les degres des polynomes Ai sont stockes dans nd[i].                     */
/* Si la constante est non nulle, il n'y a pas de facteur commun et on      */
/* renvoi l'index I du dernier polynome reste de degre nd[I] = 0            */
/* Si la constante est nulle, on renvoi l'index I = I-1 du polynome         */
/* precedent qui est facteur commun a P et P' de degre nd[I]                */
/* Remarque : La valeur max de I est nd[0].                                 */
/* Renvoi -1 si echec allocation                                            */
/*--------------------------------------------------------------------------*/
int sturm1 (double *A, int *nd)
{
int I, nq ;
double *Q ;
int dim = nd[0]+1;

	Q = (double *) malloc(dim*sizeof(double));
	if (Q == NULL) { perreur("Echec allocation dans sturm1"); return -1 ;}

    I = 1 ; nd[I] = derpol(A+I*dim, A+(I-1)*dim, nd[I-1]) ;
    while (nd[I] > 0)
    { 
		I++ ; 
		nd[I] = divpol(A+(I-2)*dim, nd[I-2], A+(I-1)*dim, nd[I-1], Q, &nq, A+I*dim) ; 
	}
    if (A[I*dim] == 0.) I-- ;
	free(Q);
    return I ;
}

/*--------------------------------------------------------------------------*/
/*    Calcul du nombre de changements de signes de la suite de Sturm en X   */
/*--------------------------------------------------------------------------*/
/*  Les coefficients en degre croissant du polynome doivent etre ranges     */
/*  dans A0. Son degre M est specifie dans nd[0].                           */
/*  Si I vaut 0, la routine calcule la suite polynomiale de Sturm qui est   */
/*  rangee dans Ai, degre dans nd[i], pour i = 1 a I qui est calcule.       */
/*  Si I est different de 0, c'est que la suite polynomiale de sturm, et I  */
/*  ont deja ete calcules.                                                  */
/*  Le nombre de changements de signes de la suite de sturm en x, pour les  */
/*  racines simples est "returne".                                          */
/*--------------------------------------------------------------------------*/
int sgn_sturm1 (double *A, int *nd, double x, int *I)
{
int i, s, as, n ;
int dim = nd[0]+1;

   if (*I == 0) *I = sturm1 (A, nd) ;
   as = valpol (A, nd[0], x) >= 0. ? 1 : 0 ;
   n = 0 ;
   for (i = 1 ; i <= *I ; i++)
   {
       s = valpol (A+i*dim, nd[i], x) >= 0. ? 1 : 0 ;
       if (s != as) n++ ;
       as = s ;
   }
   return n ;
}

/*--------------------------------------------------------------------------*/
/*  Calcul d'une suite de sturm jusqu'au reste de degre zero                */
/*--------------------------------------------------------------------------*/
/* Le tableau A[(M+1)*(M+1)] doit contenir en A[0] le polynome a traiter    */
/* Son degre doit etre indique dans nd[0].                                  */
/* La routine calcule le polynome A1 polynome derivee de A0, puis           */
/* Ai reste de la division euclidienne Ai-2 = Ai-1*Q - Ai pour              */
/* i = 2, jusqu'a l'iteration i = I ou le reste est une constante.          */
/* Les degres des polynomes Ai sont stockes dans nd[i].                   */
/* Si la constante est non nulle, il n'y a pas de facteur commun et on      */
/* renvoi l'index I du dernier polynome reste de degre nd[I] = 0            */
/* Si la constante est nulle, on recommence le processus complet a partir   */
/* du polynome precedent qui est facteur commun, et ainsi de suite jusqu'a  */
/* que le reste soit effectivement une constante non nulle ou que le        */
/* degre du dernier facteur commun soit de degre 1.                         */
/* Remarque : La valeur max de I est nd[0].                                 */
/* Renvoi -1 si echec allocation                                            */
/*--------------------------------------------------------------------------*/
int sturmM (int *I0, double *A, int *nd)
{
int I, nq, fI0 = 1 ;
double *Q ;
int dim = nd[0]+1;

	Q = (double *) malloc(dim*sizeof(double));
	if (Q == NULL) { perreur("Echec allocation dans sturmM"); return -1 ;}
    I = 0 ;
    while (nd[I] > 0)
    {
      I++ ; nd[I] = derpol(A+I*dim, A+(I-1)*dim, nd[I-1]) ;
      while (nd[I] > 0)
      { 
		  I++ ; 
		  nd[I] = divpol(A+(I-2)*dim, nd[I-2], A+(I-1)*dim, nd[I-1], Q, &nq, A+I*dim) ;
	  }
      if (A[I*dim] == 0.) I-- ;
      if (fI0) { fI0 = 0 ; *I0 = I ; }
    }
	free(Q);
    return I ;
}

/*--------------------------------------------------------------------------*/
/*    Calcul du nombre de changements de signes de la suite de Sturm en X   */
/*--------------------------------------------------------------------------*/
/*  Les coefficients en degre croissant du polynome doivent etre ranges     */
/*  dans A0. Son degre M est specifie dans nd[0].                           */
/*  Si I vaut 0, la routine calcule la suite polynomiale de Sturm qui est   */
/*  rangee dans Ai, degre dans nd[i], pour i = 1 a I qui est calcule ainsi  */
/*  que I0.                                                                 */
/*  Si I est different de 0, c'est que la suite polynomiale de sturm, I et  */
/*  I0 ont deja ete calcules.                                               */
/*  Le nombre de changements de signes de la suite de sturm en x, (pour les */
/*  racines simples) est "returne". Celui qui comptabilise les racines      */
/*  multiples est mis dans ntot"                                            */
/*--------------------------------------------------------------------------*/
int sgn_sturmM (int *ntot, double *A, int *nd, double x, int *I,
								      int *I0)
{
int i, k, s, as, n ;
int dim = nd[0]+1;

   if (*I == 0) *I = sturmM (I0, A, nd) ;
   as = valpol (A, nd[0], x) >= 0. ? 1 : 0 ;
   n = 0 ;
   for (i = 1 ; i <= *I0 ; i++)
   {
       s = valpol (A+i*dim, nd[i], x) >= 0. ? 1 : 0 ;
       if (s != as) n++ ;
       as = s ;
   }
   k = n ;
   for (i = *I0+1 ; i <= *I ; i++)
   {
       s = valpol (A+i*dim, nd[i], x) >= 0. ? 1 : 0 ;
       if (s != as) n++ ;
       as = s ;
   }
   *ntot = n  ;
   return k ;
}

/*--------------------------------------------------------------------------*/
/*                 Calcul des racines reelles d'un polynome                 */
/*    Renvoi le nombre de racines calculees ou -1 si echec allocation       */
/*--------------------------------------------------------------------------*/
int roots_sturm (double *RacR, double *Coef, int M, double Epsx)
{
double *A ;
int *nd;
int i, k, n, ni, nf, mi, mf, mt, ntot, nt, m, I = 0, I0, flag = 1 ;
double *Xs, *mu, *Dc ;
double Xi, Xf, Xa, Xb, x, dx, Pol, Dpol, pas, mult ;
int dim = M+1;

    A = (double *) malloc ((dim+4)*dim*sizeof(double)) ;
	if (A == NULL) { perreur("Echec allocation dans roots_sturm"); return -1 ;}
	Xs = A + dim*dim;
	mu = Xs + dim;
	Dc = mu + M;
	nd = (int *) (Dc+M);

    nd[0] = M ;
/* Estimation des bornes de la recherche */
    Xf = fabs(Coef[0]) ;
    for (i = 1 ; i <= M ; i++) {x = fabs(Coef[i]) ; if (x > Xf) Xf = x ; }
    Xf = 1. + Xf/x ; Xi = - Xf ;
/* Calcul nombre de racines par les suites de sturm */
    for (i = 0 ; i <= M ; i++) A[i] = Coef[i] ;
    ni = sgn_sturmM (&mi, A, nd, Xi, &I, &I0) ;
    nf = sgn_sturmM (&mf, A, nd, Xf, &I, &I0) ;
    ntot = nt = ni - nf ;

/* Isolation des racines dans des intervalles */
    k = 0 ;
    while (nt != 0)
    {
      dx = Xf - Xi ;
      if (flag)
      {
      /* Intervalle pour la racine la + gde */
	 x = Xi ; n = ni ; mt = mi - mf;
	 while (nt != 1)
	 {
	   dx /= 2. ; x = Xf - dx ;
	   n = sgn_sturmM (&m, A, nd, x, &I, &I0) ;
	   nt = n - nf ; mt = m - mf ;
	   if (nt == 0) { Xf = x ; nf = n ; mf = m ; }
	 }
	 Xs[ntot-k] = Xf ; Xs[ntot-k-1] = x ;
	 mu[ntot-k-1] = mt ;
	 Xf = x ; nf = n ; mf = m ;       /* Pour recherche racine suivante */
      }
      else
      {
      /* Intervalle pour la racine la + petite */
	 x = Xf ; n = nf ; mt = mi - mf ;
	 while (nt != 1)
	 {
	    dx /= 2. ; x = Xi + dx ;
	    n = sgn_sturmM (&m, A, nd, x, &I, &I0) ;
	    nt = ni - n ; mt = mi - m ;
	    if (nt == 0) { Xi = x ; ni = n ; mi = m ; }
	 }
	 mu[k] = mt ;
	 Xs[k] = Xi ; Xs[++k] = x ;
	 Xi = x ; ni = n ; mi = m ;       /* Pour recherche racine suivante */
      }
      flag = 1 - flag ;
      nt = ni - nf ;
    }
/* Calcul des racines par Newton. On pourrait ameliorer l'algorithme en     */
/* baissant le degre a chaque racine trouvee.                               */
    for (i = 1 ; i <= M ; i++) Dc[i-1] = i*Coef[i] ;
    n = 0 ;
    for (k = 0 ; k < ntot ; k++)
    {
      Xa = Xs[k] ; Xb = Xs[k+1] ; mult = mu[k] ;
      pas = dx = (Xb-Xa)/2. ; x = (Xa+Xb)/2. ;
      while (fabs(pas) > Epsx)
      {
      /* Calcul de la racine dans l'intervalle par Newton */
	Pol = valpol(Coef, M, x) ;
	Dpol = valderpol(Coef, M, x) ;
	if (Dpol != 0.)
	{ pas = -mult*Pol/Dpol ; x += pas ; }
	if (Dpol == 0. || x < Xa || x > Xb)
	{
	    dx /= 2. ;
	    x = Xb - dx ;
	    nt = sgn_sturm1 (A, nd, x, &I0) - sgn_sturm1(A, nd, Xb, &I0) ;
	    if (nt == 0) Xb = x ;
	    else Xa = x ;
	    pas = dx = (Xb-Xa)/2. ; x = (Xa+Xb)/2. ;
	}
      }
      for (i = 0 ; i < mu[k] ; i++) RacR[n++] = x ;
    }
    free (A) ;
    return n ;
}

/*--------------------------------------------------------------------------*/
/*   Calcul des racines reelles d'un polynome dans l'intervalle (Xi, Xf)    */
/*   Les racines multiples ne sont comptes qu'une fois                     */
/*    Renvoi le nombre de racines calculees ou -1 si echec allocation       */
/*--------------------------------------------------------------------------*/
int roots_sturmod (double *RacR, double *Coef, int M, 
								double Xi, double Xf, double Epsx) 
{
int *nd ;
int i, k, n, ni, nf, mi, mf, mt, ntot, nt, m, I = 0, I0, flag = 1 ;
double *A, *Xs, *mu, *Dc ;
double Xa, Xb, x, dx, Pol, Dpol, pas, mult ;
int dim = M+1;

    A = (double *) malloc ((dim+4)*dim*sizeof(double)) ;
	if (A == NULL) { perreur("Echec allocation dans roots_sturm"); return -1 ;}
	Xs = A + dim*dim;
	mu = Xs + dim;
	Dc = mu + M;
	nd = (int *) (Dc+M);

    nd[0] = M ;
/* Calcul nombre de racines par les suites de sturm */
    for (i = 0 ; i <= M ; i++) A[i] = Coef[i] ;
    ni = sgn_sturmM (&mi, A, nd, Xi, &I, &I0) ;
    nf = sgn_sturmM (&mf, A, nd, Xf, &I, &I0) ;
    ntot = nt = ni - nf ;
/* Isolation des racines dans des intervalles */
    k = 0 ;
    while (nt != 0)
    {
      dx = Xf - Xi ;
      if (flag)
      {
		  /* Intervalle pour la racine la + gde */
		 x = Xi ; n = ni ; mt = mi - mf;
		 while (nt != 1)
		 {
		   dx /= 2. ; x = Xf - dx ;
		   n = sgn_sturmM (&m, A, nd, x, &I, &I0) ;
		   nt = n - nf ; mt = m - mf ;
		   /* if (dx < Epsx) break ; */
		   if (nt == 0) { Xf = x ; nf = n ; mf = m ; }
		 }
		 Xs[ntot-k] = Xf ; Xs[ntot-k-1] = x ;
		 mu[ntot-k-1] = mt ;
		 Xf = x ; nf = n ; mf = m ;       /* Pour recherche racine suivante */
      }
      else
      {
      /* Intervalle pour la racine la + petite */
	 x = Xf ; n = nf ; mt = mi - mf ;
	 while (nt != 1)
	 {
	    dx /= 2. ; x = Xi + dx ;
	    n = sgn_sturmM (&m, A, nd, x, &I, &I0) ;
	    nt = ni - n ; mt = mi - m ;
	    /* if (dx < Epsx) break ; */
	    if (nt == 0) { Xi = x ; ni = n ; mi = m ; }
	 }
	 mu[k] = mt ;
	 Xs[k] = Xi ; Xs[++k] = x ;
	 Xi = x ; ni = n ; mi = m ;       /* Pour recherche racine suivante */
      }
      flag = 1 - flag ;
      nt = ni - nf ;
    }
/* Calcul des racines par Newton. On pourrait ameliorer l'algorithme en     */
/* baissant le degre a chaque racine trouvee.                               */
    for (i = 1 ; i <= M ; i++) Dc[i-1] = i*Coef[i] ;
    n = 0 ;
    for (k = 0 ; k < ntot ; k++)
    {
      Xa = Xs[k] ; Xb = Xs[k+1] ; mult = mu[k] ;
      pas = dx = (Xb-Xa)/2. ; x = (Xa+Xb)/2. ;
      while (fabs(pas) > Epsx)
      {
      /* Calcul de la racine dans l'intervalle par Newton */
	Pol = valpol(Coef, M, x) ;
	Dpol = valderpol(Coef, M, x) ;
	if (Dpol != 0.)
	{ pas = -mult*Pol/Dpol ; x += pas ; }
	if (Dpol == 0. || x < Xa || x > Xb)
	{
	    dx /= 2. ;
	    x = Xb - dx ;
	    nt = sgn_sturm1 (A, nd, x, &I0) - sgn_sturm1(A, nd, Xb, &I0) ;
	    if (nt == 0) Xb = x ;
	    else Xa = x ;
	    pas = dx = (Xb-Xa)/2. ; x = (Xa+Xb)/2. ;
	}
      }
      RacR[n++] = x ;
    }
    free (A) ;
    return n ;
}


/*--------------------------------------------------------------------------*/
/*  Calcul du nombre de racines reelles d'un polynome dans l'intervalle X,Y */
/*--------------------------------------------------------------------------*/
/*  Les coefficients en degre croissant du polynome doivent etre ranges     */
/*  dans A[0]. Son degre est specifie dans nd[0].                           */
/*  Si I vaut 0, la routine calcule la suite polynomiale de Sturm qui est   */
/*  rangee dans A[i], degre dans nd[i], pour i = 1 a I qui est calcule ainsi*/
/*  que I0.                                                                 */
/*  Si I est different de 0, c'est que la suite polynomiale de sturm, I et  */
/*  I0 ont deja ete calcules.                                               */
/*  A partir de cette suite, par analyse du nombre des changements de signe */
/*  de ces polynomes en X et en Y, on en deduit le nombre total de racines  */
/*  reelles dans cet intervalle -> dans ntot. Le nombre de racines reelles  */
/*  distinctes dans cet intervalle est retourn par le return.              */
/*--------------------------------------------------------------------------*/
int nbracs (int *ntot, double *A, int *nd, double X, double Y,
							      int *I, int *I0)
{
int n, m ;
    n = sgn_sturmM (ntot, A, nd, X, I, I0) - sgn_sturmM (&m, A, nd, Y, I, I0) ;
    *ntot -= m ;
    return n ;
}


/*--------------------------------------------------------------------------*/
/* Routine interne : Methode Newton-Raphson en complexe.                    */
/*--------------------------------------------------------------------------*/
/* Recherche de la racine du polynome C[N]+C[N-1].X+ ...+C[0].X^N a partir  */
/* de la valeur initiale Z = (rz) + i.(iz) . Renvoi la valeur trouvee dans  */
/* la valeur initiale et la derniere valeur de |Delta_z| par return.        */
/* On considere qu'il y a convergence lorsque |Delta_z| < Eps1lim. La       */
/* recherche de la limite n'est toutefois pas arretee lorsque cette         */
/* condition est atteinte. Elle est poursuivie tant que |Delta_z| diminue.  */
/* On arrete la recherche lorsqu'il y a convergence et que |Delta_z| se met */
/* a augmenter, ou bien lorsque le nombre maxit d'iterations est atteint.   */
/* Dans ce cas on renvoi la valeur qui correspond au plus petit |Delta_z|   */
/* rencontre pendant la recherche.                                          */
/*--------------------------------------------------------------------------*/
static double newton_cpx (double *rz, double *iz, double *C, int N, int maxit,
							       double Eps1lim)
{
double RZ, IZ, PolR, PolI, DPolR, DPolI, IZNm1, RZNm1, RZN, IZN,
    bid, DPRsDPI, Den, DPIsDPR, DX, DY, Epsz, Epszmem, Epsmin, Rmin, Imin ;
int I, iter;

    Rmin = RZ = *rz ; Imin = IZ = *iz ; Epsmin = Epszmem = DBL_MAX ;
    for (iter = 0 ;iter < maxit ; iter++)
    {
    /* Calcul P(Z) et P'(Z) en Z = RZ+iIZ */
	RZNm1 = 1.0 ; IZNm1 = 0.0 ;                      /* Z puissance 0 */
	PolR = C[N] ; PolI = 0.0 ;                    /* Ini calcul P(Z) */
	DPolR = 0.0 ; DPolI = 0.0 ;                   /* Ini calcul P'(Z) */
	for (I = 1 ; I <= N ; I++)
	{
	    /* Il faudrait tester les overflow et underflow sur RZN et IZN  */
	    RZN=RZ*RZNm1-IZ*IZNm1 ; IZN=RZ*IZNm1+IZ*RZNm1 ;/* Z puissance I */
	    bid = C[N-I] ;
	    PolR += bid*RZN ; PolI += bid*IZN ;              /* Calcul P(Z) */
	    bid *= I ;
	    DPolR += bid*RZNm1 ; DPolI -= bid*IZNm1 ;       /* Calcul P'(Z) */
	    RZNm1 = RZN ; IZNm1 = IZN ;
	}
	if (DPolR != 0. || DPolI != 0.)   /* Calcul Delta_Z = - P(Z)/P'(Z)  */
	{                                 /* sinon conserve ancien Delta_Z  */
	    if (fabs(DPolR) < fabs(DPolI))
	    {
	      DPRsDPI = DPolR/DPolI ; Den = DPolI + DPolR*DPRsDPI ;
	      DX = (PolI - PolR*DPRsDPI)/Den ;
	      DY = -(PolR + PolI*DPRsDPI)/Den ;
	    }
	    else
	    {
	      DPIsDPR = DPolI/DPolR ; Den = DPolR + DPolI*DPIsDPR ;
	      DX = (PolI*DPIsDPR - PolR)/Den ;
	      DY = -(PolR*DPIsDPR + PolI)/Den ;
	    }
	}
	/* Convergence lorsqu'arret de la decroissance de Epsz (< Eps1lim) */
	Epsz = fabs(DY)+fabs(DX) ;
	if (Epsz >= Epszmem && Epsz <= Eps1lim)
	{ *rz = RZ ; *iz = IZ ; return Epsz ; }

	/* Z = Z + Delta_Z */
	RZ += DX ; IZ += DY ; Epszmem = Epsz ;
	if (Epsz < Epsmin) { Epsmin = Epsz ; Rmin = RZ ; Imin = IZ ; }
    }
    *rz = Rmin ; *iz = Imin ; return Epsmin ;
}

/*--------------------------------------------------------------------------*/
int roots_newton (double *RacR, double *RacI, double *Coef, int M)
/*--------------------------------------------------------------------------*/
/* Calcul de toutes les racines (partie reelle RacR et imaginaire RacI)     */
/* d'un polynome de degre M a coefficients reels.                           */
/* Les coefficientss sont donnes avec les degres dans l'ordre croissant.    */
/* RETOUR : Nb de racines calculees                                         */
/*--------------------------------------------------------------------------*/
/* La methode employee consiste a calculer une racine en complexe a partir  */
/* d'une initialisation complexe proche de zero par la methode de Newton    */
/* Raphson (procedure suivante).                                            */
/* Si elle ne converge pas une initialisation plus grande, dans une autre   */
/* direction, est essayee. Il y a abandon apres plusieurs echecs.           */
/* Lorsqu'il y a convergence, on effectue un abaissement de degre en        */
/* divisant par un polynome du 1er ou 2eme degre suivant que la racine      */
/* trouvee etait reelle ou complexe, et on itere la meme procedure sur le   */
/* polynome de degre inferieur jusqu'a ce qu'il ne reste plus qu'une poly   */
/* de degre 0.                                                              */
/* Pour les polynomes resultant de plusieurs divisions successives, une     */
/* procedure de raffinage de la racine trouvee est entreprise. On essaie de */
/* faire converger le polynome initial (avant toute division) a partir de   */
/* la racine trouvee sur le polynome divise (une seule tentative, avec un   */
/* nombre d'iterations reduit). S'il y a convergence et que la limite       */
/* atteinte n'est pas trop eloignee de la racine trouvee sur le polynome    */
/* divise, on considere que c'est la meme racine, plus precise et elle lui  */
/* est sustituee. On peut en effet reconverger vers une racine differente   */
/* deja traitee, qu'il ne faut pas prendre en compte, car on trouverait une */
/* racine double inexistante.                                               */
/*--------------------------------------------------------------------------*/
{
/*      PARAMETRES A REGLER DEPENDANT DES MACHINES OU DES APPLICATIONS      */
#define MAXIT 100 /* Nombre max d'iterations avec une initialisation donnee */
#define MAXRAF 20 /* Idem pour la phase de raffinage						*/
#define MAXINIT 5 /* Nombre max d'initialisation des iterations				*/
double Epsz, ZtoR, Separ ;
int N_non_raf ;
/*            ---            ----           ----            ---             */
double fabs (double) ;
double RZ, IZ, RZO, IZO, bid, RZmem, IZmem, Som2, Prod2, *C, *CMemo ;
double eps ;
int I, Inis, maxit, N, Nmem, K, Rafine, ndiv ;

    K = 0 ;                                  /* Nombre de racines calculees */
    /* Retire terme de haut degre inutiles */
    while (M > 0 && Coef[M]+Coef[M-1] == Coef[M-1]) M-- ;
    if (M < 1) return 0 ;

	C = (double *) malloc((2*M+2)*sizeof(double));
	if (C == NULL) { perreur("Echec allocation dans roots_newton"); return -1 ;}
	CMemo = C + M +1;

    /* Copie de travail des coefficients */
    for (I = 0 ; I <= M ; I++) CMemo[I] = Coef[I] ;
    /* Collecte racines nulles */
    while (CMemo[0] == 0. && M >= 0)
    {
      RacR[K] = 0. ; RacI[K++] = 0.;
      for (I = 1 ; I <= M ; I++) CMemo[I-1] = CMemo[I] ;
      M-- ;
    }

	if (M < kRootsIter || M < 5)
	{
		switch (M) {
		case 1 : RacI[K] = 0 ; RacR[K] = - CMemo[0]/CMemo[1] ; K++ ; break;
		case 2 : racpol2 (RacR+K, RacI+K, CMemo) ; K+=2 ; break;
		case 3 : racpol3 (RacR+K, RacI+K, CMemo) ; K+=3 ; break;
		case 4 : racpol4 (RacR+K, RacI+K, CMemo) ; K+=4 ; break; 
		}
		free(C); 
		return K ;
	}

    /* Renversement ordre (decroissant)  C : poly de travail de degre N a 1 */
    for (I = 0 ; I <= M ; I++) C[M-I] = CMemo[I] ;
    for (I = 0 ; I <= M ; I++) CMemo[I] = C[I] ;   /* Poly original degre N */

/*      PARAMETRES A REGLER DEPENDANT DES MACHINES OU DES APPLICATIONS      */
    Epsz = 1.0e-5 ;                        /* Seuil d'examen de convergence */
    ZtoR = Epsz ;		     /* Si Imag/Reel < ZtoR ==> Rac. reelle */
    N_non_raf = 3 ;                /* Nombre de divisions avant de raffiner */
    Separ = 1.e-3 ;          /* Separation des limites directes et rafinees */
/*            ---            ----           ----            ---             */
    ndiv = 0 ;

    Nmem = N = M ;
    /* Recherche racine du polynome de degre N restant */
    while (N > 0)
    {
      /* Point complexe de depart des iterations */
      IZmem = IZ = IZO = -.0500101 ; RZmem = RZ = RZO = -0.1000101 ;
      Inis = 1 ;                        /* Numero initialisation differente */
      /* Calcul sur pol. reduit puis sur pol. complet pour + gde precision  */
      Rafine = 0 ;
      while (1) /* Sortie par break ; */
      {
	maxit = (Rafine == 0) ? MAXIT : MAXRAF ;
	eps = newton_cpx (&RZ, &IZ, C, N, maxit, Epsz) ;
	if (eps > Epsz) /* Pas converge */
	{
	  if (Rafine == 0) /* Si pas conv : essai un nouveau point de depart */
	  {
	    if (++Inis > MAXINIT) break ; /* return K ;*/                     /* Abandon */
	    RZ = RZO ; RZO = -10.0*IZO ; IZ = IZO = -10.0*RZ ; RZ = RZO ;
	    continue ;                    /* Continue recherche 1ere limite */
	  }
	  else { RZ = RZmem; IZ = IZmem; break ;}/* Sinon conserve anc. lim */
	}
	else /* On a converge */
	{
	  if (Rafine == 0)                             /* Trouve 1ere limite */
	  {
	  /* Boucle supplementaire avec polynome complet */
	    if (N != M && ndiv > N_non_raf)
	    {
	      for (I = 0 ; I <= M ; I++)
	      {bid = C[I] ; C[I] = CMemo[I] ; CMemo[I] = bid ;}
	      RZmem = RZ ; IZmem = IZ ; N = M ;
	      Rafine = 1 ;                           /* Deja trouve une limite */
	    }
	    else break ;
	  }
	  else
	  {
	    if (fabs(RZmem-RZ)+fabs(IZmem-IZ) > Separ)
	    { RZ = RZmem ; IZ = IZmem ; }
	    break ;        /* Trouve 2eme limite => sortie recherche limite */
	  }
	}
      }
   /* Sortie recherche limite */
      /* Recupere polynome reduit */
      if (Rafine)
      {
	 N = Nmem ;
	 for (I = 0 ; I <= M ; I++)
	 {bid = C[I] ; C[I] = CMemo[I] ; CMemo[I] = bid ;}
      }
      /* Teste si Z reel ou complexe et fait la division polynomiale */
      if (IZ == 0. || (RZ != 0. && (fabs(IZ/RZ)  <= ZtoR)) || N == 1)
      {
	  N-- ;
	  for (I = 1 ; I <= N ; I++) { C[I] += RZ*C[I-1] ; }
	  RacR[K] = RZ ; RacI[K++] = 0. ;
      }
      else
      {
	  Som2 = RZ + RZ ; Prod2 = RZ*RZ + IZ*IZ ; N -= 2 ;
	  C[1] += Som2*C[0] ;
	  for (I = 1 ; I < N ; I++)
	  { C[I+1] += Som2*C[I] - Prod2*C[I-1] ; }
	  RacR[K] = RZ ; RacI[K++] = IZ ;
	  RacR[K] = RZ ; RacI[K++] = -IZ ;
      }
      Nmem = N ;
      ndiv++ ;
    } /* while N > 1 */
	free(C);
    return K ;
}


/*--------------------------------------------------------------------------*/
/*   Algorithme quotient-difference de RUTISHAUSER         MLL - 14/04/94   */
/*   Adaptation de l'algorithme fortran PRQD de la bibli mathematique IBM   */
/*--------------------------------------------------------------------------*/
static int qd_rutis (double *A, double *B, int N, double *coef)
/*--------------------------------------------------------------------------*/
/* Dans un premier temps, le polynome P(x) est norme (coef de + haut degre  */
/* mis a 1). Soit A0(x) ce polynome de degre N. Sa derivee est est egalement*/
/* normee. Soit B1(x) ce polynome de degre N-1. On decompose le quotient    */
/*  A0        z1 |   r1 |   z2 |   r2 |         zN-1|  rN-1                 */
/* ---- = x - ---- - ---- - ---- - ---- - ... - ---- - ----                 */
/*  B1          1      x      1      x            1    x-zN                 */
/* en fraction continuee au moyen de divisions euclidiennes successives.    */
/* A0 = B1 (x-z1) + R2 = x B1 - z1 (B1 - R2/z1) fournit z1 et le polynome   */
/* norme A1 = B1 - R2/z1 de degre N-1. Par ailleurs on pose r1 B2 = R2/z1   */
/* ou B2 est un polynome norme de degre N-2, ce qui definit r1. Et ainsi de */
/* suite :                                                                  */
/* AN-2 = BN-1 (x-zN-1) + RN = x BN-1 - zN-1 (BN-1 - RN/zN-1) fournit zN-1  */
/* et le polynome norme AN-1 = BN-1 - RN/zN-1 de degre 1. Par ailleurs la   */
/* relation rN-1 BN = RN/zN-1 ou BN est la constante 1 definit rN-1.        */
/* Finalement AN-1 = x - zN definit directement zN.                         */
/* En fait pour caluler les termes successifs, on utilise les relations :   */
/* Aj-1 = x Bj - zj Aj  ==>  zj Aj = x Bj - Aj-1                            */
/* Aj = Bj + rj Bj+1    ==>  rj Bj+1 = Aj - Bj                              */
/* Ensuite des iterations sont effectues sur les termes zj et rj qui amenent*/
/* les zj a converger vers les racines.                                     */
/*--------------------------------------------------------------------------*/
/* Entree :                                                                 */
/* 	N	: degre du polynome.                                        */
/*	coef	: coefficients du polynome donnes en croissant              */
/*	A,B	: tableaux de dimension N+1 !!!!!                           */
/* Sortie :                                                                 */
/*	A,B 	: parties reelles et imaginaires de la solution.            */
/*		  A(N+1) contient l'erreur relative max sur le polynome     */
/*		  reconstitue a partir des racine, calcule uniquement si    */
/*		  le renvoi est <= 0.                                       */
/*	retour  :                                                           */
/*	      0	: Pas d'erreur                                              */
/*	      1 : Ne converge pas                                           */
/*	      2 : polynome nul                                              */
/*	      3 : division par zero                                         */
/*	      4 : pas de fraction continue                                  */
/*	     -1 : precision tres faible                                     */
/*--------------------------------------------------------------------------*/
{
double fabs (double), sqrt (double) ;
int Np1, ier, k1, kM, jdeb, i, j, k, iter, iend, jend, kM1, maxit ;
int flag_mult ;
double peps[36], d, eps, tol, esav, decx, teps, tmt,
	p, s, r, t, u, v, w = 0.;

/* Adressage a partir de 1 comme en fortran */
    A -= 1 ; B -= 1 ; coef -= 1 ;
/* Retire les coefficients nuls de degre max */
    Np1 = N+1 ;
    while (coef[Np1] == 0.)
    { Np1-- ; if (Np1 <= 1) return 2 ;}
    N = Np1-1 ;

/* Normalisation du polynome dans A et extraction des racines nulles */
    d = 1./coef[Np1] ;
  /*
	A,B contiendront les racines nulles pour j = 1 a k1-1.
	Les coefficients normes du plus haut degre au plus bas, pour
	j = k1 a Np1 sont dans A(j) avec A(k1) = 1 (donc inutile).
  */
    k1 = 1 ; k = Np1 + 1 ; jdeb = 1 ;
    for (i = 1 ; i <= Np1 ; i++)
    {
		if (jdeb == 1 && coef[i] == 0.)            /* collecte racines nulles */
		{ k++ ; A[k1] = B[k1] = 0. ; k1++ ; }
		else                    /* A[j] : j de Np1 a 1 en sens inverse de i */
		{ jdeb = 2 ; coef[i] *= d ; A[k-i] = coef[i] ; }
    }
/* Initialisations */
    ier = 0 ;
    eps = 1.e-100 ;	                          /* test division par zero */
    teps = 1.e-14 ;				        /* test convergence */
    tol = 1.e-4  ;
    maxit = 50*Np1 ;
    iter = 0 ;
    N = Np1 - 1 ;			                               /* N */
    esav = 0. ;
    A[k1] = 0. ;                    /* car coef de + haut degre connu = 1 */
/* Debut rebouclage */
/*
    k1 : indice 1er coef du polynome en cours a diviser
    kM : indice du dernier coef. Si un facteur commun est trouve au cours de
    la division, kM sera recule, de telle maniere que le polynome a traiter
    en premier soit de k1 a kM, et le facteur commun a traiter ensuite soit
    de kM a Np1.
*/
deb_quotient:
    kM = Np1 ;
/* Calcul du polynome derive B */
    decx = Np1 - k1 ;                     /* degre effectif pour derivation */
    B[k1] = decx ;               /* Coef plus haut degre du polynome derive */
    flag_mult = 0 ;      	                   /* Pas de facteur commun */
    for (i = k1+1 ; i <= Np1 ; i++)
    {
		decx -= 1. ;
		peps[i] = eps * fabs (A[i]) + eps ;      /* Epsilon pour test a zero */
		B[i] = A[i] * decx ;                 /* Coefs suivant du pol derive */
    }                                                /* decx = 0. en sortie */
/* Test dimension restante */
    if (N < k1) goto fin_ok ;
    jend = N - 1 ;				                     /* N-1 */
/* Calcul de la fraction continuee */
    for (i = k1 ; i <= jend ; i++)                            /* 1 a N -1 */
    {
		/* Test si racines multiples (inutile pour i = k1) */
		if (fabs(B[i]) <= peps[i+1])
		{ /* Facteur commun traite ulterieurement dans A(kM=i) a A(Np1) */
			kM = i ;
			/* Teste si les autres coefs sont =/= 0. */
			for (k = i+1 ; k <= jend ; k++)
			{ if (fabs(B[k]) > peps[k+1]) return 4 ;	}
			flag_mult = 1 ;     /* Facteur commun, la division est terminee */
			break ;
		}
		/* Division euclidienne */
		/* Les zI et les rI sont stockes decales de 1 : z1 en A(2), ..., zN
		   en A(N+1), r0 en B(1), ...,rN-1 en B(N-1). r0 est inutilise
		   mais sert a normaliser B0 = P' */
		for (k = i ; k <= N ; k++)
		{
			B[k+1] /= B[i] ;           /* Normalise B, rI-1 = B[I] en place */
			A[k+1] = B[k+1] - A[k+1] ;               /* zI AI = x BI - AI-1 */
			if (k == i) /* teste zI */
			{ if (fabs(A[i+1]) <= peps[i+1]) return 4 ; }
			else
			{
				A[k+1] /= A[i+1] ;    /* Normalise AI, zI = A[i+1] en place */
				peps[k+1]  /= fabs(A[i+1]) ;               /* Pour les tests */
				B[k] = A[k+1] - B[k] ;               /* rI+1 BI+1 = AI - BI */
			}
		}
    }
/* ----- */
	if (fabs(B[i]) <= teps)
	{ /* Facteur commun traite ulterieurement dans A(kM=i) a A(Np1) */
	    kM = i ;
	    /* Teste si les autres coefs sont =/= 0. */
	    for (k = i+1 ; k <= jend ; k++)
	    { if (fabs(B[k]) > teps) return 4 ;	}
	    flag_mult = 1 ;     /* Facteur commun, la division est terminee */
	}
/* ----- */
    if (!flag_mult) A[Np1] = -A[Np1] ;/* zN en place : oppose du dernier coef */
    kM1 = kM-1 ;
/*
    La partie traitee en difference va de k1 a kM1+1.
    La convergence n'est examinee qu'en queue. kM1 est decremente
    au fur et a mesure que des racines convergent, ce qui permet de ne
    faire les calculs iteratifs que les les parties restantes. C'est la
    deflation.
    La ligne du tableau QD traitee est :
    B(k1)=0 ,  A(k1+1),B(k1+1),...,B(kM1),A(kM1+1) , B(kM1+1)=0
    Le deplacement decx est initialement mis a 0 lors du calcul de P'(x).
*/
    B[k1] = 0. ;				                  /* r0 = 0 */
deb_diff:
    B[kM1+1] = 0. ;                                               /* rM = 0 */
  /* EXamen s'il ne reste qu'une racine ou une constante */
    if (kM1 <= k1) 			                          /* kM > 1 */
    {
		if (kM1 == k1)                       /* kM = 1 => Facteur 1er degre */
		{ A[k1+1] += decx ; B[k1+1] = 0. ; }        /* Rac. reelle en place */
	   /* Test si existe diviseur commun non factorise */
		B[k1] = esav ;      /* Recupere partie imaginaire racine precedente */
		if (kM == Np1) goto fin_ok ;
		/* Initialise algo_qd pour diviseur commun */
		k1 = kM ;                          /* Debut polynome facteur commun */
		esav = B[k1] ;  /* Memo im racine deja calculee car sera mis a zero */
		goto deb_quotient ;
    }

/* Debut iteration algorithme QD */
	jdeb = k1+1 ;
	jend = kM1-1 ;   					             /* M-1 */
	tmt = 0.01 ;
do	{
		iter++ ;
		p = A[kM1+1] ;                                                   /* zM */
		r = fabs(B[kM1]) ;                                             /* rM-1 */
	  /* Test convergence vers racine simple : |rM-1| < eps */
		if (r <= teps)          /* Calcul racine simple : reduction degre par 1 */
		{
			A[kM1+1] += decx ;
			kM1-- ; goto deb_diff ;
		}
		s = fabs(B[jend]) ;					            /* rM-2 */
	  /* Si |rM-2| < |rM-1| */
		if (s <= r)
		{
		/* Calcul du deplacement pour une racine double
		   Equation quadratique :
		   X**2 - [zM-1 + zM + rM-1]*X + zM*zM-1 = 0 */
			p = 0.5*(A[kM1]+ A[kM1+1] + B[kM1]) ;	              /* demi-somme */
			d = p*p - A[kM1]*A[kM1+1] ;                         /* discriminant */
			t = sqrt(fabs(d)) ;                                      /* radical */
		  /* test pour convergence vers couple de racines |rM-2| < eps */
			if (s <= teps)
			{ /* Calcule racines couplees */
				p += decx ;			            /* recale partie reelle */
				if (d < 0.)                         /* couple racines complexes */
				{
				A[kM1] = A[kM1+1] = p ; B[kM1] = t ; B[kM1+1] = -t ;
				}
				else                                  /* couple racines reelles */
				{
				A[kM1] = p-t ; A[kM1+1] = p+t ; B[kM1] = 0. ;
				}
			  /* Reduction degre par 2 (deflation) */
				kM1 -= 2 ; goto deb_diff ;
			}
		  /* Calcul deplacement */
			if (d < 0.) /* cas racines complexes */
			{
			/* |rM-2| < 0.01 */
				if (s < tmt)
				{
					/* Initialisation */
					d = A[jdeb] + B[jdeb] - p ;
					/* test pour petit diviseur */
					if (fabs(d) <= peps[jdeb]) { ier = 3 ; goto rearrange ; }
					t /= d ; t *= t ;                     /* t = (t/d)**2 */
					u = B[jdeb]*A[jdeb+1]/(d*(1.+t)) ;
					v = d + u ;
					iter += 2 ;
					/* boucle pour deplacement complexe */
					for (j = jdeb ; j <= kM1 ; j++)
					{
						d = A[j+1] + B[j+1] - u - p ;
						/* test petit diviseur */
						if (fabs(v) <= peps[j] || fabs(d) <= peps[j+1])
						{ /* Perte precision : calcul couple de racine */
							if (j == kM1)    /* autorisee simplement a la fin */
							{
								decx += p ;
								if (fabs(B[jend]) > tol) { ier = 3 ; goto rearrange ; }
								p = 0.5*(v+d-B[jend]) ;
								d = p*p - (v-u)*(d-u*t-d*w*(1.+t)/A[jend]) ;
								t = sqrt(fabs(d)) ;
								/* Calcule racines couplees */
								p += decx ;
								if (d < 0.) /* couple racine complexe */
								{
									A[kM1] = A[kM1+1] = p ; B[kM1] = t ; B[kM1+1] = -t ;
								}
								else        /* couple racine reelle */
								{
									A[kM1] = p-t ; A[kM1+1] = p+t ; B[kM1] = 0. ;
								}
								/* Reduction degre par 2 (deflation) */
								kM1 -= 2 ;
								goto deb_diff ;
							} else { ier = 3 ; goto rearrange ; }
						}
						w = u * d / v ;
						t *= (v/d)*(v/d) ;
						A[j] = v + w - B[j-1] ;
						if (j < kM1) u = A[j+2] * B[j+1] /(d*(1.+t)) ;
						else u = 0. ;
						v = d + u - w ;
						/* test petit diviseur */
						if (fabs(A[j]) <= peps[j]) { ier = 3 ; goto rearrange ; }
						B[j] = w*v*(1.+t)/A[j] ;
					}
					A[kM1+1] = v - B[kM1] ;
				}
				/* |rM-1| > |rM-2| > 0.01 */
				else {p = 0.; goto rm2gtrm1 ;}
			}
			/* Calcul deplacement */
			else /* Cas racines reelles */
			{ p = (p >= 0.) ? p - t : p + t ;  r = s ; goto rm2gtrm1 ;}
		}
		/* |rM-2| > |rM-1| */
		else
		{
		/* Le deplacement est-il assez petit ? */
rm2gtrm1:	if (r >= tmt)
			{
				p = 0. ;
			}
			d = p ;
			for (j = jdeb ; j <= kM1 ; j++)
			{
				A[j] += B[j] - B[j-1] - d ;
				/* teste petit diviseur */
				if (fabs(A[j]) <= peps[j]) {ier = 3 ; goto rearrange ; }
				B[j] *= A[j+1]/A[j] ;
			}
			A[kM1+1] += -B[kM1] - d ;
		}
		decx += p ;
		teps *= 1.1 ;
		tmt *= 1.1 ;
	} while (iter < maxit) ;

/* convergence insatisfaisante */
    ier = 1 ;
  /* Rearrange les racines calculees */
rearrange:
    iend = kM - (kM1+1) ;
    B[k1] = esav ;
    for (i = 1 ; i <= iend ; i++)
    {
		j = k1 + i ; k = kM1 + 1 + i ;
		B[j] = B[k] ; A[j] = A[k] ;
    }
    Np1 = k1 + iend ;
    N = Np1 - 1 ;
/* Retour fin normale */
fin_ok:
    if (N <= 0) return ier ;
    for (i = 1 ; i <= N ; i++)
    { A[i] = A[i+1] ; B[i] = B[i+1] ; }
/* Calcul coefs a partir des racines */
    peps[Np1] = 1. ;
    iend = N-1 ;
    jdeb = 1 ;
    for (j = 1 ; j <= N ; j++)
    {
		k1 = Np1-j ;
		d = 0. ;
		p = A[k1] ;
		t = B[k1] ;
		if (t == 0.)
		{
			for (i = k1 ; i <= N ; i++)
			{ peps[i] = d - p * peps[i+1] ; d = peps[i+1] ; }
			continue ;
		}
		if (jdeb == 1)
		{ jdeb = 2 ; peps[k1] = 0. ;  continue ; }
		jdeb = 1 ;
		u = p*p + t*t ;
		p += p ;
		for (i = k1 ; i <= iend ; i++)
		{ peps[i] = d - p*peps[i+1] + u*peps[i+2] ; d = peps[i+1] ; }
		peps[N] = d - p ;
    }
    if (ier != 0) return ier ;
/* Comparaison avec coef pour test precision */
    p = 0. ;
    for (i = 1 ; i <= N ; i++)
    {
		if (coef[i] == 0.) d = fabs(peps[i]) ;
		else d = fabs((peps[i]-coef[i])/coef[i]) ;
		if (d > p) p = d ;
    }
    if (p >= tol) ier = -1 ;
    A[Np1] = p ; B[Np1] = 0. ;
    return ier ;
}

/*--------------------------------------------------------------------------*/
/*        Racines par l'algorithme quotient-difference de RUTISHAUSER       */
/*--------------------------------------------------------------------------*/
int roots_qd (double *RacR, double *RacI, double *Coef, int M)
/*--------------------------------------------------------------------------*/
/*	retour  : >= 0 : Nombre de racines calcules							*/
/*	      0	: Pas d'erreur													*/
/*	     -2 : Echec allocation												*/
/*--------------------------------------------------------------------------*/
{
double *A, *B, *CMemo;
int i, j, K ;

    K = 0 ;                                  /* Nombre de racines calculees */
    /* Retire terme de haut degre inutiles */
    while (M > 0 && Coef[M]+Coef[M-1] == Coef[M-1]) M-- ;
    if (M < 1) return 0 ;

	A = (double *) malloc((3*M+3)*sizeof(double));
	if (A == NULL) { perreur("Echec allocation dans roots_qd"); return -1 ;}
	B = A + M + 1;
	CMemo = B + M +1;

    /* Copie de travail des coefficients */
    for (i = 0 ; i <= M ; i++) CMemo[i] = Coef[i] ;
    /* Collecte racines nulles */
    while (CMemo[0] == 0. && M >= 0)
    {
      RacR[K] = 0. ; RacI[K++] = 0.;
      for (i = 1 ; i <= M ; i++) CMemo[i-1] = CMemo[i] ;
      M-- ;
    }

	if (M < kRootsIter || M < 5)
	{
		switch (M) {
		case 1 : RacI[K] = 0 ; RacR[K] = - CMemo[0]/CMemo[1] ; K++ ; break;
		case 2 : racpol2 (RacR+K, RacI+K, CMemo) ; K+=2 ; break;
		case 3 : racpol3 (RacR+K, RacI+K, CMemo) ; K+=3 ; break;
		case 4 : racpol4 (RacR+K, RacI+K, CMemo) ; K+=4 ; break; 
		}
		free(A); 
		return K ;
	}


  j = qd_rutis (A, B, M, CMemo) ;
  if (j <= 0)
  {
	for (i = 0 ; i < M ; i++) { RacR[K+i] = A[i] ; RacI[K+i] = B[i] ; }
	if (j == 0 || A[M] < 0.01) K += M ;
  }
  free(A);
  return K;
}


