/************************************************************************/
/* FICHIER QUATERN.C : Utilitaires pour calcul avec les quaternions     */
/************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "quatern.h"
#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif
#define DEUXPI 6.28318530717958647692
/*************************************************************************
   quaternion standard : q[0] est rendu positif.
***************************************************************************/
double *qnstd( double *q)
{
    if (q[0] < 0.)
    {
	q[0] = -q[0] ;
	q[1] = -q[1] ;
	q[2] = -q[2] ;
	q[3] = -q[3] ;
    }
return q ;
}
/*************************************************************************
   quaternion meme nappe : q est rendu dans la meme nappe que qref.
***************************************************************************/
double *qnprox (double *q, double *qref)
{
double s = q[0]*qref[0]+q[1]*qref[1]+q[2]*qref[2]+q[3]*qref[3] ;

    if (s < 0.)
    {
	q[0] = -q[0] ;
	q[1] = -q[1] ;
	q[2] = -q[2] ;
	q[3] = -q[3] ;
    }
return q ;
}
/***************************************************************************
  Conjugue d'un quaternion
****************************************************************************/
double *tqn(double *q)
{
	q[1] = -q[1];
	q[2] = -q[2];
	q[3] = -q[3];
	return q ;
}
/***************************************************************************
  Inverse d'un quaternion
****************************************************************************/
double *iqn(double *q)
{
double d = q[0]*q[0]+q[1]*q[1]+q[2]*q[2]+q[3]*q[3];

	if (d!=0);
	{
		q[0] =  q[0]/d;
		q[1] = -q[1]/d;
		q[2] = -q[2]/d;
		q[3] = -q[3]/d;
	}
	return q ;
}
/**************************************************************************
   Produit quaternionique : a=b*c
***************************************************************************/
double *qnqn (double *a, double *b, double *c)
{
double a0, a1, a2 ;

	a0   = b[0]*c[0] - b[1]*c[1] - b[2]*c[2] - b[3]*c[3] ;
	a1   = b[1]*c[0] + b[0]*c[1] - b[3]*c[2] + b[2]*c[3] ;
	a2   = b[2]*c[0] + b[3]*c[1] + b[0]*c[2] - b[1]*c[3] ;
	a[3] = b[3]*c[0] - b[2]*c[1] + b[1]*c[2] + b[0]*c[3] ;
	a[0] = a0 ;
	a[1] = a1 ;
	a[2] = a2 ;
return a ;
}
/***************************************************************************
   Produit quaternionique : a=b'*c
***************************************************************************/
double *tqnqn( double *a, double *b, double *c)
{
double a0, a1, a2 ;

	a0  = b[0]*c[0] + b[1]*c[1] + b[2]*c[2] + b[3]*c[3] ;
	a1  =-b[1]*c[0] + b[0]*c[1] + b[3]*c[2] - b[2]*c[3] ;
	a2  =-b[2]*c[0] - b[3]*c[1] + b[0]*c[2] + b[1]*c[3] ;
	a[3]=-b[3]*c[0] + b[2]*c[1] - b[1]*c[2] + b[0]*c[3] ;
	a[0]=a0 ;
	a[1]=a1 ;
	a[2]=a2 ;
return a ;
}
/***************************************************************************
   Produit quaternionique : a=b*c'
***************************************************************************/
double *qntqn(double *a, double *b, double *c)
{
double a0, a1, a2;

	a0  =b[0]*c[0] + b[1]*c[1] + b[2]*c[2] + b[3]*c[3] ;
	a1  =b[1]*c[0] - b[0]*c[1] + b[3]*c[2] - b[2]*c[3] ;
	a2  =b[2]*c[0] - b[3]*c[1] - b[0]*c[2] + b[1]*c[3] ;
	a[3]=b[3]*c[0] + b[2]*c[1] - b[1]*c[2] - b[0]*c[3] ;
	a[0]=a0 ;
	a[1]=a1 ;
	a[2]=a2 ;
return a ;
}
/***************************************************************************
   Produit quaternionique : a=inv(b)*c
***************************************************************************/
double *iqnqn( double *a, double *b, double *c)
{
double ib[4];
return qnqn(a, iqn(cpyqn(ib,b)),c) ;
}
/***************************************************************************
   Produit quaternionique : a=b*inv(c)
***************************************************************************/
double *qniqn( double *a, double *b, double *c)
{
double ic[4];
return qnqn(a, b, iqn(cpyqn(ic,c))) ;
}
/***************************************************************************
   Produit quaternionique : dq=.5*q*w        utilise dans le
   calcul de la derivee d'un quaternion en fonction de omega dans Rm.
***************************************************************************/
double *omm2dqn (double *dq, double *w, double *q)
{
	dq[0] = (-q[1]*w[0] - q[2]*w[1] - q[3]*w[2] )/2. ;
	dq[1] = ( q[0]*w[0] - q[3]*w[1] + q[2]*w[2] )/2. ;
	dq[2] = ( q[3]*w[0] + q[0]*w[1] - q[1]*w[2] )/2. ;
	dq[3] = (-q[2]*w[0] + q[1]*w[1] + q[0]*w[2] )/2. ;
return dq ;
}
/***************************************************************************
   Produit quaternionique : w = 2.*q'*dq        utilise dans le
   calcul de omega dans Rm en fonction du quaternion et de sa derivee.
***************************************************************************/
double *dqn2omm(double *w, double *dq,double *q)
{

	w[0] = (-q[1]*dq[0] + q[0]*dq[1] + q[3]*dq[2] - q[2]*dq[3] )*2. ;
	w[1] = (-q[2]*dq[0] - q[3]*dq[1] + q[0]*dq[2] + q[1]*dq[3] )*2. ;
	w[2] = (-q[3]*dq[0] + q[2]*dq[1] - q[1]*dq[2] + q[0]*dq[3] )*2. ;
return w ;
}
/***************************************************************************
   Produit quaternionique : dq=.5*w*q        utilise dans le
   calcul de la derivee d'un quaternion en fonction de omega dans Ro.
***************************************************************************/
double *omo2dqn (double *dq, double *w, double *q)
{
	dq[0] = (-q[1]*w[0] - q[2]*w[1] - q[3]*w[2] )/2. ;
	dq[1] = ( q[0]*w[0] + q[3]*w[1] - q[2]*w[2] )/2. ;
	dq[2] = (-q[3]*w[0] + q[0]*w[1] + q[1]*w[2] )/2. ;
	dq[3] = ( q[2]*w[0] - q[1]*w[1] + q[0]*w[2] )/2. ;
return dq ;
}
/***************************************************************************
   Produit quaternionique : w = 2.*dq*q'        utilise dans le
   calcul de omega dans Ro en fonction du quaternion et de sa derivee.
***************************************************************************/
double *dqn2omo(double *w, double *dq,double *q)
{

	w[0] = (-q[1]*dq[0] + q[0]*dq[1] - q[3]*dq[2] + q[2]*dq[3] )*2. ;
	w[1] = (-q[2]*dq[0] + q[3]*dq[1] + q[0]*dq[2] - q[1]*dq[3] )*2. ;
	w[2] = (-q[3]*dq[0] - q[2]*dq[1] + q[1]*dq[2] + q[0]*dq[3] )*2. ;
return w ;
}
/***************************************************************************
  Generation de la matrice de passage A associee au quaternion unitaire q
***************************************************************************/
double *qnu2mrot (double *A, double *q)
{
double q02, q12, q22, q32, q0q1, q0q2, q0q3, q1q2, q1q3, q2q3 ;

	q02 = q[0]*q[0] ;
	q12 = q[1]*q[1] ;
	q22 = q[2]*q[2] ;
	q32 = q[3]*q[3] ;

	q0q1 = q[0]*q[1] ;
	q0q2 = q[0]*q[2] ;
	q0q3 = q[0]*q[3] ;
	q1q2 = q[1]*q[2] ;
	q1q3 = q[1]*q[3] ;
	q2q3 = q[2]*q[3] ;
    /* Diagonale */
	A[0] = 2.*(q02 + q12) - 1. ;
	A[4] = 2.*(q02 + q22) - 1. ;
	A[8] = 2.*(q02 + q32) - 1. ;

	A[1] = 2.*(q1q2 + q0q3)  ;
	A[3] = 2.*(q1q2 - q0q3)  ;

	A[2] = 2.*(q1q3 - q0q2)  ;
	A[6] = 2.*(q1q3 + q0q2)  ;

	A[5] = 2*(q2q3 + q0q1)  ;
	A[7] = 2*(q2q3 - q0q1)  ;
return A ;
}
/**************************************************************************
  Calcul du quaternion q correspondant a la matrice A
  la composante de plus grande valeur absolue est positive.
***************************************************************************/
double *mrot2qn(double *q, double *a)
{
double us4qi ;
int i, j ;

    q[1] =  a[0] - a[4] - a[8] ;
    q[2] = -a[0] + a[4] - a[8] ;
    q[3] = -a[0] - a[4] + a[8] ;
    q[0] =  a[0] + a[4] + a[8] ;

    i = 0 ;
    for (j = 1 ; j <= 3 ; j++)  if ( q[j] > q[i] ) i = j ;

    switch (i)
    {
    case 0 : /* q0 est le plus grand */
	q[0] = sqrt(q[0]+1.)/2. ;
	us4qi = 0.25/q[0] ;
	q[1] = us4qi*( a[5] - a[7] ) ;
	q[2] = us4qi*( a[6] - a[2] ) ;
	q[3] = us4qi*( a[1] - a[3] ) ;
	break ;
    case 1 : /*  q1 est le plus grand */
	q[1] = sqrt(q[1]+1.)/2. ;
	us4qi = 0.25/q[1] ;
	q[0] = us4qi*( a[5] - a[7]) ;
	q[2] = us4qi*( a[1] + a[3]) ;
	q[3] = us4qi*( a[6] + a[2]) ;
	break ;
    case 2 : /*  q2 est le plus grand */
	q[2] = sqrt(q[2]+1.)/2. ;
	us4qi = 0.25/q[2] ;
	q[0] = us4qi*( a[6] - a[2]) ;
	q[3] = us4qi*( a[5] + a[7]) ;
	q[1] = us4qi*( a[1] + a[3]) ;
	break ;
    case 3 : /* q3 est le plus grand */
	q[3] = sqrt(q[3]+1.)/2. ;
	us4qi = 0.25/q[3] ;
	q[0] = us4qi*( a[1] - a[3]) ;
	q[1] = us4qi*( a[6] + a[2]) ;
	q[2] = us4qi*( a[5] + a[7]) ;
	break ;
    }
return q ;
}

/***************************************************************************
  Vecteur rotation associ au quaternion unitaire q
***************************************************************************/
double *qnu2v3 (double *v, double *q)
{
double coef, sas2 = sqrt(q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);

	if (sas2 != 0.) coef = 2.*asin(sas2)/sas2;
	else coef = 2.;
	v[0] = q[1]*coef;
	v[1] = q[2]*coef;
	v[2] = q[3]*coef;
	return v;
}

/***************************************************************************
  Vecteur rotation et angle associ au quaternion unitaire q
***************************************************************************/
double qnu2av3 (double *v, double *q)
{
double ang, coef, sas2 = sqrt(q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);

	ang = 2.*asin(sas2);
	if (sas2 != 0.) coef = ang/sas2;
	else coef = 2.;
	v[0] = q[1]*coef;
	v[1] = q[2]*coef;
	v[2] = q[3]*coef;
	return ang;
}
/***************************************************************************
   Fournit la norme du quaternion q
****************************************************************************/
double carreqn(double *q)
{
    return (q[1]*q[1] + q[2]*q[2] + q[3]*q[3] + q[0]*q[0]) ;
}
/***************************************************************************
   Fournit la longueur du quaternion q
****************************************************************************/
double longqn(double *q)
{
    return (sqrt(q[1]*q[1] + q[2]*q[2] + q[3]*q[3] + q[0]*q[0])) ;
}
/***************************************************************************
  Normalisation a l'unite d'un quaternion : Renvoi la longueur
****************************************************************************/
double lunitqn(double *q)
{
double s ;
	s = longqn(q) ;
	if(s == 0.)
	{
	   q[0]=1. ;
	   q[1]=0. ;
	   q[2]=0. ;
	   q[3]=0. ;
	}
	else
	{
	   q[0]=q[0]/s ;
	   q[1]=q[1]/s ;
	   q[2]=q[2]/s ;
	   q[3]=q[3]/s ;
	}
	return (s) ;
}
/***************************************************************************
  Normalisation a l'unite d'un quaternion : Renvoi le quaternion
****************************************************************************/
double *unitqn(double *q)
{
double s ;
	s = longqn(q) ;
	if(s == 0.)
	{
	   q[0]=1. ;
	   q[1]=0. ;
	   q[2]=0. ;
	   q[3]=0. ;
	}
	else
	{
	   q[0]=q[0]/s ;
	   q[1]=q[1]/s ;
	   q[2]=q[2]/s ;
	   q[3]=q[3]/s ;
	}
	return q ;
}
/***************************************************************************
   Fournit la valeur absolue de l'angle de rotation associe a l'ecart
  d'orientation entre les attitudes definies par les 2 quaternions unitaires
****************************************************************************/
double eraqnus(double *qa,double *qb)
{
double qe[4],qt[4],er ;

	qt[0] = qa[0] ;
	qt[1] = - qa[1] ;
	qt[2] = - qa[2] ;
	qt[3] = - qa[3] ;
	qnqn(qe,qb,qt) ;
	er = asin(sqrt(qe[1]*qe[1] + qe[2]*qe[2] + qe[3]*qe[3])) ;
	if(qe[0] < 0.) er = M_PI - er ;
	er = 2.*er ;
	if(er > M_PI) er = DEUXPI - er ;
	return (er) ;
}
/***************************************************************************
  Duplication de quaternion : qs = qe
****************************************************************************/
double *cpyqn(double *qs, double *qe)
{
	qs[0]=qe[0] ;
	qs[1]=qe[1] ;
	qs[2]=qe[2] ;
	qs[3]=qe[3] ;
return qs ;
}
/***************************************************************************
  Calcul de la variation du quaternion par la methode d' Edward2

  Entrees: some[3]     integrale des p,q,r (ds sat.) sur un pas.
	   asom[3]     valeur de somme au pas predcedent.

  Sortie : deltaq[4]   Variation du quaternion
****************************************************************************/
double *dedw(double *deltaq, double *some, double *asom)
{
double swx,t0,t02,t0s2 ;

	 t02 = some[1]*some[1] + some[2]*some[2] + some[0]*some[0] ;
	 t0 = sqrt(t02) ;
	 t0s2 = .5*t0 ;
	 deltaq[0] = cos(t0s2) ;
	 if(t0 > 1.e-5) swx = sin(t0s2)/t0 ;
	 else swx = 0.5 ;
	 deltaq[1] = swx*some[0]+(asom[1]*some[2]-asom[2]*some[1])/24. ;
	 deltaq[2] = swx*some[1]+(asom[2]*some[0]-asom[0]*some[2])/24. ;
	 deltaq[3] = swx*some[2]+(asom[0]*some[1]-asom[1]*some[0])/24. ;
return deltaq ;
}

/***************************************************************************
  Quaternion de rotation autour de Ox
****************************************************************************/
double *qrotx(double *q, double ang) 
{
    ang /= 2. ;
	q[0] = cos(ang) ;
	q[1] = sin(ang) ;
	q[2] = q[3] = 0. ;
return q ;
}
/***************************************************************************
  Quaternion de rotation autour de Oy
****************************************************************************/
double *qroty(double *q, double ang) 
{
    ang /= 2. ;
	q[0] = cos(ang) ;
	q[2] = sin(ang) ;
	q[1] = q[3] = 0. ;
return q ;
}

/***************************************************************************
  Quaternion de rotation autour de Oz
****************************************************************************/
double *qrotz(double *q, double ang) 
{
    ang /= 2. ;
	q[0] = cos(ang) ;
	q[3] = sin(ang) ;
	q[1] = q[2] = 0. ;
return q ;
}

static double *crossv3(double* a, double* b, double* c)
{
double p, q ;
      p  = b[1] * c[2] - b[2] * c[1];
      q  = b[2] * c[0] - b[0] * c[2];
    a[2] = b[0] * c[1] - b[1] * c[0];
    a[0] = p ;
    a[1] = q ;
	return a;
}

/***************************************************************************
  Rotation d'un vecteur par un quaternion unitaire
****************************************************************************/
double *qnuv3tqnu(double *a, double *q, double *b)
{
	crossv3(a, q+1, b);
	a[0] += q[0]*b[0];
	a[1] += q[0]*b[1];
	a[2] += q[0]*b[2];
	crossv3(a, q+1, a);
	a[0] = b[0]+2.*a[0];
	a[1] = b[1]+2.*a[1];
	a[2] = b[2]+2.*a[2];
	return a;
}
