#include "racines.h"
#include "deferreur.h"

#include <math.h>
#include <stdlib.h>
/*--------------------------------------------------------------------------*/
/*      CALCUL DE TOUTES LES RACINES D'UN POLYNOME : REELLES ET CPX         */
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/* Degre a partir duquel les procedures roots_hqr et roots balhqr calculent */
/* les racines par iteration plutot que par les formules analytiques        */
/* Ex: Si kRootsIter = 2, les iterations sont utilisees meme pour le degre 2*/
/*--------------------------------------------------------------------------*/
static int kRootsIter = 4;      /* On n'utilise pas les formules de Ferrari */
#include "roots234.c"
/*--------------------------------------------------------------------------*/
/* Calcul racines polynomes par HQR avec balancement prealable              */
/*--------------------------------------------------------------------------*/
int roots_balhqr (double *RacR, double *RacI, double *Coef, int M)
/*--------------------------------------------------------------------------*/
/* Calcul de toutes les racines (partie reelle RacR et imaginaire RacI)     */
/* d'un polynome de degre M  a coefficients reels.                          */
/* Les coefficientss sont donnes avec les degres dans l'ordre croissant.    */
/* RETOUR : Nb de racines calculees                                         */
/*--------------------------------------------------------------------------*/
{
void balanc (int N, double *A, int *low, int *igh, double *scale);
void elmhes (int N, int LOW, int IGH, double *A, int *INT);
int hqr(int N, int LOW, int IGH, double *H, double *wr, double *wi);
int K, I, J, low, igh, *ibid ;
double CM, *A, *CMemo, *vbid ; 
/* double A[DIMENSION_MAX*DIMENSION_MAX], CMemo[DIMENSION_MAX+1], vbid[DIMENSION_MAX+1] ;
int ibid[DIMENSION_MAX+1] ; */

    K = 0 ;                                  /* Nombre de racines calculees */
    /* Teste degre */
    if (M < 1) return K ;
    /* Retire termes de haut degre inutiles */
    while (M > 0 && Coef[M]+Coef[M-1] == Coef[M-1]) M-- ;

	A = (double *) malloc((M*M+M+M+2)*sizeof(double)+(M+1)*sizeof(int));
	if (A == NULL) {perreur("Echec allocation dans roots_balhqr"); return -1 ;}
	CMemo = A + M*M;
	vbid = CMemo+M+1;
	ibid = (int *) (vbid+M+1);

    /* Copie de travail des coefficients */
    for (I = 0 ; I <= M ; I++) CMemo[I] = Coef[I] ;
    /* Collecte racines nulles */
    while (M >= 0 && CMemo[0] == 0.)
    {
      RacR[K] = 0. ; RacI[K++] = 0.;
      for (I = 1 ; I <= M ; I++) CMemo[I-1] = CMemo[I] ;
      M-- ;
    }

    /* Cas racines rationnelles */
	if (M < kRootsIter || M < 5)
	{
		switch (M) {
		case 1 : RacI[K] = 0 ; RacR[K] = - CMemo[0]/CMemo[1] ; K++ ; break;
		case 2 : racpol2 (RacR+K, RacI+K, CMemo) ; K+=2 ; break;
		case 3 : racpol3 (RacR+K, RacI+K, CMemo) ; K+=3 ; break;
		case 4 : racpol4 (RacR+K, RacI+K, CMemo) ; K+=4 ; break; 
		}
		free(A); 
		return K ;
	}

    CM = CMemo[M] ;
    /* Transposee de la forme compagne verticale */
    for (J=0; J<M*M;J++) A[J] = 0.;
    for (I = 1, J = 0 ; I < M ; J += M, I++)
    { A[J] = -CMemo[M-I]/CM ; A[J+I] = 1. ; }
    A[J] = -CMemo[0]/CM ;
    /* Valeurs propres de la matrice de Hessemberg */
    balanc (M, A, &low, &igh, vbid) ;
    elmhes (M, low, igh, A, ibid) ;
    K += M - hqr (M, low, igh, A, RacR+K, RacI+K) ;
    free (A);
    return K ;
}

/*--------------------------------------------------------------------------*/
/*  Calcul racines polynomes par HQR sans balancement                       */
/*--------------------------------------------------------------------------*/
int roots_hqr (double *RacR, double *RacI, double *Coef, int M)
/*--------------------------------------------------------------------------*/
/* Calcul de toutes les racines (partie reelle RacR et imaginaire RacI)     */
/* d'un polynome de degre M a coefficients reels.                           */
/* Les coefficientss sont donnes avec les degres dans l'ordre croissant.    */
/* RETOUR : Nb de racines calculees                                         */
/*--------------------------------------------------------------------------*/
{
int hqr(int N, int LOW, int IGH, double *H, double *wr, double *wi);
int K, I, J ;
double *A, *CMemo, CM ;

    K = 0 ;                                  /* Nombre de racines calculees */
    /* Teste degre */
    if (M < 1) return K ;
    /* Retire termes de haut degre inutiles */
    while (M > 0 && Coef[M]+Coef[M-1] == Coef[M-1]) M-- ;

	A = (double *) malloc((M*M+M+1)*sizeof(double));
	if (A == NULL) {perreur("Echec allocation dans roots_hqr"); return -1 ;}
	CMemo = A + M*M;

    /* Copie de travail des coefficients */
    for (I = 0 ; I <= M ; I++) CMemo[I] = Coef[I] ;
    /* Collecte racines nulles */
    while (M >= 0 && CMemo[0] == 0.)
    {
      RacR[K] = 0. ; RacI[K++] = 0.;
      for (I = 1 ; I <= M ; I++) CMemo[I-1] = CMemo[I] ;
      M-- ;
    }
    /* Cas racines rationnelles */
	if (M < kRootsIter || M < 5)
	{
		switch (M) {
		case 1 : RacI[K] = 0 ; RacR[K] = - CMemo[0]/CMemo[1] ; K++ ; break;
		case 2 : racpol2 (RacR+K, RacI+K, CMemo) ; K+=2 ; break;
		case 3 : racpol3 (RacR+K, RacI+K, CMemo) ; K+=3 ; break;
		case 4 : racpol4 (RacR+K, RacI+K, CMemo) ; K+=4 ; break; 
		}
		free(A); 
		return K ;
	}

    CM = CMemo[M] ;
    /* Transposee de la forme compagne verticale */
    for (J=0; J<M*M;J++) A[J] = 0.;
    for (I = 1, J = 0 ; I < M ; J += M, I++)
    { A[J] = -CMemo[M-I]/CM ; A[J+I] = 1. ; }
    A[J] = -CMemo[0]/CM ;
    /* Valeurs propres de la matrice de Hessemberg */
    K += M - hqr (M, 1, M, A, RacR+K, RacI+K) ;
    free (A) ;
    return K ;
}

