/*--------------------------------------------------------------------------*/
/*      CALCUL DES RACINES DES POLYNOMES DE DEGRE 2,3 et 4                  */
/*--------------------------------------------------------------------------*/
#include <math.h>
#define PIS3     1.04719755119659774615
#define DEUXPIS3 2.09439510239319549230
#define SQRT_3   1.7320508075688772935
/* 
Racines d'un polynome de degre 2, coefs ordre degre croissant
renvoi le nombre de racines valides
*/
static int racpol2 (double *pr, double *pi, double *coefs)
{
double delta ;
double div = coefs[2] ;
    
    if (div == 0.)
	{
		if (coefs[1] == 0.) return 0 ;
		else
		{
			pi[0] = 0.;
			pr[0] = -coefs[0]/coefs[1] ;
			return 1 ;
		}
	}
    pr[0] = pr[1] = - 0.5*coefs[1]/div ;
    delta = pr[0]*pr[0] - coefs[0]/div ;
    if (delta < 0.)
    {  delta = sqrt(-delta) ; pi[0] = delta ; pi[1] = -delta ; }
    else
    {
       delta = sqrt(delta) ; pi[0] = pi[1] = 0. ;
       pr[0] += delta ; pr[1] -= delta ;
    }
	return 2 ;
}

/* 
   Racines d'un polynome de degre 3, coefs ordre degre croissant
   renvoi le nombre de racines valides
*/
static int racpol3 (double *pr, double *pi, double *coefs)
{
double P, Q, R, ps3, p2s3, a, b, as3, bs2, a3s27, r1 ;
double div = coefs[3] ;

   if (div == 0.) return racpol2 (pr, pi, coefs) ;
   P = coefs[2]/div ;
   Q = coefs[1]/div ;
   R = coefs[0]/div ;
   ps3 = P / 3. ;
   p2s3 = P * ps3 ;
   a = Q - p2s3 ;
   b = P * p2s3 / 4.5 - Q * ps3 + R ;
   bs2 = b / 2. ; as3 = a / 3. ;
   a3s27 = as3 * as3 * as3 ;
   r1 = bs2 * bs2 + a3s27 ;
   pi[0] = 0. ;
   if (r1 > 0.)
   {
   double ga, gb, apb ;
	ga = -b / 2. + sqrt (r1) ;
	if (ga >= 0.) ga = pow (ga, 1./3.) ;
	else ga = - pow (-ga, 1./3.) ;
	gb = -b / 2. - sqrt (r1) ;
	if (gb >= 0.) gb = pow (gb, 1./3.) ;
	else gb = - pow (-gb, 1./3.) ;
	/* gb = - a / (3. * ga) ; */
	apb = ga + gb ;
	pr[0] = apb - ps3 ;
	pr[2] = pr[1] = - apb/2. - ps3 ;
	pi[1] = SQRT_3 * (ga - gb) / 2. ;
	pi[2] = - pi[1] ;
    }
    else
    {
	pi[2] = pi[1] = 0. ;
	if (as3 != 0.)
	{
	double cph, phis3 ;
	    r1 = 2.*sqrt ( -as3) ;
	    cph = 3.*b/(a*r1) ;
	    if (cph >= 1.) phis3 = 0. ;
	    else if (cph <= -1.) phis3 = PIS3 ;
	    else phis3 = acos (3.*b/(a*r1)) / 3. ;
	    pr[0] = r1 * cos (phis3) ;
	    pr[1] = r1 * cos (phis3 + DEUXPIS3) ;
	    pr[2] = -(pr[0] + pr[1]) - ps3 ;
	    pr[0] -= ps3 ;
	    pr[1] -= ps3 ;
	}
	else { pr[2] = pr[1] = pr[0] = - ps3 ; }
    }
	return 3 ;
}


/* Calcul de l'unique racine relle ou de la plus grande de X3+PX2+QX+R=0 */
static double ur3 (double P, double Q, double R)         /* pour Ferrari */
{
double ps3, p2s3 ;
double a, b, as3, bs2, a3s27, r1 ;

    ps3 = P / 3. ;
    p2s3 = P * ps3 ;
    a = Q - p2s3 ;
    b = P * p2s3 / 4.5 - Q * ps3 + R ;
    bs2 = b / 2. ; as3 = a / 3. ;
    a3s27 = as3 * as3 * as3 ;
    r1 = bs2 * bs2 + a3s27 ;
    if (r1 > 0.)
    {
    double ga, gb ;

	ga = -b / 2. + sqrt (r1) ;
	if (ga >= 0.) ga = pow (ga, 1./3.) ;
	else ga = - pow (-ga, 1./3.) ;
	gb = -b / 2. - sqrt (r1) ;
	if (gb >= 0.) gb = pow (gb, 1./3.) ;
	else gb = - pow (-gb, 1./3.) ;
	/* gb = - a / (3. * ga) ; */
	return (ga + gb - ps3) ;
    }
    else
    {
	if (as3 != 0.)
	{
	double cph, phis3 ;
	    r1 = 2.*sqrt ( -as3) ;
	    cph = 3.*b/(a*r1) ;
	    if (cph >= 1.) phis3 = 0. ;
	    else if (cph <= -1.) phis3 = PIS3 ;
	    else phis3 = acos (3.*b/(a*r1)) / 3. ;
	    return (r1 * cos (phis3) - ps3) ;
	}
	else return (-ps3) ;
    }
}

/* 
	Racines d'un polynome de degre 4, coefs ordre degre croissant 
    Renvoi le nombre de racines valides
*/
static int racpol4 (double *pr, double *pi, double *coefs)
{
double a, b, c, d ;
double as4, a2s4, R, R2, S, S2, RS2, X, Y, D2, E2, DX, DY, EX, EY ;
int fdr, fer ;
double div = coefs[4] ;

    if (div == 0.) return racpol3 (pr, pi, coefs) ;
    a = coefs[3]/div ;
    b = coefs[2]/div ;
    c = coefs[1]/div ;
    d = coefs[0]/div ;
    as4 = a / 4. ;
    a2s4 = as4*a ;
    Y = ur3 (-b , a*c-4.*d, d*(4.*b - a*a) - c*c) ;
    R2 = a2s4 + Y - b ;
    S2 = Y*Y/4. - d ;
    if (R2 > 0.) R = sqrt(R2) ; else R = R2 = 0. ;
    if (S2 > 0.) S = sqrt(S2) ; else S = S2 = 0. ;
    if (a*Y < 2.*c) S = -S ;
    X = a2s4 + R2 - 2.*Y ;
    Y = a * R - 4.* S;
    D2 = X+Y ;
    if (D2 < 0.) { fdr = 0 ; DX = 0. ; DY = sqrt (-D2) ; }
    else         { fdr = 1 ; DX = sqrt ( D2) ; DY = 0. ; }
    E2 = X-Y ;
    if (E2 < 0.) { fer = 0 ; EX = 0. ; EY = sqrt (-E2) ; }
    else         { fer = 1 ; EX = sqrt ( E2) ; EY = 0. ; }
    DX /= 2. ; DY /= 2. ; EX /= 2. ; EY /= 2. ; RS2 = R/2. ;
    pr[0] = -as4 - RS2 ;
    pr[2] = -as4 + RS2 ;
    if (fdr) { pi[0] = pi[1] = 0. ; pr[1] = pr[0] - DX ; pr[0] += DX ; }
    else     { pi[0] = DY ; pi[1] = -DY ; pr[1] = pr[0] ; }
    if (fer) { pi[2] = pi[3] = 0. ; pr[3] = pr[2] - EX ; pr[2] += EX ; }
    else     { pi[2] = EY ; pi[3] = -EY ; pr[3] = pr[2] ; }
	return 4 ;
}
