#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <float.h>
#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif
#include "sp3.h"

static int printformat=1113;
static char strfmt[10]="%11.3lg";
/**********************************************************************/
/*           UTILITAIRES MATRICIELS DIMENSION 3                       */
/**********************************************************************/
/*---------------------------------------------------------------------
	Impression d'un vecteur de R3
	-----------------------------

	Entres :
		char *s		: Texte  imprimer (pas de Rc ajout  la fin).
		double *a	: Adresse du v3 imprim sur la ligne
----------------------------------------------------------------------*/
void printv3(char *s, double *a)
{
int i;

	sprintf(strfmt+1,"%d.%dl%c",(printformat/10)%100,(printformat%10),
		printformat/1000 ? 'g' : 'f');

	printf("%s",s) ;
	for (i = 0 ; i < 3 ; i++) printf(strfmt,a[i]) ;
	printf("\n");
}

/*---------------------------------------------------------------------
	Affection d'un scalaire aux composantes d'un vecteur de R3
	----------------------------------------------------------

	Entrees:
		double *u	: Adresse du v3 affect
		double s    : Scalaire  affecter

	Sorties:
		u est affect
		Retour		: Adresse du v3 affect

----------------------------------------------------------------------*/
double *setv3(double* u, double s)
{
    u[0] = s ;
    u[1] = s ;
    u[2] = s ;
	return u;
}

/*---------------------------------------------------------------------
	Affection des composantes  un vecteur de R3
	--------------------------------------------

	Entrees:
		double *u	: Adresse du v3 affect
		double x,y,z: Composantes  affecter

	Sorties:
		u est affect
		Retour		: Adresse du v3 affect

----------------------------------------------------------------------*/
double *iniv3(double *v, double x, double y, double z)
{ v[0] = x; v[1] = y; v[2] = z; return v;}

/*---------------------------------------------------------------------
	Recopie d'un vecteur de R3 : U = V
	----------------------------------

	Entrees:
		double *u	: Adresse du v3 affect
		double *v	: Adresse vecteur donn

	Sorties:
		u est affect
		Retour		: Adresse du v3 affect
----------------------------------------------------------------------*/
double *cpyv3(double* u, double* v)
{
    u[0] = v[0] ;
    u[1] = v[1] ;
    u[2] = v[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Ngation d'un vecteurs de R3 : U = -V
	-------------------------------------

	Entrees:
		double *u	: Adresse du v3 affect
		double *v	: Adresse vecteur donn

	Sorties:
		u est affect
		Retour		: Adresse du v3 affect

    Remarque : u et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *negv3(double* u, double* v)
{
    u[0] = -v[0] ;
    u[1] = -v[1] ;
    u[2] = -v[2] ;
	return u;
}
/*---------------------------------------------------------------------
    Valeur absolue d'un vecteurs de R3 : U = |V|
	--------------------------------------------

	Entrees:
		double *u	: Adresse du v3 affect
		double *v	: Adresse vecteur donn

	Sorties:
		u est affect
		Retour		: Adresse du v3 affect

    Remarque : u et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *absv3(double* u, double* v)
{
    u[0] = fabs(v[0]) ;
    u[1] = fabs(v[1]) ;
    u[2] = fabs(v[2]) ;
	return u;
}
/*---------------------------------------------------------------------
	Composante maximale d'un vecteur de R3 : s = max(Ui)
	----------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn

	Sorties:
		Retour		: Composante maximale du vecteur
----------------------------------------------------------------------*/
double xmaxv3(double* u)
{
	double s = u[0];
	if (u[1]>s)s=u[1];
	if (u[2]>s)s=u[2];
	return s;
}

/*---------------------------------------------------------------------
	Composante max. en val. abs. d'un vecteur de R3 : s = max(|Ui|)
	---------------------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn

	Sorties:
		Retour		: Composante maximale du vecteur, en val. abs.
----------------------------------------------------------------------*/
double xamaxv3(double* u)
{
	double s = fabs(u[0]);
	if (fabs(u[1])>s)s=fabs(u[1]);
	if (fabs(u[2])>s)s=fabs(u[2]);
	return s;
}
/*---------------------------------------------------------------------
	Index de la composante max d'un vecteur de R3
	---------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn

	Sorties:
		Retour		: index de la composante maximale du vecteur
----------------------------------------------------------------------*/
int imaxv3(double* u)
{
	int i = 0 ;
	double s = u[0];
	if (u[1]>s) { s=u[1]; i = 1;}
	if (u[2]>s) i = 2;
	return i;
}

/*---------------------------------------------------------------------
	Index de la composante max en v.abs. d'un vecteur de R3
	-------------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn

	Sorties:
		Retour		: index de la composante maximale en v.a.
----------------------------------------------------------------------*/
int iamaxv3(double* u)
{
	int i = 0 ;
	double s = fabs(u[0]);
	if (fabs(u[1])>s) { s=fabs(u[1]); i = 1;}
	if (fabs(u[2])>s) i = 2;
	return i;
}

/*---------------------------------------------------------------------
	Composante minimale d'un vecteurs de R3 : s = min(Ui)
	-----------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn

	Sorties:
		Retour		: Composante minimale du vecteur
----------------------------------------------------------------------*/
double xminv3(double* u)
{
	double s = u[0];
	if (u[1]<s)s=u[1];
	if (u[2]<s)s=u[2];
	return s;
}

/*---------------------------------------------------------------------
	Composante min. en val. abs. d'un vecteur de R3 : s = min(|Ui|)
	---------------------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn

	Sorties:
		Retour		: Composante minimale du vecteur, en val. abs.
----------------------------------------------------------------------*/
double xaminv3(double* u)
{
	double s = fabs(u[0]);
	if (fabs(u[1])<s)s=fabs(u[1]);
	if (fabs(u[2])<s)s=fabs(u[2]);
	return s;
}

/*---------------------------------------------------------------------
	Index de la composante min d'un vecteur de R3
	---------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn

	Sorties:
		Retour		: index de la composante maximale du vecteur
----------------------------------------------------------------------*/
int iminv3(double* u)
{
	int i = 0 ;
	double s = u[0];
	if (u[1]<s) { s=u[1]; i = 1;}
	if (u[2]<s) i = 2;
	return i;
}

/*---------------------------------------------------------------------
	Index de la composante min en v.abs. d'un vecteur de R3
	-------------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn

	Sorties:
		Retour		: index de la composante maximale en v.a.
----------------------------------------------------------------------*/
int iaminv3(double* u)
{
	int i = 0 ;
	double s = fabs(u[0]);
	if (fabs(u[1])<s) { s=fabs(u[1]); i = 1;}
	if (fabs(u[2])<s) i = 2;
	return i;
}


/*---------------------------------------------------------------------
	Multiplication d'un vecteur de R3 par un scalaire : U = sV
	----------------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du vecteur de R3 donn
		double s	: Scalaire donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V
----------------------------------------------------------------------*/
double *scav3(double* u, double s, double* v)
{
    u[0] = s * v[0] ;
    u[1] = s * v[1] ;
    u[2] = s * v[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Addition de vecteurs de R3 : U = V + W
	--------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *addv3(double* u, double* v, double* w)
{
    u[0] = v[0] + w[0] ;
    u[1] = v[1] + w[1] ;
    u[2] = v[2] + w[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Soustraction de vecteurs de R3 : U = V - W
	------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *subv3(double* u, double* v, double* w)
{
    u[0] = v[0] - w[0] ;
    u[1] = v[1] - w[1] ;
    u[2] = v[2] - w[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Multiplication terme a terme de vecteurs de R3 : U = V x W
	----------------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *mulv3(double* u, double* v, double* w)
{
    u[0] = v[0] * w[0] ;
    u[1] = v[1] * w[1] ;
    u[2] = v[2] * w[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Division terme a terme de vecteurs de R3 : U = V / W
	----------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect
----------------------------------------------------------------------*/
double *divv3(double* u, double* v, double* w)
{
    u[0] = v[0] / w[0] ;
    u[1] = v[1] / w[1] ;
    u[2] = v[2] / w[2] ;
	return u;
}
/*---------------------------------------------------------------------
	Cumulation de vecteurs de R3 : U = V + aW
	-----------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double a	: Scalaire donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *cumv3(double* u, double* v,double a, double* w)
{
    u[0] = v[0] + a*w[0] ;
    u[1] = v[1] + a*w[1] ;
    u[2] = v[2] + a*w[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Produit scalaire de vecteurs de R3 : s = <A . B>
	------------------------------------------------

	Entrees:
		double *A	: Adresse du premier vecteur de R3 donn
		double *B	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		Retour		: produit scalaire
----------------------------------------------------------------------*/
double dotv3(double* a, double* b)
{
    return(a[0]*b[0]+a[1]*b[1]+a[2]*b[2]);
}

/*---------------------------------------------------------------------
	Produit tensoriel de vecteurs de R3 : M = V1 * V2'
	--------------------------------------------------

	Entrees:
		double *M	: Adresse de la matrice de R3 affecte
		double *V1	: Adresse du premier vecteur de R3 donn
		double *V2	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		M est affecte
		Retour		: Adresse de la matrice de R3 affecte
----------------------------------------------------------------------*/
double *tensv3 (double *mato, double *v1, double *v2)
{
    mato[0] = v1[0] * v2[0] ;
    mato[1] = v1[1] * v2[0] ;
    mato[2] = v1[2] * v2[0] ;
    mato[3] = mato[1] ;
    mato[4] = v1[1] * v2[1] ;
    mato[5] = v1[2] * v2[1] ;
    mato[6] = mato[2] ;
    mato[7] = mato[5] ;
    mato[8] = v1[2] * v2[2] ;
	return mato;
}
/*---------------------------------------------------------------------
	Produit vectoriel de vecteurs de R3 : A = B ^ C
	-----------------------------------------------

	Entrees:
		double *A	: Adresse du vecteur affect
		double *B	: Adresse du premier vecteur de R3 donn
		double *C	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		A est affect
		Retour		: Adresse du vecteur affect

	Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *crossv3(double* a, double* b, double* c)
{
double p, q ;
      p  = b[1] * c[2] - b[2] * c[1];
      q  = b[2] * c[0] - b[0] * c[2];
    a[2] = b[0] * c[1] - b[1] * c[0];
    a[0] = p ;
    a[1] = q ;
	return a;
}

/*---------------------------------------------------------------------
	Somme des composantes d'un vecteurs de R3 : s = Sigma Ui
	--------------------------------------------------------

	Entrees:
		double *A	: Adresse vecteur donn

	Sorties:
		Retour		: Somme des composantes
----------------------------------------------------------------------*/
double xsumv3(double* a)
{
    return(a[0]+a[1]+a[2]);
}

/*---------------------------------------------------------------------
	Carr d'un vecteurs de R3 : s = A'*A
	------------------------------------

	Entrees:
		double *A	: Adresse vecteur donn

	Sorties:
		Retour		: Somme des carrs des composantes
----------------------------------------------------------------------*/
double carrev3(double* a)
{
    return(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]);
}

/*---------------------------------------------------------------------
	Longueur d'un vecteurs de R3 : s = sqrt(A'*A)
	---------------------------------------------

	Entrees:
		double *A	: Adresse vecteur donn

	Sorties:
		Retour		: Longueur du vecteur
----------------------------------------------------------------------*/
double longv3(double* a)
{
    return(sqrt(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]));
}

/*---------------------------------------------------------------------
	Normalisation d'un vecteur de R3 : A = B / ||B||
	------------------------------------------------

	Entrees:
		double *A	: Adresse vecteur unitaire rsultat
		double *B	: Adresse vecteur donn

	Sorties:
		A est affect
		Retour		: Adresse vecteur unitaire rsultat

	Remarque : A et B peuvent occuper la mme mmoire
		       si ||B|| = 0  -> A[0] = 1, les autres nuls.

	Remarque : A peut occuper la mme mmoire que B
----------------------------------------------------------------------*/
double *unitv3(double* a, double* b)
{
    double l ;
    l = sqrt(b[0]*b[0]+b[1]*b[1]+b[2]*b[2]);
    if (l == 0.)
    {
        a[0]=1.; a[1]=0.; a[2]=0.;
    }
    else
    {
        a[0] = b[0]/l;
		a[1] = b[1]/l;
        a[2] = b[2]/l;
    }
    return a;
}

/*---------------------------------------------------------------------
	Normalisation d'un vecteur de R3 : A = B / ||B||
	------------------------------------------------

	Entrees:
		double *A	: Adresse vecteur unitaire rsultat
		double *B	: Adresse vecteur donn

	Sorties:
		A est affect
		Retour		: Longueur du vecteur B

	Remarque : A et B peuvent occuper la mme mmoire
		       si ||B|| = 0  -> A[0] = 1, les autres nuls.

	Remarque : A peut occuper la mme mmoire que B
----------------------------------------------------------------------*/
double lunitv3(double* a, double* b)
{
    double l ;
    l = sqrt(b[0]*b[0]+b[1]*b[1]+b[2]*b[2]);
    if (l == 0.)
    {
        a[0]=1.; a[1]=0.; a[2]=0.;
    }
    else
    {
        a[0] = b[0]/l;
		a[1] = b[1]/l;
        a[2] = b[2]/l;
    }
    return(l);
}

/*---------------------------------------------------------------------
	Affection d'un scalaire aux composantes d'une matrice de R3
	-----------------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice m3 affecte
		double s    : Scalaire  affecter

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice m3 affecte

----------------------------------------------------------------------*/
double *setm3(double* A, double s)
{
int i ; for(i=0;i<9;i++) A[i]=s;
	return A;
}

/*---------------------------------------------------------------------
	Recopie d'une matrice de R3 : A = B
	-----------------------------------

	Entrees:
		double *A	: Adresse de la matrice m3 affecte
		double *B	: Adresse de la matrice m3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice m3 affecte
----------------------------------------------------------------------*/
double *cpym3(double* A, double* B)
{
int i ; for(i=0;i<9;i++) A[i]=B[i];
	return A;
}

/*---------------------------------------------------------------------
	Ngation d'une matrice de R3 : A = - B
	--------------------------------------

	Entrees:
		double *A	: Adresse de la matrice m3 affecte
		double *B	: Adresse de la matrice m3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice m3 affecte

    Remarque : A et B peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *negm3(double* A, double* B)
{
int i ; for(i=0;i<9;i++) A[i]=-B[i];
	return A;
}

/*---------------------------------------------------------------------
    Multiplication d'une matrice de R3 par un scalaire : A = sB
	-----------------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice m3 affecte
		double *B	: Adresse de la matrice m3 donne
		double s	: Scalaire donn

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice m3 affecte

    Remarque : A et B peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *scam3(double* A, double s, double* B)
{
int i ; for(i=0;i<9;i++) A[i]=s*B[i];
	return A;
}

/*---------------------------------------------------------------------
	Addition de matrices de R3 : A = B + C
	--------------------------------------

	Entrees:
		double *A	: Adresse de la matrice m3 affecte
		double *B	: Adresse de la premire matrice m3 donne
		double *C	: Adresse de la deuxime matrice m3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice m3 affecte

    Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *addm3(double* A, double* B, double *C)
{
int i ; for(i=0;i<9;i++) A[i]=B[i]+C[i];
	return A;
}

/*---------------------------------------------------------------------
Soustraction de matrices de R3 : A = B - C
------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice m3 affecte
		double *B	: Adresse de la premire matrice m3 donne
		double *C	: Adresse de la deuxime matrice m3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice m3 affecte

	Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *subm3(double* A, double* B, double *C)
{
int i ; for(i=0;i<9;i++) A[i]=B[i]-C[i];
	return A;
}
/*---------------------------------------------------------------------
	Cumulation de matrices de R3 : A = B + sC
	-----------------------------------------

	Entrees:
		double *A	: Adresse de la matrice m3 affecte
		double *B	: Adresse de la premire matrice m3 donne
		double s	: Scalaire donn
		double *C	: Adresse de la deuxime matrice m3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice m3 affecte

	Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *cumm3(double* A, double* B, double s, double *C)
{
int i ; for(i=0;i<9;i++) A[i]=B[i]+s*C[i];
	return A;
}

/*---------------------------------------------------------------------
	Matrice Identite de R3 :  A = I
	-------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte

    Sorties:
		A est affecte
		Retour		: Adresse de A
----------------------------------------------------------------------*/
double *eyem3(double *a)
{
    a[0] = a[4] = a[8] = 1. ;
    a[1] = a[2] = a[3] = 0. ;
    a[5] = a[6] = a[7] = 0. ;
	return a ;
}

/*---------------------------------------------------------------------
	Gnration d'une matrice diagonale  partir d'un vecteur de R3
	--------------------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double *V	: Adresse du vecteur donn

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice affecte
----------------------------------------------------------------------*/
double *diag2m3(double *a, double *v)
{
    a[0] = v[0]; a[3] = 0.  ; a[6] = 0.  ;
    a[1] = 0.  ; a[4] = v[1]; a[7] = 0.  ;
    a[2] = 0.  ; a[5] = 0.  ; a[8] = v[2];
	return a ;
}

/*---------------------------------------------------------------------
	Transpose d'une matrice de R3 : A = A'
	--------------------------------------

	Entrees:
		double *A	: Adresse de la matrice

	Sorties:
		A est transpose
		Retour		: Adresse de A
----------------------------------------------------------------------*/
double *tm3(double *a)
{
double s ;
	s = a[3]; a[3] = a[1]; a[1] = s;
	s = a[6]; a[6] = a[2]; a[2] = s;
	s = a[7]; a[7] = a[5]; a[5] = s;
	return a ;
}

/*---------------------------------------------------------------------
	Extraction du vecteur diagonal d'une matrice de R3
	--------------------------------------------------

	Entrees:
		double *V	: Adresse vecteur affect
		double *A	: Adresse de la matrice donne

	Sorties:
		V est affect
		Retour		: Adresse vecteur affect
----------------------------------------------------------------------*/
double *m3todiag(double *v, double *a)
{
    v[0] = a[0];
    v[1] = a[4];
	v[2] = a[8];
	return v ;
}

/*---------------------------------------------------------------------
	Gnration matrice antisymtrique  partir d'un vecteur de R3
	-------------------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double *V	: Adresse vecteur donn

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice affecte
----------------------------------------------------------------------*/
double *v3toasm3(double *a, double *v)
{
    a[0] =   0. ; a[3] = -v[2]; a[6] =  v[1];
    a[1] =  v[2]; a[4] =  0.  ; a[7] = -v[0];
    a[2] = -v[1]; a[5] =  v[0]; a[8] =   0. ;
	return a ;
}

/*---------------------------------------------------------------------
	Extraction du vecteur antisymtrique d'une matrice de R3
	--------------------------------------------------------

	Entrees:
		double *V	: Adresse vecteur affect
		double *A	: Adresse de la matrice donn

	Sorties:
		V est affect
		Retour		: Adresse vecteur affect
----------------------------------------------------------------------*/
double *asm3tov3(double *v, double *a)
{
    v[0] = 0.5*(a[5]-a[7]);
    v[1] = 0.5*(a[6]-a[2]);
	v[2] = 0.5*(a[1]-a[3]);
	return v ;
}

/*---------------------------------------------------------------------
	Produit matrice x vecteurs de R3 : w = A v
	-----------------------------------------------

	Entrees:
		double *w	: Adresse du vecteur affect
		double *A	: Adresse de la matrice de R3 donn
		double *v	: Adresse du vecteur de R3 donn

	Sorties:
		w est affect
		Retour		: Adresse du vecteur affect

    Remarque : w et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *m3v3(double* w, double* a, double* v)
{
double p, q ;
      p  = a[0] * v[0] + a[3] * v[1] + a[6] * v[2];
      q  = a[1] * v[0] + a[4] * v[1] + a[7] * v[2];
    w[2] = a[2] * v[0] + a[5] * v[1] + a[8] * v[2];
    w[0] = p ;
    w[1] = q ;
	return w;
}

/*---------------------------------------------------------------------
	Produit matrice transpose x vecteurs de R3 : w = A' v
	------------------------------------------------------

	Entrees:
		double *w	: Adresse du vecteur affect
		double *A	: Adresse de la matrice de R3 donn
		double *v	: Adresse du vecteur de R3 donn

	Sorties:
		w est affect
		Retour		: Adresse du vecteur affect

    Remarque : w et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *tm3v3(double* w, double* a, double* v)
{
double p, q ;
      p  = a[0] * v[0] + a[1] * v[1] + a[2] * v[2];
      q  = a[3] * v[0] + a[4] * v[1] + a[5] * v[2];
    w[2] = a[6] * v[0] + a[7] * v[1] + a[8] * v[2];
    w[0] = p ;
    w[1] = q ;
	return w;
}

/*---------------------------------------------------------------------
	Produit vecteur transpos x matrice de R3 : a' = b' C
	------------------------------------------------------

	Entrees:
		double *a	: Adresse du vecteur affect
		double *b	: Adresse du vecteur de R3 donn
		double *c	: Adresse de la matrice de R3 donn

	Sorties:
		a est affect
		Retour		: Adresse du vecteur affect

    Remarque : a et b peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *tv3m3( double* a,double* b, double* c)
{
double d,e ;
    d    = b[0] * c[0] + b[1] * c[1] + b[2] * c[2];
    e    = b[0] * c[3] + b[1] * c[4] + b[2] * c[5];
    a[2] = b[0] * c[6] + b[1] * c[7] + b[2] * c[8];
	a[1] = e; a[0] = d;
	return a;
}

/*---------------------------------------------------------------------
	Produit matrice x matrice de R3 : A = B C
	-----------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double *B	: Adresse de la premire matrice de R3 donne
		double *C	: Adresse de la deuxime matrice de R3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice affecte

    Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *m3m3(double* a,double* b, double* c)
{
register int i ;
double d[8] ;
    d[0] = b[0] * c[0] + b[3] * c[1] + b[6] * c[2];
    d[1] = b[1] * c[0] + b[4] * c[1] + b[7] * c[2];
    d[2] = b[2] * c[0] + b[5] * c[1] + b[8] * c[2];

    d[3] = b[0] * c[3] + b[3] * c[4] + b[6] * c[5];
    d[4] = b[1] * c[3] + b[4] * c[4] + b[7] * c[5];
    d[5] = b[2] * c[3] + b[5] * c[4] + b[8] * c[5];

    d[6] = b[0] * c[6] + b[3] * c[7] + b[6] * c[8];
    d[7] = b[1] * c[6] + b[4] * c[7] + b[7] * c[8];
    a[8] = b[2] * c[6] + b[5] * c[7] + b[8] * c[8];
    for(i = 0 ; i < 8 ; ++i) a[i] = d[i] ;
	return a;
}

/*---------------------------------------------------------------------
	Produit matrice x matrice transpose de R3 : A = B C'
	----------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double *B	: Adresse de la premire matrice de R3 donne
		double *C	: Adresse de la deuxime matrice de R3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice affecte

    Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *m3tm3(double* a, double* b, double* c)
{
register int i ;
double d[8] ;
    d[0] = b[0] * c[0] + b[3] * c[3] + b[6] * c[6];
    d[1] = b[1] * c[0] + b[4] * c[3] + b[7] * c[6];
    d[2] = b[2] * c[0] + b[5] * c[3] + b[8] * c[6];

    d[3] = b[0] * c[1] + b[3] * c[4] + b[6] * c[7];
    d[4] = b[1] * c[1] + b[4] * c[4] + b[7] * c[7];
    d[5] = b[2] * c[1] + b[5] * c[4] + b[8] * c[7];

    d[6] = b[0] * c[2] + b[3] * c[5] + b[6] * c[8];
    d[7] = b[1] * c[2] + b[4] * c[5] + b[7] * c[8];
    a[8] = b[2] * c[2] + b[5] * c[5] + b[8] * c[8];
    for(i = 0 ; i < 8 ; ++i) a[i] = d[i] ;
	return a;
}

/*---------------------------------------------------------------------
	Produit matrice transpose x matrice de R3 : A = B'C
	----------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double *B	: Adresse de la premire matrice de R3 donne
		double *C	: Adresse de la deuxime matrice de R3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice affecte

    Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *tm3m3( double* a,double* b, double* c)
{
register int i;
double d[8] ;
    d[0] = b[0] * c[0] + b[1] * c[1] + b[2] * c[2];
    d[1] = b[3] * c[0] + b[4] * c[1] + b[5] * c[2];
    d[2] = b[6] * c[0] + b[7] * c[1] + b[8] * c[2];

    d[3] = b[0] * c[3] + b[1] * c[4] + b[2] * c[5];
    d[4] = b[3] * c[3] + b[4] * c[4] + b[5] * c[5];
    d[5] = b[6] * c[3] + b[7] * c[4] + b[8] * c[5];

    d[6] = b[0] * c[6] + b[1] * c[7] + b[2] * c[8];
    d[7] = b[3] * c[6] + b[4] * c[7] + b[5] * c[8];
    a[8] = b[6] * c[6] + b[7] * c[7] + b[8] * c[8];
    for(i = 0 ; i < 8 ; ++i) a[i] = d[i] ;
	return a;
}

/*---------------------------------------------------------------------
	Trace d'une matrice de R3 : s = Sigma de Aii
	--------------------------------------------

	Entrees:
		double *A	: Adresse matrice donne

	Sorties:
		Retour		: trace de la matrice
----------------------------------------------------------------------*/
double tracem3(double* a)
{
    return(a[0]+a[4]+a[8]);
}

/*---------------------------------------------------------------------
	Dterminant d'une matrice de R3 : s = |A|
	-----------------------------------------

	Entrees:
		double *A	: Adresse matrice donne

	Sorties:
		Retour		: dterminant de la matrice
----------------------------------------------------------------------*/
double detm3(double* a)
{
    return(a[0]*(a[4]*a[8]-a[7]*a[5])
		  +a[3]*(a[7]*a[2]-a[1]*a[8])
		  +a[6]*(a[1]*a[5]-a[4]*a[2]));
}

/*---------------------------------------------------------------------
	Inverse d'une matrice de R3 : inv(A)
	------------------------------------

	Entrees:
		double *A	: Adresse matrice

	Sorties:
	    A = inv(A) si det != 0.
	Retour			: Adresse de A

   ATTENTION : EXTREMEMENT DANGEREUX SI DETERMINANT NUL
   FAIRE PRECEDER DU CALCUL DU DETERMINANT
----------------------------------------------------------------------*/
double *invm3(double* a)
{
double delta ;
double c[9] ;

    crossv3 (c, a+3, a+6) ;
    delta = dotv3(a, c) ;
    if (delta == 0.) return a ;
    crossv3 (c+3, a+6, a) ;
    crossv3 (c+6, a, a+3) ;
    tm3(scam3 (a, 1./delta, c)) ;
    return a ;
}

/*---------------------------------------------------------------------
	Dterminant et Inverse d'une matrice de R3 : s = |A|, inv(A)
	------------------------------------------------------------

	Entrees:
		double *A	: Adresse matrice

	Sorties:
	    A = inv(A) si det != 0.
		Retour		: dterminant de la matrice
----------------------------------------------------------------------*/
double dinvm3(double* a)
{
double delta ;
double c[9] ;

    crossv3 (c, a+3, a+6) ;
    delta = dotv3(a, c) ;
    if (delta == 0.) return delta ;
    crossv3 (c+3, a+6, a) ;
    crossv3 (c+6, a, a+3) ;
    tm3(scam3 (a, 1./delta, c)) ;
    return delta ;
}

/*---------------------------------------------------------------------
	Impression d'une matrice de R3
	------------------------------

	Entrees:
		char *s		: Texte  impimer (RC ajout en fin du texte)
		double *A	: Adresse matrice donne  imprimer
----------------------------------------------------------------------*/
void printm3(char *s, double *a)
{
int i, j ;

	sprintf(strfmt+1,"%d.%dl%c",(printformat/10)%100,(printformat%10),
		printformat/1000 ? 'g' : 'f');
	printf("%s\n",s);
	for (i = 0 ; i < 3 ; i++)
	{
		for(j=i ; j<9 ; j += 3) printf(strfmt,a[j]) ; printf("\n");
	}
}