#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <float.h>
#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif
#include "spn.h"

static int printformat=1113;
static char strfmt[10]="%11.3lg";
/**********************************************************************/
/*           UTILITAIRES MATRICIELS DIMENSION N                       */
/**********************************************************************/
/*---------------------------------------------------------------------
	Impression d'une matrice de Rn
	------------------------------

	Entrees:
		char *s		: Texte  impimer (RC ajout en fin du texte)
		double *A	: Adresse de la matrice donne  imprimer
		int li		: Nombre de lignes
		int co		: Nombre de colonnes
----------------------------------------------------------------------*/
void printmat(char *s, double *a, int li, int co)
{
int i, j, k, n = li*co ;

	sprintf(strfmt+1,"%d.%dl%c",(printformat/10)%100,(printformat%10),
		printformat/1000 ? 'g' : 'f');

	printf("%s\n",s);
	for (i = 0 ; i < li ; i++)
	{
		k = 0 ;
		for(j=i ; j<n ; j += li)
		{
			printf(strfmt,a[j]) ;
			if (++k == 7 && j+li<n) { printf("\n  "); k = 0;}
		}
		printf("\n");
	}
}

/*---------------------------------------------------------------------
	Affectation d'un scalaire aux composantes vecteur de Rn : Ui = s
	----------------------------------------------------------------

	Entrees:
		double *u	: Adresse du vecteur affect
		double s	: scalaire donn
		int n		: dimension de v

	Sorties:
		u est affect
		Retour		: Adresse du vecteur affect
----------------------------------------------------------------------*/
double *setvec (double* u, double s, int n)
{
    if (n > 0) while (n--) u[n] = s;
	return u;
}

/*---------------------------------------------------------------------
	Recopie d'un vecteur de Rn : U = V
	----------------------------------

	Entrees:
		double *u	: Adresse du vecteur affect
		double *v	: Adresse vecteur donn
		int n		: dimension de v

	Sorties:
		u est affect
		Retour		: Adresse du vecteur affect
----------------------------------------------------------------------*/
double *cpyvec (double* u, double* v, int n)
{
    if (n > 0) while (n--) u[n] = v[n];
	return u;
}

/*---------------------------------------------------------------------
	Echange de 2 vecteurs de Rn : U <-> V
	-------------------------------------

	Entrees:
		double *u	: Adresse du 1er vecteur
		double *v	: Adresse du 2eme vecteur donn
		int n		: dimension de u et v

	Sorties:
		u et v changs
		Retour		: Adresse de u
----------------------------------------------------------------------*/
double *swapvec (double* u, double* v, int n)
{
double t;
if (n > 0) while (n--) {t=u[n]; u[n]=v[n]; v[n]=t;}
	return u;
}
/*---------------------------------------------------------------------
	Ngation d'un vecteur de Rn : U = -V
	------------------------------------

	Entrees:
		double *u	: Adresse du vecteur affect
		double *v	: Adresse vecteur donn
		int n		: dimension de v

	Sorties:
		u est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V
----------------------------------------------------------------------*/
double *negvec (double* u, double* v, int n)
{
    if (n > 0) while (n--) u[n] = -v[n];
	return u;
}

/*---------------------------------------------------------------------
	Addition de vecteurs de Rn : U = V + W
	--------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *addvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] + w[n];
	return u;
}

/*---------------------------------------------------------------------
	Soustraction de vecteurs de Rn : U = V - W
	------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *subvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] - w[n];
	return u;
}

/*---------------------------------------------------------------------
	Addition de vecteurs de Rn : U = V + W
	--------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double s	: Scalaire donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *cumvec (double* u, double* v, double s, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] + s*w[n];
	return u;
}

/*---------------------------------------------------------------------
	Multiplication terme  terme de vecteurs de Rn : U = V * W
	----------------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *mulvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] * w[n];
	return u;
}

/*---------------------------------------------------------------------
	Division terme  terme de vecteurs de Rn : U = V / W
	----------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W

    ATTENTION : DIVISION PAR ZERO NON TESTEE !!!!!!!!!!!!!!!
----------------------------------------------------------------------*/
double *divvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] / w[n];
	return u;
}

/*---------------------------------------------------------------------
    Valeur absolue d'un vecteurs de Rn : U = |V|
	--------------------------------------------

	Entrees:
		double *u	: Adresse du vecteur affect
		double *v	: Adresse vecteur donn
		int n		: dimension des vecteurs

	Sorties:
		u est affect
		Retour		: Adresse du vecteur affect

    Remarque : u et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *absvec (double* u, double* v, int n)
{
    if (n > 0) while (n--) u[n] = fabs(v[n]);
	return u;
}

/*---------------------------------------------------------------------
	Composante maximale d'un vecteur de Rn : s = max(Ui)
	-----------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur


	Sorties:
		Retour		: Composante maximale du vecteur
----------------------------------------------------------------------*/
double xmaxvec (double* u, int n)
{
double s = -DBL_MAX;
    if (n > 0) while (n--) if (u[n]>s) s = u[n];
	return s;
}

/*---------------------------------------------------------------------
	Composante max. en val. abs. d'un vecteur de Rn : s = max(|Ui|)
	---------------------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		Retour		: Composante max. du vecteur, en val. abs.
----------------------------------------------------------------------*/
double xamaxvec (double* u, int n)
{
double s = -DBL_MAX;
    if (n > 0) while (n--) if (fabs(u[n]) > s) s = fabs(u[n]);
	return s;
}


/*---------------------------------------------------------------------
	index du + grand lment (0  n-1)
	----------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		Retour		: index
----------------------------------------------------------------------*/
int imaxvec (double *u, int n)
{
int i, id ;
double dmax;

    if (n <= 1) return (n-1) ;
    dmax = u[0] ; id = 0 ;
    for (i = 1 ; i < n ; i++)
    { if (u[i] > dmax) {id = i ; dmax = u[i] ; } }
    return id ;
}

/*---------------------------------------------------------------------
	index du + grand lment en valeur absolue (0  n-1)
	----------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		Retour		: index
----------------------------------------------------------------------*/
int iamaxvec (double *u, int n)
{
int i, id ;
double dmax, a ;

    if (n <= 1) return (n-1) ;
    dmax = fabs(u[0]) ; id = 0 ;
    for (i = 1 ; i < n ; i++)
    { a = fabs(u[i]) ; if (a > dmax) {id = i ; dmax = a ; } }
    return id ;
}

/*---------------------------------------------------------------------
	index du + petit lment en valeur absolue (0  n-1)
	----------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		Retour		: index
----------------------------------------------------------------------*/
int iaminvec (double *u, int n)
{
int i, id ;
double dmin, a ;

    if (n <= 1) return (n-1) ;
    dmin = fabs(u[0]) ; id = 0 ;
    for (i = 1 ; i < n ; i++)
    { a = fabs(u[i]) ; if (a < dmin) {id = i ; dmin = a ; } }
    return id ;
}

/*---------------------------------------------------------------------
	Composante minimale d'un vecteur de Rn : s = max(Ui)
	-----------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur


	Sorties:
		Retour		: Composante minimale du vecteur
----------------------------------------------------------------------*/
double xminvec (double* u, int n)
{
double s = DBL_MAX;
    if (n > 0) while (n--) if (u[n]<s) s = u[n];
	return s;
}

/*---------------------------------------------------------------------
	Composante min. en val. abs. d'un vecteur de Rn : s = min(|Ui|)
	---------------------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		Retour		: Composante min. du vecteur, en val. abs.
----------------------------------------------------------------------*/
double xaminvec (double* u, int n)
{
double s = DBL_MAX;
    if (n > 0) while (n--) if (fabs(u[n]) < s) s = fabs(u[n]);
	return s;
}

/*---------------------------------------------------------------------
	index du + petit lment (0  n-1)
	----------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		Retour		: index
----------------------------------------------------------------------*/
int iminvec (double *u, int n)
{
int i, id ;
double dmin;

    if (n <= 1) return (n-1) ;
    dmin = u[0] ; id = 0 ;
    for (i = 1 ; i < n ; i++)
    { if (u[i] < dmin) {id = i ; dmin = u[i] ; } }
    return id ;
}

/*---------------------------------------------------------------------
	Somme des composantes d'un vecteurs de Rn : s = Sigma Ui
	--------------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		Retour		: Somme des composantes
----------------------------------------------------------------------*/
double xsumvec (double* u, int n)
{
double s = 0.;
    if (n > 0) while (n--) s += u[n];
	return s;
}

/*---------------------------------------------------------------------
	Longueur d'un vecteurs de Rn : s = sqrt(U'*U)
	---------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		Retour		: Longueur du vecteur
----------------------------------------------------------------------*/
double longvec (double* u, int n)
{
double s = 0.;
    if (n > 0) while (n--) s += u[n]*u[n];
	return sqrt(s);
}

/*---------------------------------------------------------------------
	Normalisation d'un vecteur de Rn : U = V / ||V||
	------------------------------------------------

	Entrees:
		double *U	: Adresse vecteur affect
		double *V	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		U est affect
		Retour		: Adresse vecteur affect

	Remarque : U et V peuvent occuper la mme mmoire
		       si ||V|| = 0  -> U[0] = 1, les autres nuls.

	Remarque : U peut occuper la mme mmoire que V
----------------------------------------------------------------------*/
double *unitvec (double* u, double *v, int n)
{
int m = n ;
double s = 0.;
    if (n > 0)
	{
		while (n--) s += u[n]*u[n];
		if (s==0.) u[0] = 1.;
		else
		{
			s = sqrt(s) ;
			while (m--) u[m] = v[m]/s;
		}
	}
	return u;
}

/*---------------------------------------------------------------------
	Normalisation d'un vecteur de Rn : U = V / ||V||
	------------------------------------------------

	Entrees:
		double *U	: Adresse vecteur affect
		double *V	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		U est affect
		Retour		: Longueur du vecteur V

	Remarque : U et V peuvent occuper la mme mmoire
		       si ||V|| = 0  -> U[0] = 1, les autres nuls.

	Remarque : U peut occuper la mme mmoire que V
----------------------------------------------------------------------*/
double lunitvec (double* u, double *v, int n)
{
int m = n ;
double s = 0.;
    if (n > 0)
	{
		while (n--) s += u[n]*u[n];
		if (s==0.)
		{ u[0] = 1. ; while (--m) u[m] = 0. ; }
		else
		{
			s = sqrt(s) ;
			while (m--) u[m] = v[m]/s;
		}
	}
	return s;
}

/*---------------------------------------------------------------------
	Multiplication d'un vecteur de Rn par un scalaire : U = aV
	----------------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double a	: Scalaire donn
		double *V	: Adresse du vecteur de Rn donn
		int n		: dimension du vecteur

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V
----------------------------------------------------------------------*/
double *scavec (double *u, double a, double *v, int n)
{
    if (n > 0) while (n--) u[n] = a * v[n] ;
	return u;
}

/*---------------------------------------------------------------------
	Produit scalaire de vecteurs de Rn : s = <U . V>
	------------------------------------------------

	Entrees:
		double *U	: Adresse du premier vecteur de Rn donn
		double *V	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		Retour		: produit scalaire
----------------------------------------------------------------------*/
double dotvec (double* u, double* v, int n)
{
double s ;
    s = 0. ;
    if (n > 0) while (n--) s += u[n] * v[n] ;
    return(s);
}

/*---------------------------------------------------------------------
	Carr d'un vecteurs de Rn : s = U'*U
	------------------------------------

	Entrees:
		double *U	: Adresse vecteur donn
		int n		: dimension des vecteurs


	Sorties:
		Retour		: Somme des carrs des composantes
----------------------------------------------------------------------*/
double carrevec (double* u, int n)
{
double s ;
    s = 0. ;
    if (n > 0) while (n--) s += u[n] * u[n] ;
    return(s);
}

/*---------------------------------------------------------------------
	Produit tensoriel de vecteurs de Rn : A = B * C'
	--------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice de Rn affecte
		double *B	: Adresse du premier vecteur de Rn donn
		double *C	: Adresse du deuxime vecteur de Rn donn
		int	m		: dimension du premier vecteur
		int p		: dimension du deuxime vecteur

	Sorties:
		A est affecte (dimension m * p)
		Retour		: Adresse de la matrice de Rn affecte
----------------------------------------------------------------------*/
double *tensvec (double *A, double *B, double *C, int m, int p)
{
double *a, *b, *c, r ;
int i, j /* , k */;


    for (i = 0 ; i < m ; ++i)
    {
	a = A + i ;      			/* A(i,0) */
	c = C ;                                 /* C(0,0) */
	for (j = 0 ; j < p ; ++j)
	{
	    r = 0 ;
	    b = B + i ;				/* B(i,0) */
	  /*  for (k = 0 ; k < 1 ; ++k) */
	    {
		r += *b * *c++ ;		/* B(i,k) * C(k,j) */
		b += m ;
	    }
	    *a = r ;
	    a += m ;				/* A(i,j) */
	}
    }
	return A;
}

/*---------------------------------------------------------------------
	Gnration d'une matrice avec des 1 sur la diagonale
	----------------------------------------------------

	Entrees:
		double *A	: Adresse de A
		int    li	: Nombre de lignes de A
		int    co	: Nombre de colonnes de A

    Sortie :
		A est calcule
		Retour	: Adresse de A			:
----------------------------------------------------------------------*/
double *eyemat(double *a, int li, int co)
{
int i,j, k = 0;

	for(j = 0 ; j < co; j++)
		for (i = 0 ; i < li ; i++) a[k++] = (i==j ? 1. : 0.) ;
	return a ;
}

/*---------------------------------------------------------------------
	Gnration d'une matrice diagonale A  partir d'un vecteur V
	------------------------------------------------------------

	Entrees:
		double *A	: Adresse de A
		double *V   : Adresse de V
		int    li	: Nombre de lignes de A
		int    co	: Nombre de colonnes de A

    Sortie :
		A est calcule
		Retour	: Adresse de A			:
----------------------------------------------------------------------*/
double *diag2mat(double *a, double *v, int li, int co)
{
int i,j, k = 0, l = 0;

	for(j = 0 ; j < co; j++)
		for (i = 0 ; i < li ; i++)
		{
			if (i == j) a[k++] = v[l++] ;
			else a[k++] = 0.;
		}
	return a ;
}

/*---------------------------------------------------------------------
	Extraction de la diagonale d'une matrice diagonale
	--------------------------------------------------

	Entrees:
		double *V   : Adresse de V affect
		double *A	: Adresse de A
		int    li	: Nombre de lignes de A
		int    co	: Nombre de colonnes de A

    Sortie :
		V est calcule
		Retour	: Adresse de V			:
----------------------------------------------------------------------*/
double *mat2diag(double *v, double *a, int li, int co)
{
int i, lico = li*co, li1 = li+1, k = 0 ;

	for (i = 0 ; i < lico; i += li1) v[k++] = a[i];
	return v ;
}

/*---------------------------------------------------------------------
	Gnration colonne contenant la comp. max de chaque ligne
	---------------------------------------------------------

	Entrees:
		double *V   : Adresse de V colonne affect
		double *A	: Adresse de A
		int    li	: Nombre de lignes de A
		int    co	: Nombre de colonnes de A

    Sortie :
		V est calcule
		Retour	: Adresse de V			:
----------------------------------------------------------------------*/
double *cmaxmat(double *v, double *a, int li, int co)
{
int i, j, k = 0 ;
double s ;

	cpyvec(v,a,li) ;
	k = li ;
	for (j = 1 ; j < co; j++)
	{
		for (i = 0 ; i < li ; i++)
		{
			s = a[k++] ; if (s > v[i]) v[i] = s;
		}
	}
	return v ;
}

/*---------------------------------------------------------------------
	Gnration colonne contenant la comp. min de chaque ligne
	---------------------------------------------------------

	Entrees:
		double *V   : Adresse de V colonne affect
		double *A	: Adresse de A
		int    li	: Nombre de lignes de A
		int    co	: Nombre de colonnes de A

    Sortie :
		V est calcule
		Retour	: Adresse de V			:
----------------------------------------------------------------------*/
double *cminmat(double *v, double *a, int li, int co)
{
int i, j, k = 0 ;
double s;

	cpyvec(v,a,li) ;
	k = li ;
	for (j = 1 ; j < co; j++)
	{
		for (i = 0 ; i < li ; i++)
		{
			s = a[k++] ; if (s < v[i]) v[i] = s;
		}
	}
	return v ;
}

/*---------------------------------------------------------------------
	Gnration colonne contenant la somme des comp. de chaque ligne
	---------------------------------------------------------------

	Entrees:
		double *V   : Adresse de V colonne affect
		double *A	: Adresse de A
		int    li	: Nombre de lignes de A
		int    co	: Nombre de colonnes de A

    Sortie :
		V est calcule
		Retour	: Adresse de V			:
----------------------------------------------------------------------*/
double *csummat(double *v, double *a, int li, int co)
{
int i, j, k = 0 ;

	cpyvec(v,a,li) ;
	k = li ;
	for (j = 1 ; j < co; j++)
	{
		for (i = 0 ; i < li ; i++)
		{
			v[i] += a[k++];
		}
	}
	return v ;
}

/*---------------------------------------------------------------------
	Gnration ligne contenant la comp. max de chaque colonne
	---------------------------------------------------------

	Entrees:
		double *V   : Adresse de V ligne affect
		double *A	: Adresse de A
		int    li	: Nombre de lignes de A
		int    co	: Nombre de colonnes de A

    Sortie :
		V est calcule
		Retour	: Adresse de V			:
----------------------------------------------------------------------*/
double *lmaxmat(double *v, double *a, int li, int co)
{
int i, j, k = 0 ;

	for (j = 0 ; j < co; j++) { v[j] = a[k] ; k += li ;}
	for (i = 1 ; i < li ; i++)
	{
		k = i ;
		for (j = 0 ; j < co; j++)
		{
			if (a[k] > v[j]) v[j] = a[k] ;
			k += li ;
		}
	}
	return v ;
}

/*---------------------------------------------------------------------
	Gnration ligne contenant la comp. min de chaque colonne
	---------------------------------------------------------

	Entrees:
		double *V   : Adresse de V ligne affect
		double *A	: Adresse de A
		int    li	: Nombre de lignes de A
		int    co	: Nombre de colonnes de A

    Sortie :
		V est calcule
		Retour	: Adresse de V			:
----------------------------------------------------------------------*/
double *lminmat(double *v, double *a, int li, int co)
{
int i, j, k = 0 ;

	for (j = 0 ; j < co; j++) { v[j] = a[k] ; k += li ;}
	for (i = 1 ; i < li ; i++)
	{
		k = i ;
		for (j = 0 ; j < co; j++)
		{
			if (a[k] < v[j]) v[j] = a[k] ;
			k += li ;
		}
	}
	return v ;
}

/*---------------------------------------------------------------------
	Gnration ligne contenant la somme des comp.de chaque colonne
	--------------------------------------------------------------

	Entrees:
		double *V   : Adresse de V ligne affect
		double *A	: Adresse de A
		int    li	: Nombre de lignes de A
		int    co	: Nombre de colonnes de A

    Sortie :
		V est calcule
		Retour	: Adresse de V			:
----------------------------------------------------------------------*/
double *lsummat(double *v, double *a, int li, int co)
{
int i, j, k = 0 ;

	for (j = 0 ; j < co; j++) { v[j] = a[k] ; k += li ;}
	for (i = 1 ; i < li ; i++)
	{
		k = i ;
		for (j = 0 ; j < co; j++)
		{
			v[j] += a[k] ;
			k += li ;
		}
	}
	return v ;
}

/*---------------------------------------------------------------------
	Produit de matrice : Amp = Bmn x Cnp
	------------------------------------

	Entrees:
		double *A	: Adresse de A
		double *B	: Adresse de B
		double *C	: Adresse de C
		int    m	: Nombre de lignes de A et de B
		int    n	: Nombre de colonnes de B et de lignes de C
		int    p	: Nombre de colonnes de A et de C

    Sortie :
		A est calcule
		Retour	: Adresse de A			:

    Remarque : premire et dernire dimension : celles du rsultat
	A ne doit pas occuper la mme zone mmoire que B ou C
----------------------------------------------------------------------*/
double *matmat (double *A, double *B, double *C, int m, int n, int p)
{
double *a, *b, *c, r ;
int i, j, k ;


    for (i = 0 ; i < m ; ++i)
    {
	a = A + i ;      			/* A(i,0) */
	c = C ;                                 /* C(0,0) */
	for (j = 0 ; j < p ; ++j)
	{
	    r = 0 ;
	    b = B + i ;				/* B(i,0) */
	    for (k = 0 ; k < n ; ++k)
	    {
		r += *b * *c++ ;		/* B(i,k) * C(k,j) */
		b += m ;
	    }
	    *a = r ;
	    a += m ;				/* A(i,j) */
	}
    }
	return A;
}

/*---------------------------------------------------------------------
	Produit de matrice : Amp = Bmn x [C']np
	---------------------------------------

	Entrees:
		double *A	: Adresse de A
		double *B	: Adresse de B
		double *C	: Adresse de C
		int    m         : Nombre de lignes de A et de lignes de B
		int    n         : Nombre de colonnes de B et de colonnes de C
		int    p         : Nombre de colonnes de A et de lignes de C

    Sortie :
		A est calcule
		Retour	: Adresse de A			:

    Remarque : premire et dernire dimension : celles du rsultat
	A ne doit pas occuper la mme zone mmoire que B ou C
----------------------------------------------------------------------*/
double *mattmat (double *a, double *b, double *c, int m, int n, int p)
{
int i, j, k, ij, ik, jk ;

    ij = -1 ;
    for (j = 0 ; j < p ; j++)
	for (i = 0 ; i < m ; i++)
	{
	    a[++ij] = 0. ;
		for (k = 0, ik = i, jk = j ; k < n ; k++, ik += m, jk += p)
		    a[ij] += b[ik]*c[jk] ;
	}
	return a;
}

/*---------------------------------------------------------------------
	Produit de matrice : Amp = [B']mn x Cnp
	---------------------------------------

	Entrees:
		double *A	: Adresse de A
		double *B	: Adresse de B
		double *C	: Adresse de C
		int    m         : Nombre de lignes de A et de colonnes de B
		int    n         : Nombre de lignes de B et  de C
		int    p         : Nombre de colonnes de A et de C

    Sortie :
		A est calcule
		Retour	: Adresse de A			:

    Remarque : premire et dernire dimension : celles du rsultat
	A ne doit pas occuper la mme zone mmoire que B ou C
----------------------------------------------------------------------*/
double *tmatmat (double *a, double *b, double *c, int m, int n, int p)
{
int i, j, k, ij, ki, kj, kj0 ;

    for (j = 0, ij = 0, kj0 = 0 ; j < p ; j++, kj0 += n)
    {
	for (i = 0, ki = 0 ; i < m ; i++, ij++)
	{
	    a[ij] = 0. ;
	    for (k = 0, kj = kj0 ; k < n ; k++) a[ij] += b[ki++] * c[kj++] ;
	}
    }
	return a;
}

/*---------------------------------------------------------------------
    Produit de matrice : Amm = [B']mn x Bnm
    ---------------------------------------
    Entrees:
	double *A        : Adresse de A
	double *B        : Adresse de B
	int    m         : Nombre de lignes de A et de colonnes de B
	int    n         : Nombre de lignes de B

    Sortie :
	A est calculee

    Attention : par convention MLL, les elements successifs en memoire
    ---------   sont supposes etre successifs en colonne.

    Attention : le tableau A ne doit pas occuper la meme place que le
    ---------   tableau B.

----------------------------------------------------------------------*/
void tAAmat (double *a, double *b, int m, int n)
{
int i, j, k, ij0, ij, mp1 = m+1, ki00, ki0, ki, kj0, kj, ji0, ji ;

    for (j = 0, ij0 = 0   , ki00 = 0 , kj0 = 0 ; j < m ;
	 j++  , ij0 += mp1, ki00 += n, kj0 += n)
	for (i = j, ij = ij0, ki0 = ki00 ; i < m ; i++, ij++, ki0 += n)
	{
	    a[ij] = 0. ;
	    for (k = 0, ki = ki0, kj = kj0 ; k < n ; k++)
		a[ij] += b[ki++] * b[kj++] ;
	}
    for (j = 1, ij0 = m, ji0 = 1 ; j < m ; j++, ij0 += m, ji0++)
	for (i = 0, ij = ij0, ji = ji0 ; i < j ; i++, ij++, ji += m)
	    a[ij] = a[ji] ;
}

/*---------------------------------------------------------------------
    Produit de matrice : Ann = Bnm x [B']mn
    ---------------------------------------
    Entrees:
	double *A        : Adresse de A
	double *B        : Adresse de B
	int    n         : Nombre de lignes de A et de B
	int    m         : Nombre de colonnes de B

    Sortie :
	A est calculee

    Attention : par convention MLL, les elements successifs en memoire
    ---------   sont supposes etre successifs en colonne.

    Attention : le tableau A ne doit pas occuper la meme place que le
    ---------   tableau B.

----------------------------------------------------------------------*/
void AAtmat (double *a, double *b, int n, int m)
{
int i, j, k, ij0, ij, np1 = n+1, ik, jk ,ji, ji0;

    for (i=ij0=0;i<n;i++,ij0+=np1)
	for (j=i,ij=ij0;j<n;j++,ij+=n)
	{
	    a[ij] = 0. ;
	    for (k=0,ik=i,jk=j; k<m ;k++,ik+=n, jk+=n) a[ij] += b[ik] * b[jk] ;
	}
    for (i=1,ji0=n;i<n;i++,ji0+=n)
	for (j=0, ij=i,ji=ji0;j<i;j++,ij+=n) a[ij] = a[ji++];
}


/*---------------------------------------------------------------------
	Trace d'une matrice
	-------------------

	Entrees:
		double *A	: Adresse de A
		int    li	: Nombre de lignes de A
		int    co	: Nombre de colonnes de A

    Sortie :
		Retour	: trace de A			:
----------------------------------------------------------------------*/
double tracemat(double *a, int li, int co)
{
int i, lico = li*co, li1 = li+1 ;
double s = 0.;

	for (i = 0 ; i < lico; i += li1) s += a[i];
	return s ;
}

/*---------------------------------------------------------------------
	Transpose d'une matrice : A = B'
	----------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double *B	: Adresse de la matrice donne
		int m		: Nombre de lignes de A, colonnes de B
		int n		: Nombre de colonnes de A, lignes de B

	Sorties:
		A est transpose
		Retour		: Adresse de A

    ATTENTION : A et B ne doivent pas occuper la mme zone mmoire
----------------------------------------------------------------------*/
double *trmat(double *a, double *b, int m, int n)
{
int j , ij = 0, ji, mn = m*n ;
    for (j = 0 ; j < n ; j++)
    for (ji = j ; ji < mn ; ji += n) a[ij++] = b[ji] ;
	return a ;
}

/*---------------------------------------------------------------------
	Transpose d'une matrice carre : A = A'
	----------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		int n		: Dimension de la matrice

	Sorties:
		A est transpose
		Retour		: Adresse de A
----------------------------------------------------------------------*/
double *tmat(double *a, int n)
{
int i,in,j,jn;
double s ;

	for (i = 1, in = n ; i < n ; i++, in += n)
		for (j = 0, jn = 0 ; j < i ; j++, jn += n)
		{s = a[i+jn] ; a[i+jn]=a[j+in]; a[j+in]=s;}

	return a ;
}

