#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "linsolve.h"
#include "eigen.h"
#include "racines.h"
#include "polynomes.h"

char *terreur;
#define DIM 6

#define TESTLINPACK

#ifdef TESTLINPACK 

double memoA[DIM*2*DIM] ;
double memoS[DIM*2*DIM] ;
double A[DIM*2*DIM] ;
double B[DIM*2*DIM] ;
double xy[2*DIM*DIM];
double I[DIM*2*DIM];
double Noy[DIM*2*DIM];
int jpvt[2*DIM];
double U[DIM*DIM], V[4*DIM*DIM], S[DIM];

#endif

//#define TESTEIGEN
#ifdef TESTEIGEN 

double memoA[DIM*DIM] ;
double A[DIM*DIM] ;
double B[DIM*DIM] ;
double wr[DIM+DIM] ;
double wrbis[DIM+DIM] ;
double I[DIM*DIM];
double M[DIM*DIM];
double IM[DIM*DIM];
#endif

//#define TESTROOTS
#ifdef TESTROOTS
double memoA[DIM+1] ;
double wr[DIM+DIM] ;
double wrbis[DIM+DIM] ;
double wrter[DIM+DIM] ;
#endif



int main(void)
{
/*--------------------------------------------------------------*/
#ifdef TESTROOTS
int i;

//	MessageBox(NULL,"Texte","Caption",0);
    srand (12745) ;
/* Matrice carree DIM x DIM aleatoire */
    for ( i = 0 ; i < DIM+DIM ; i++) 
		wr[i] = 3.9*rand() / RAND_MAX - 2.1;
	for (i = 1 ; i < DIM/2; i+=2)
	{
		wr[i-1]=wr[i]; wr[i+DIM-1]=-wr[i+DIM];
	}
	--i;
	for(;i<DIM;i++) wr[i+DIM]=0.; /* reelles */
	trmat(wrbis,wr,2,DIM);
	printmat("Racines initiales (RE/IM) :",wrbis,2,DIM);

	rac2pol (memoA, wr, wr+DIM, DIM);

	printmat("Polynome initial :",memoA,1,DIM+1);

	i = roots_newton (wr, wr+DIM, memoA, DIM) ;
	trmat(wrter,wr,2,DIM);
	printmat("Racines Newton (RE/IM) :",wrter,2,DIM);

	i = roots_qd (wr, wr+DIM, memoA, DIM) ;
	trmat(wrter,wr,2,DIM);
	printmat("Racines Rutihauser (RE/IM) :",wrter,2,DIM);

	i = roots_hqr (wr, wr+DIM, memoA, DIM) ;
	trmat(wrter,wr,2,DIM);
	printmat("Racines HQR (RE/IM) :",wrter,2,DIM);

	i = roots_balhqr (wr, wr+DIM, memoA, DIM) ;
	trmat(wrter,wr,2,DIM);
	printmat("Racines BALHQR (RE/IM) :",wrter,2,DIM);

	printmat("Polynome initial :",memoA,1,DIM+1);
//    i= roots_sturm (wr, memoA, DIM, 1.e-6) ;
    i= roots_sturmod (wr, memoA, DIM ,-100., 100., 1.e-6) ;
	printmat("Racines roots_sturm (RE) :",wr,1,i);
#endif
/*--------------------------------------------------------------*/

/*--------------------------------------------------------------*/
#ifdef TESTEIGEN
clock_t start, finish, duree;
int i, mode;
double err, errmax ;

    srand (12745) ;
/* Matrice carree DIM x DIM aleatoire */
    for ( i = 0 ; i < DIM*DIM ; i++) 
		memoA[i] = 3.9*rand() / RAND_MAX - 2.1;

	// printmat("Matrice initiale :",memoA,DIM,DIM);

	for (mode = 0; mode < 8; mode++)
	{
		cpyvec (A, memoA, DIM*DIM) ;
		i= eigvec (wr, wr+DIM, M, A, DIM, mode) ;
		if (i != 0)
		{
			printf("Abandon aprs calcul de %d val.p.\n",i);
			continue ;
		}
		trmat(wrbis,wr,2,DIM);
		//printmat("Valeurs propres (RE/IM) :",wrbis,2,DIM);
		//printmat("Vecteurs propres :",M,DIM,DIM);
		invmat(cpyvec (IM, M, DIM*DIM), DIM);
		matmat(B,A,IM,DIM,DIM,DIM);
		matmat(A,M,B,DIM,DIM,DIM);
		errmax = 0.; for (i=0; i<DIM*DIM;i++) {err = fabs(A[i]-memoA[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}

#endif
/*--------------------------------------------------------------*/
#ifdef TESTLINPACK

clock_t start, finish, duree;
int i, rank;
double err, errmax, det ;
double epsilon = 1.e-6;
int nlin = DIM, ncol;

    srand (12745) ;
/* Matrice carree DIM x DIM aleatoire */
    for ( i = 0 ; i < DIM*DIM ; i++) 
		memoA[i] = 3.9*rand() / RAND_MAX - 2.1;
    for (i = 0 ; i < DIM*DIM ; i++) 
		memoA[i+DIM*DIM] = memoA[i];
	tAAmat (memoS, memoA, DIM, DIM) ;

/*-----------------------------------------------------------------*/
	puts("------------------------------------------");
	puts("Resolutions systemes symetriques");
	ncol = DIM;

	cpyvec (A, memoS, DIM*DIM) ;
	printf("Determinant = %lg\n",detmat(A,DIM));

	cpyvec (A, memoS, DIM*DIM) ;
	eyemat (xy, DIM,DIM) ; /* Second membre = matrice identite */
    start = clock();
	det = solvchsq(xy, A, DIM, DIM);
	finish = clock(); duree = finish - start;
	printf("\nDuree Cholevski AR  = %d clocks\n", duree );
	printf("Determinant = %lg\n",det);
    if (det != 0.)
	{
		matmat (A, xy, memoS, DIM, DIM, DIM) ;
		eyemat (I, DIM,DIM) ; /* matrice identite */
		errmax = 0.; for (i=0; i<DIM*DIM;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}

	cpyvec (A, memoS, DIM*DIM) ;
	eyemat (xy, DIM,DIM) ; /* Second membre = matrice identite */
    start = clock();
	det = solvchud(xy, A, DIM, DIM);
	finish = clock(); duree = finish - start;
	printf("\nDuree Cholevski UDU = %d clocks\n", duree );
	printf("Determinant = %lg\n",det);
    if (det != 0.)
	{
		matmat (A, xy, memoS, DIM, DIM, DIM) ;
		eyemat (I, DIM,DIM) ; /* matrice identite */
		errmax = 0.; for (i=0; i<DIM*DIM;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}

	cpyvec (A, memoS, DIM*DIM) ;
	eyemat (xy, DIM,DIM) ; /* Second membre = matrice identite */
    start = clock();
    det = solvlu (xy, A, DIM, DIM, 1) ;
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition LU = %d clocks\n", duree );
	printf("Determinant = %lg\n",det);
    if (det!=0.) ;
	{
		matmat (A, xy, memoS, DIM, DIM, DIM) ;
		eyemat (I, DIM,DIM) ; /* Second membre = matrice identite */
		errmax = 0.; for (i=0; i<DIM*DIM;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}
/*-----------------------------------------------------------------*/
	puts("------------------------------------------");
	puts("Resolution systemes carres");
	ncol = DIM;

	cpyvec (A, memoA, nlin*ncol) ;
	mat2QR (B, A, nlin, ncol);
	matmat(I,B,A,nlin, nlin,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QR : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2QRE (B, A, jpvt, nlin, ncol, epsilon);
	printf("rang QRE : %d\n",rank);
	det = 1.; for(i=0;i<nlin*nlin;i+=nlin+1) det *=-A[i];
	printf("Determinant = %lg\n",det);
	matmat(I,B,A,nlin, nlin, ncol);
	permcol(I, jpvt, nlin, ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QRE : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2SVD (U, S, V, A,nlin,ncol, epsilon);
	printf("rang SVD : %d\n",rank);
	det = 1.; for(i=0;i<nlin;i++) det *=S[i];
	printf("Val. abs. determinant = %lg\n",det);
	diag2mat(I,S,nlin,ncol); matmat(xy,U,I,nlin,nlin,ncol);
	mattmat(I,xy,V,nlin,ncol,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur SVD : %lg\n",errmax);

	cpyvec (A, memoA, DIM*DIM) ;
	eyemat (xy, DIM,DIM) ; /* Second membre = matrice identite */
    start = clock();
	det = solvpi (xy, A, xy, nlin, ncol, ncol, epsilon) ;
	finish = clock(); duree = finish - start;
	printf("\nDuree presqu'inverse   = %d clocks\n", duree );
	printf("Determinant AAt+epsI = %lg\n",det);
    if (det != 0.)
	{
		matmat (A, xy, memoA, DIM, DIM, DIM) ;
		eyemat (I, DIM,DIM) ; /* matrice identite */
		errmax = 0.; for (i=0; i<DIM*DIM;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}

	cpyvec (A, memoA, DIM*DIM) ;
	eyemat (xy, DIM,DIM) ; /* Second membre = matrice identite */
    start = clock();
    det = solvlu (xy, A, DIM, DIM, 0) ;
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition LU = %d clocks\n", duree );
	printf("Determinant = %lg\n",det);
    if (det!=0.) ;
	{
		matmat (A, xy, memoA, DIM, DIM, DIM) ;
		eyemat (I, DIM,DIM) ; /* Second membre = matrice identite */
		errmax = 0.; for (i=0; i<DIM*DIM;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}

	cpyvec (A, memoA, DIM*DIM) ;
	eyemat (xy, DIM,DIM) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvqr (xy, A, DIM, DIM, DIM, epsilon) ;
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition QR = %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
	det = 1.; for(i=0;i<nlin*nlin;i+=nlin+1) det *=-A[i];
	printf("Determinant = %lg\n",det);
	matmat (A, xy, memoA, DIM, DIM, DIM) ;
	eyemat (I, DIM,DIM) ; /* matrice identite */
	errmax = 0.; for (i=0; i<DIM*DIM;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur maximale : %lg\n",errmax);

	cpyvec (A, memoA, DIM*DIM) ;
	eyemat (xy, DIM,DIM) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvsvd (xy, A, DIM, DIM, DIM, epsilon) ; 
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition SVD= %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
	det = 1.; for(i=0;i<nlin*nlin;i+=nlin+1) det *=A[i];
	printf("Val. abs. determinant = %lg\n",det);
    if (rank>0)
	{
		matmat (A, xy, memoA, DIM, DIM, DIM) ;
		eyemat (I, DIM,DIM) ; /* matrice identite */
		errmax = 0.; for (i=0; i<DIM*DIM;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}

/*-----------------------------------------------------------------*/
	puts("------------------------------------------");
	puts("Resolution systemes surdetermines");
	ncol = (3*nlin)/4;

	cpyvec (A, memoA, nlin*ncol) ;
	mat2QR (B, A, nlin, ncol);
	matmat(I,B,A,nlin, nlin,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QR : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2QRE (B, A, jpvt, nlin, ncol, epsilon);
	printf("rang QRE : %d\n",rank);
	matmat(I,B,A,nlin, nlin,ncol);
	permcol(I, jpvt, nlin, ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QRE : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2SVD (U, S, V, A,nlin,ncol, epsilon);
	printf("rang SVD : %d\n",rank);
	diag2mat(I,S,nlin,ncol); matmat(xy,U,I,nlin,nlin,ncol);
	mattmat(I,xy,V,nlin,ncol,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur SVD : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	eyemat (xy, nlin, nlin) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvqr (xy, A, nlin, ncol, nlin, epsilon) ;
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition QR = %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
	matmat (A, xy, memoA, ncol, nlin, ncol) ;
    eyemat (I, ncol, ncol) ; 
	errmax = 0.; for (i=0; i<ncol*ncol;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur maximale : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	eyemat (xy, nlin,nlin) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvsvd (xy, A, nlin, ncol, nlin, epsilon) ; 
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition SVD= %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
    if (rank>0)
	{
		matmat (A, xy, memoA, ncol, nlin, ncol) ;
		eyemat (I, ncol,ncol) ; 
		errmax = 0.; for (i=0; i<ncol*ncol;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}

/*-----------------------------------------------------------------*/
	puts("------------------------------------------");
	puts("Resolution systemes redondants.");
	ncol = (4*nlin)/3;

	cpyvec (A, memoA, nlin*ncol) ;
	mat2QR (B, A, nlin, ncol);
	matmat(I,B,A,nlin, nlin,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QR : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2QRE (B, A, jpvt, nlin, ncol, epsilon);
	printf("rang QRE : %d\n",rank);
	matmat(I,B,A,nlin, nlin,ncol);
	permcol(I, jpvt, nlin, ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QRE : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2SVD (U, S, V, A,nlin,ncol, epsilon);
	printf("rang SVD : %d\n",rank);
	diag2mat(I,S,nlin,ncol); matmat(xy,U,I,nlin,nlin,ncol);
	mattmat(I,xy,V,nlin,ncol,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur SVD : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	eyemat (xy, nlin,nlin) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvqr (xy, A, nlin, ncol, nlin, epsilon) ;
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition QR = %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
	matmat (A, memoA, xy, nlin, ncol, nlin) ;
	eyemat (I, nlin,nlin) ; /* matrice identite */
	errmax = 0.; for (i=0; i<nlin*nlin;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur maximale : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	eyemat (xy, nlin,nlin) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvsvd (xy, A, nlin, ncol, nlin, epsilon) ; 
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition SVD= %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
    if (rank>0)
	{
		matmat (A, memoA, xy, nlin, ncol, nlin) ;
		eyemat (I, nlin,nlin) ; /* Second membre = matrice identite */
		errmax = 0.; for (i=0; i<nlin*nlin;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}


/*-----------------------------------------------------------------*/
	for (i=0;i+3<DIM; i += 4)
		cpyvec(&memoA[(i+3)*DIM],&memoA[i*DIM],DIM);

	puts("------------------------------------------");
	puts("Resolution systemes carres SINGULIERS");
	ncol = DIM;

	cpyvec (A, memoA, nlin*ncol) ;
	mat2QR (B, A, nlin, ncol);
	matmat(I,B,A,nlin, nlin,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QR : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2QRE (B, A, jpvt, nlin, ncol, epsilon);
	printf("rang QRE : %d\n",rank);
	det = 1.; for(i=0;i<nlin*nlin;i+=nlin+1) det *=-A[i];
	printf("Determinant = %lg\n",det);
	matmat(I,B,A,nlin, nlin,ncol);
	permcol(I, jpvt, nlin, ncol); 
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QRE : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2SVD (U, S, V, A,nlin,ncol, epsilon);
	printf("rang SVD : %d\n",rank);
	det = 1.; for(i=0;i<nlin;i++) det *=S[i];
	printf("Val. abs. determinant = %lg\n",det);
	diag2mat(I,S,nlin,ncol); matmat(xy,U,I,nlin,nlin,ncol);
	mattmat(I,xy,V,nlin,ncol,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur SVD : %lg\n",errmax);

	cpyvec (A, memoA, DIM*DIM) ;
	eyemat (xy, DIM,DIM) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvqr (xy, A, DIM, DIM, DIM, epsilon) ;
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition QR = %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
	det = 1.; for(i=0;i<nlin*nlin;i+=nlin+1) det *=-A[i];
	printf("Determinant = %lg\n",det);
	cpyvec(A,memoA,DIM*DIM);
	rank = nullqr (Noy, A, DIM, DIM, epsilon);
	AAtmat(B,Noy,DIM,DIM-rank);
	matmat (A, memoA, xy, DIM, DIM, DIM) ;
	addvec(A,A,B,DIM*DIM);
	eyemat (I, DIM,DIM) ; /* matrice identite */
	errmax = 0.; for (i=0; i<DIM*DIM;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur maximale : %lg\n",errmax);

	cpyvec (A, memoA, DIM*DIM) ;
	eyemat (xy, DIM,DIM) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvsvd (xy, A, DIM, DIM, DIM, epsilon) ; 
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition SVD= %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
	det = 1.; for(i=0;i<nlin*nlin;i+=nlin+1) det *=A[i];
	printf("Val. abs. determinant = %lg\n",det);
    if (rank>0)
	{
		cpyvec (A, memoA, DIM*DIM) ;
		rank = nullsvd (Noy, A, DIM, DIM, epsilon,1);
		AAtmat(B,Noy,DIM,DIM-rank);
		matmat (A, memoA, xy, DIM, DIM, DIM) ;
		addvec(A,A,B,DIM*DIM);
		eyemat (I, DIM,DIM) ; /* matrice identite */
		errmax = 0.; for (i=0; i<DIM*DIM;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}

/*-----------------------------------------------------------------*/
	puts("------------------------------------------");
	puts("Resolution systemes surdetermines SINGULIERS");
	ncol = (3*nlin)/4;

	cpyvec (A, memoA, nlin*ncol) ;
	mat2QR (B, A, nlin, ncol);
	matmat(I,B,A,nlin, nlin,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QR : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2QRE (B, A, jpvt, nlin, ncol, epsilon);
	printf("rang QRE : %d\n",rank);
	matmat(I,B,A,nlin, nlin,ncol);
	permcol(I, jpvt, nlin, ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QRE : %lg\n",errmax);
	
	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2SVD (U, S, V, A,nlin,ncol, epsilon);
	printf("rang SVD : %d\n",rank);
	diag2mat(I,S,nlin,ncol); matmat(xy,U,I,nlin,nlin,ncol);
	mattmat(I,xy,V,nlin,ncol,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur SVD : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	eyemat (xy, nlin, nlin) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvqr (xy, A, nlin, ncol, nlin, epsilon) ;
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition QR = %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
	cpyvec(A,memoA,nlin*ncol);
	rank = nullqr (Noy, A, nlin, ncol, epsilon);
    printf ("Rang de la matrice = %d\n",rank) ;
	AAtmat(B,Noy,nlin,nlin-rank);
	matmat (A, memoA, xy, nlin, ncol, nlin) ;
	addvec(A,A,B,nlin*nlin);
	eyemat (I, nlin,nlin) ; /* matrice identite */
	errmax = 0.; for (i=0; i<nlin*nlin;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur maximale : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	eyemat (xy, nlin,nlin) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvsvd (xy, A, nlin, ncol, nlin, epsilon) ; 
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition SVD= %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
    if (rank>0)
	{
		cpyvec(A,memoA,nlin*ncol);
		rank = nullsvd (Noy, A, nlin, ncol, epsilon,1);
		printf ("Rang de la matrice = %d\n",rank) ;
		AAtmat(B,Noy,nlin,nlin-rank);
		matmat (A, memoA, xy, nlin, ncol, nlin) ;
		addvec(A,A,B,nlin*nlin);
		eyemat (I, nlin,nlin) ; /* matrice identite */
		errmax = 0.; for (i=0; i<nlin*nlin;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}

/*-----------------------------------------------------------------*/
	puts("------------------------------------------");
	puts("Resolution systemes redondants SINGULIERS.");
	ncol = (4*nlin)/3;

	cpyvec (A, memoA, nlin*ncol) ;
	mat2QR (B, A, nlin, ncol);
	matmat(I,B,A,nlin, nlin,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QR : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2QRE (B, A, jpvt, nlin, ncol, epsilon);
	matmat(I,B,A,nlin, nlin,ncol);
	permcol(I, jpvt, nlin, ncol); 
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur QRE : %lg\n",errmax);
	printf("rang QRE : %d\n",rank);

	cpyvec (A, memoA, nlin*ncol) ;
	rank = mat2SVD (U, S, V, A,nlin,ncol, epsilon);
	printf("rang SVD : %d\n",rank);
	diag2mat(I,S,nlin,ncol); matmat(xy,U,I,nlin,nlin,ncol);
	mattmat(I,xy,V,nlin,ncol,ncol);
	errmax = 0.; for (i=0; i<nlin*ncol;i++) {err = fabs(memoA[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur SVD : %lg\n",errmax);

	cpyvec (A, memoA, nlin*ncol) ;
	eyemat (xy, nlin,nlin) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvqr (xy, A, nlin, ncol, nlin, epsilon) ;
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition QR = %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
	cpyvec(A,memoA,nlin*ncol);
	rank = nullqr (Noy, A, nlin, ncol, epsilon);
    printf ("Rang de la matrice = %d\n",rank) ;
	AAtmat(B,Noy,nlin,nlin-rank);
	matmat (A, memoA, xy, nlin, ncol, nlin) ;
	addvec(A,A,B,nlin*nlin);
	eyemat (I, nlin,nlin) ; /* matrice identite */
	errmax = 0.; for (i=0; i<nlin*nlin;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
	printf("Erreur maximale : %lg\n",errmax);


	cpyvec (A, memoA, nlin*ncol) ;
	eyemat (xy, nlin,nlin) ; /* Second membre = matrice identite */
    start = clock();
    rank = solvsvd (xy, A, nlin, ncol, nlin, epsilon) ; 
	finish = clock(); duree = finish - start;
	printf("\nDuree decomposition SVD= %d clocks\n", duree );
    printf ("Rang de la matrice = %d\n",rank) ;
    if (rank>0)
	{
		cpyvec(A,memoA,nlin*ncol);
		rank = nullsvd (Noy, A, nlin, ncol, epsilon,1);
		printf ("Rang de la matrice = %d\n",rank) ;
		AAtmat(B,Noy,nlin,nlin-rank);
		matmat (A, memoA, xy, nlin, ncol, nlin) ;
		addvec(A,A,B,nlin*nlin);
		eyemat (I, nlin,nlin) ; /* matrice identite */
		errmax = 0.; for (i=0; i<nlin*nlin;i++) {err = fabs(A[i]-I[i]); if (err > errmax) errmax = err;} 
		printf("Erreur maximale : %lg\n",errmax);
	}
/*-----------------------------------------------------------------*/
#endif

// #define TESTQUATERN  
#ifdef TESTQUATERN
#include "quatern.h"
#include "rotation.h"
int i;
double mrota[9], mrotb[9], mrotc[9], mcou[9], mbid[9];
double qna[4], qnb[4], qnc[4], qbid[4], qcou[4];
double carola[3] = {0.15,-0.16,0.17};
double carolb[3] = {-0.25,0.26,-0.27};
double errat[3], d;
double q1[4], q2[4], qatt[4];
double v1[3], v2[3], vbid[3];

	/* angles carolA -> matrice */
	carol2mrot(mrota, carola); 
	/* matrice -> quatern */
	mrot2qn(qna, mrota);

	/* idem angles carolB */
	carol2mrot(mrotb, carolb); 
	mrot2qn(qnb, mrotb);

	/* Produit de matrice C = A*B */
	m3m3(mrotc,mrota,mrotb);
	
	/* Produit de quaternion C = A*B */	
	qnqn(qnc,qna,qnb); 


	printm3("",mrotc);
	qnu2mrot(mbid,qnc);
	printm3("",mbid);

	/* Erreur entre A et B : C = inv(A)*B */
	tm3m3(mrotc,mrota,mrotb);
	tqnqn(qnc,qna,qnb); 

	printm3("",mrotc);
	qnu2mrot(mbid,qnc);
	printm3("",mbid);

	asm3tov3(errat,mrotc);
	printv3("Erreur approche attitude matrice : ",errat);

	mrot2v3(errat,mrotc);
	printv3("Erreur exacte attitude matrice : ",errat);

	scav3(errat,2.,qnc+1);
	printv3("Erreur approche attitude quatern : ",errat);

	qnu2v3(errat,qnc);
	printv3("Erreur exacte attitude quatern : ",errat);

	/* changement de base */
	m3v3(vbid, mrotc, errat);
	printv3("V dans c :", vbid);

	qnuv3tqnu(vbid, qnc, errat);
	printv3("Q dans c :", vbid);

	/* Vecteur d'erreur */

#endif
}